@@ -44,7 +44,7 @@ WELCOME=`cat <<EOF
 Welcome to the SAMHAIN configuration tool
 
 This script is meant to make installing SAMHAIN as easy as
-possible.  Just read the text below, hit ENTER, and you're
+possible.  Just read the text below, hit ENTER, and you are
 on your way. 
 
 SAMHAIN ships with NO WARRANTY whatsoever, without
@@ -290,12 +290,12 @@ if [ "x$I386_LINUX" = "xyes" ]; then
 
 $dialog \
 --title "Kernel module rootkit detection" \
---separate-output \
 --backtitle "Samhain $VERSION" \
 --inputbox "SAMHAIN can detect kernel module rootkits if compiled with support\nfor this. If you want to enable this option, please give the path\nto your System.map file, else choose CANCEL.\n\nNOTE: this option will require root privileges for at least one\ncommand during compilation (to read from /dev/kmem)." \
 16 75 "/boot/System.map" \
 2> $tmp
 
+
 mtest=$?
 
 if [ $mtest = -1 ]; then
@@ -318,7 +318,6 @@ fi
 #------------------------------------------------------------
 $dialog \
 --title "Signed database and configuration" \
---separate-output \
 --backtitle "Samhain $VERSION" \
 --yesno "Samhain can be configured to support PGP signed database\nand configuration files. This requires a working installation\nof GnuPG.\n\nDo you want to use this option ?" \
 10 75 \
@@ -334,7 +333,6 @@ if [ $mtest = 0 ]; then
 
 $dialog \
 --title "Signed database and configuration" \
---separate-output \
 --backtitle "Samhain $VERSION" \
 --inputbox "Please enter the full path to gpg (i.e. the GnuPG binary)" \
 10 75 "/usr/bin/gpg" \
@@ -356,7 +354,6 @@ done
 
 $dialog \
 --title "Signed database and configuration" \
---separate-output \
 --backtitle "Samhain $VERSION" \
 --inputbox "Please enter the fingerprint of the key to use (one string, no spaces)" \
 10 75 "6BD9050FD8FC941B43412DCC68B7AB8957548DCD" \
@@ -384,7 +381,6 @@ fi
 #------------------------------------------------------------
 $dialog \
 --title "Stealth options" \
---separate-output \
 --backtitle "Samhain $VERSION" \
 --yesno "Samhain has some stealth options to hide its presence.\nDo you want to take advantage of these ?" \
 10 75 \
@@ -474,7 +470,6 @@ fi
 
 $dialog \
 --title "Stealth options" \
---separate-output \
 --backtitle "Samhain $VERSION" \
 --inputbox "You can set a magic string such that command line arguments will be ignored unless the first argument is this magic string, and read from stdin otherwise. If you do not want this, select CANCEL, otherwise choose a string and select OK." \
 10 75 "foo" \
@@ -494,7 +489,6 @@ fi
 if [ "x$I386_LINUX" = "xyes" ]; then
 $dialog \
 --title "Stealth options" \
---separate-output \
 --backtitle "Samhain $VERSION" \
 --yesno "SAMHAIN can compile and install a kernel module to hide the SAMHAIN daemon process. Do you want that ?"\
 2> $tmp
@@ -691,7 +685,7 @@ fi
 
 if [ "x$KCHECK" = "xyes" ]; then
     if [ `id -u` != 0 ]; then
-$dialog --msgbox "Compiling with --with-kcheck option (kernel rootkit detection). This\nrequires root privileges for at least one command during compilation,\nbut you are not running this as root. Please expect compilation to fail.\n\nYou need to run \"make\" manually after failure, and\nfollow the instructions to execute the one required\ncommand as root." 20 75
+$dialog --msgbox "Compiling with --with-kcheck option (kernel rootkit detection). This\nrequires root privileges for at least one command during compilation,\nbut you are not running this as root. Please expect compilation to fail.\n\nYou need to follow the instructions shown in the \nerror message after failure." 20 75
     fi
 fi
 
