@@ -33,20 +33,59 @@
  *
  **************************************************/
 
+/* IPv6 */
+#if defined(HAVE_GETNAMEINFO) && defined(HAVE_GETADDRINFO)
+
+#if defined(SH_COMPILE_STATIC) && defined(__linux__)
+#undef USE_IPVX
+#define SH_SOCKMAX 1
+#else
+
+#if defined(USE_IPV4)
+#undef USE_IPVX
+#else
+#define USE_IPVX 1
+#endif
+
+#define SH_SOCKMAX 8
+#endif
+
+#else
+#undef USE_IPVX
+#define SH_SOCKMAX 1
+#endif
+
+/* end IPv6 */
+
 #define REPLACE_OLD
 
-/* Standard buffer sizes.
+/* Standard buffer sizes. 
+ * IPv6 is 8 groups of 4 hex digits seperated by colons.
  */
+#define SH_IP_BUF        48
 #define SH_MINIBUF       64
 #define SH_BUFSIZE     1024
 #define SH_MAXBUF      4096
 #define SH_PATHBUF      256
+#define SH_MSG_BUF    64512
+
+#define SH_ERRBUF_SIZE   64
+
+/* MAX_PATH_STORE must be >= KEY_LEN
+ */
+#define MAX_PATH_STORE 12287
 
 /* Sizes for arrays (user, group, timestamp).
  */
-#define USER_MAX  20
-#define GROUP_MAX 20
-#define TIM_MAX   32
+#define SOCKPASS_MAX 14
+#define USER_MAX     20
+#define GROUP_MAX    20
+#define TIM_MAX      32
+
+#define CMODE_SIZE   11
+
+#define ATTRBUF_SIZE 16
+#define ATTRBUF_USED 12
 
 /* The number of bytes in a key,  
  * the number of chars in its hex repesentation,
@@ -55,15 +94,12 @@
 #define KEY_BYT   24
 #define KEY_LEN   48
 #define KEY_BLOCK 24
+#define KEYBUF_SIZE (KEY_LEN+1)
 
 /* The length of the compiled-in password.
  */
 #define PW_LEN     8
 
-/* Buffer for privileged content.
- */
-#define PRIV_MAX  4096   
-
 #undef  GOOD
 #define GOOD  1
 #undef  BAD
@@ -88,26 +124,32 @@
 #elif defined(HAVE_SHORT_32)
 #define UINT32 unsigned short
 #define SINT32 short
+#else
+#error "No 32 bit integer type found"
 #endif
 
 #ifdef HAVE_INTTYPES_H
 #include <inttypes.h>
-#else
+#endif
 #ifdef HAVE_STDINT_H
 #include <stdint.h>
 #endif
+
+#if !defined(HAVE_UINT16_T)
+#define UINT16 unsigned short
+#else
+#define UINT16 uint16_t
 #endif
 
-#if !defined(HAVE_INTTYPES_H) && !defined(HAVE_STDINT_H)
+#if !defined(HAVE_UINT64_T)
 
-#ifdef HAVE_LONG_LONG
-#define  UINT64 long long
+#ifdef HAVE_LONG_LONG_64
+#define  UINT64 unsigned long long
 #else
 #ifdef HAVE_LONG_64
-#define  UINT64 long
+#define  UINT64 unsigned long
 #else
-#define UINT64_IS_32
-#define  UINT64 long
+#error "no 64bit type found"
 #endif
 #endif
 
@@ -119,6 +161,13 @@
 
 #define UBYTE unsigned char
 
+enum {
+  SH_TIGER192    = 0, 
+  SH_SHA1        = 1,
+  SH_MD5         = 2,
+  SH_SHA256      = 3
+};
+
 
 enum {
   SH_CHECK_NONE    = 0, 
@@ -126,6 +175,48 @@ enum {
   SH_CHECK_CHECK   = 2
 };
 
+#define SH_MOD_THREAD  1
+#define SH_MOD_ACTIVE  0
+#define SH_MOD_FAILED -1
+#define SH_MOD_OFFSET 10
+
+/* Flags for file status
+ */
+#define SH_FFLAG_ALLIGNORE (1<<0)
+#define SH_FFLAG_VISITED   (1<<1)
+#define SH_FFLAG_CHECKED   (1<<3)
+#define SH_FFLAG_REPORTED  (1<<3)
+#define SH_FFLAG_SUIDCHK   (1<<4)
+
+#define SH_FFLAG_ALLIGNORE_SET(a)   (((a) & SH_FFLAG_ALLIGNORE) != 0)
+#define SET_SH_FFLAG_ALLIGNORE(a)   ((a) |= SH_FFLAG_ALLIGNORE)
+#define CLEAR_SH_FFLAG_ALLIGNORE(a) ((a) &= ~SH_FFLAG_ALLIGNORE)
+
+#define SH_FFLAG_VISITED_SET(a)     (((a) & SH_FFLAG_VISITED) != 0)
+#define SET_SH_FFLAG_VISITED(a)     ((a) |= SH_FFLAG_VISITED)
+#define CLEAR_SH_FFLAG_VISITED(a)   ((a) &= ~SH_FFLAG_VISITED)
+
+#define SH_FFLAG_CHECKED_SET(a)     (((a) & SH_FFLAG_VISITED) != 0)
+#define SET_SH_FFLAG_CHECKED(a)     ((a) |= SH_FFLAG_VISITED)
+#define CLEAR_SH_FFLAG_CHECKED(a)   ((a) &= ~SH_FFLAG_VISITED)
+
+#define SH_FFLAG_REPORTED_SET(a)    (((a) & SH_FFLAG_REPORTED) != 0)
+#define SET_SH_FFLAG_REPORTED(a)    ((a) |= SH_FFLAG_REPORTED)
+#define CLEAR_SH_FFLAG_REPORTED(a)  ((a) &= ~SH_FFLAG_REPORTED)
+
+#define SH_FFLAG_SUIDCHK_SET(a)     (((a) & SH_FFLAG_SUIDCHK) != 0)
+#define SET_SH_FFLAG_SUIDCHK(a)     ((a) |= SH_FFLAG_SUIDCHK)
+#define CLEAR_SH_FFLAG_SUIDCHK(a)   ((a) &= ~SH_FFLAG_SUIDCHK)
+
+/* Flags for inotify
+ */
+#define SH_INOTIFY_USE      (1<<0)
+#define SH_INOTIFY_DOSCAN   (1<<1)
+#define SH_INOTIFY_NEEDINIT (1<<2)
+#define SH_INOTIFY_INSCAN   (1<<3)
+#define SH_INOTIFY_IFUSED(a)   if ((sh.flag.inotify & SH_INOTIFY_USE) != 0) { a }
+
+
 /**************************************************
  *
  * TYPEDEFS
@@ -141,7 +232,10 @@ enum {
   SH_LEVEL_ATTRIBUTES  = 6,
   SH_LEVEL_USER0       = 7,
   SH_LEVEL_USER1       = 8,
-  SH_LEVEL_PRELINK     = 9
+  SH_LEVEL_USER2       = 9,
+  SH_LEVEL_USER3       = 10,
+  SH_LEVEL_USER4       = 11,
+  SH_LEVEL_PRELINK     = 12
 };
 
 typedef struct {
@@ -183,6 +277,9 @@ typedef struct {
   time_t          time_check;    /* time      last check */
   unsigned long   dirs_checked;  /* #dirs     last check */
   unsigned long   files_checked; /* #files    last check */
+  unsigned long   files_report;  /* #file reports        */
+  unsigned long   files_error;   /* #file access error   */
+  unsigned long   files_nodir;   /* #file not a directory*/
 } sh_sh_stat;
 
 typedef struct {
@@ -190,6 +287,7 @@ typedef struct {
   int    checkSum;                 /* whether to init/check checksums */
   int    update;                   /* update db                       */
   int    opts;                     /* reading cl options              */
+  int    started;                  /* finished with startup stuff     */
   int    isdaemon;                 /* daemon or not                   */
   int    loop;                     /* go in loop even if not daemon   */
   int    nice;                     /* desired nicety                  */
@@ -203,12 +301,15 @@ typedef struct {
   int    client_class;             /* TRUE if client class used       */
   int    audit;
   unsigned long aud_mask;
-  int    hidefile;                 /* TRUE if file not reveled in log */
+  int    hidefile;                 /* TRUE if file not shown in log   */
+  int    inotify;                  /* Flags for inotify               */
 } sh_sh_flag;
 
 typedef struct {
 
   char   prg_name[8];
+
+  UINT64 pid;  
  
   sh_sh_df     exec;
   sh_sh_df     conf;
@@ -239,6 +340,11 @@ typedef struct {
 
   int    looptime;                 /* timing for main loop            */
   /*@null@*//*@out@*/ char   * timezone;
+
+#ifdef SCREW_IT_UP
+  int sigtrap_max_duration;
+#endif
+
 } sh_struct;
 
 
@@ -246,8 +352,9 @@ extern volatile  int      sig_raised;
 extern volatile  int      sig_urgent;
 extern volatile  int      sig_debug_switch;       /* SIGUSR1 */
 extern volatile  int      sig_suspend_switch;     /* SIGUSR2 */
-extern volatile  int      sh_global_suspend_flag; /* SIGUSR2 */
+extern volatile  int      sh_global_suspend_flag;
 extern volatile  int      sig_fresh_trail;        /* SIGIOT  */
+extern volatile  int      sh_thread_pause_flag;
 extern volatile  int      sig_config_read_again;  /* SIGHUP  */
 extern volatile  int      sig_terminate;          /* SIGQUIT */
 extern volatile  int      sig_termfast;           /* SIGTERM */
@@ -255,11 +362,14 @@ extern volatile  int      sig_force_check;        /* SIGTTOU */
 
 extern long int eintr__result;
 
+extern int     sh_argc_store;
+extern char ** sh_argv_store;
+
 #include "sh_calls.h"
 
 
 typedef struct {
-  char   sh_sockpass[32];
+  char   sh_sockpass[2*SOCKPASS_MAX+2];
   char   sigkey_old[KEY_LEN+1];
   char   sigkey_new[KEY_LEN+1];
   char   mailkey_old[KEY_LEN+1];
@@ -279,6 +389,8 @@ typedef struct {
   UINT32 rng1[3];
   UINT32 rng2[3];
 
+  UINT32 res_vec[6];
+
   UINT32 ErrFlag[2];
 
 #ifdef SH_ENCRYPT
@@ -290,24 +402,99 @@ typedef struct {
 extern sh_struct sh; 
 /*@null@*/ extern sh_key_t  *skey; 
 
+/**************************************************
+ *
+ * macros
+ *
+ **************************************************/
+
+#if defined(__GNUC__) && (__GNUC__ >= 4)
+#define SH_GNUC_SENTINEL __attribute__((__sentinel__))
+#else
+#define SH_GNUC_SENTINEL
+#endif
+
+#if defined(__GNUC__) && (__GNUC__ >= 3)
+#undef  SH_GNUC_PURE
+#define SH_GNUC_PURE     __attribute__((pure))
+#undef  SH_GNUC_CONST
+#define SH_GNUC_CONST	 __attribute__((const))
+#undef  SH_GNUC_NORETURN
+#define SH_GNUC_NORETURN __attribute__((noreturn))
+#undef  SH_GNUC_MALLOC
+#define SH_GNUC_MALLOC   __attribute__((malloc))
+#else
+#undef  SH_GNUC_PURE
+#define SH_GNUC_PURE
+#undef  SH_GNUC_CONST
+#define SH_GNUC_CONST
+#undef  SH_GNUC_NORETURN
+#define SH_GNUC_NORETURN
+#undef  SH_GNUC_MALLOC
+#define SH_GNUC_MALLOC
+#endif
+
+
+/* The semantics of the built-in are that it is expected that expr == const
+ * for __builtin_expect ((expr), const)
+ */
+#if defined(__GNUC__) && (__GNUC__ > 2) && defined(__OPTIMIZE__)
+#define SH_LIKELY(expr)   (__builtin_expect((expr), 1))
+#define SH_UNLIKELY(expr) (__builtin_expect((expr), 0))
+#else
+#define SH_LIKELY(expr) (expr)
+#define SH_UNLIKELY(expr) (expr)
+#endif
+
+/* signal-safe log function
+ */
+int  safe_logger (int thesignal, int method, char * details);
+void safe_fatal  (const char * details, const char *f, int l);
+
+#define SH_VALIDATE_EQ(a,b) \
+     do { \
+         if ((a) != (b)) safe_fatal(#a " != " #b, FIL__, __LINE__);\
+     } while (0)
+
+#define SH_VALIDATE_NE(a,b) \
+     do { \
+         if ((a) == (b)) safe_fatal(#a " == " #b, FIL__, __LINE__);\
+     } while (0)
+
+#define SH_VALIDATE_GE(a,b) \
+     do { \
+         if ((a) < (b)) safe_fatal(#a " < " #b, FIL__, __LINE__);\
+     } while (0)
 
 #if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
+#ifdef USE_SUID
 #define MLOCK(a, b) \
       if ((skey != NULL) && skey->mlock_failed == SL_FALSE){ \
         (void) sl_set_suid(); \
-	if (sh_unix_mlock( a, b) < 0) skey->mlock_failed = SL_TRUE; \
-        (void) sl_unset_suid(); } 
+	if (sh_unix_mlock(FIL__, __LINE__, a, b) < 0) skey->mlock_failed = SL_TRUE; \
+        (void) sl_unset_suid(); }
+#else
+#define MLOCK(a, b) \
+      if ((skey != NULL) && skey->mlock_failed == SL_FALSE){ \
+        if (sh_unix_mlock(FIL__, __LINE__, a, b) < 0) skey->mlock_failed = SL_TRUE; }
+#endif 
 #else
 #define MLOCK(a, b) \
   ;
 #endif
 
 #if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
+#ifdef USE_SUID
 #define MUNLOCK(a, b) \
       if ((skey != NULL) && skey->mlock_failed == SL_FALSE){ \
         (void) sl_set_suid(); \
 	(void) sh_unix_munlock( a, b );\
-        (void) sl_unset_suid(); } 
+        (void) sl_unset_suid(); }
+#else
+#define MUNLOCK(a, b) \
+      if ((skey != NULL) && skey->mlock_failed == SL_FALSE){ \
+        (void) sh_unix_munlock( a, b ); }
+#endif 
 #else
 #define MUNLOCK(a, b) \
   ;
