@@ -13,11 +13,11 @@ void sh_socket_server_cmd(const char * srvcmd);
 #if defined (SH_WITH_SERVER)
 
 
-int    sh_socket_open_int ();
-int    sh_socket_remove ();
+int    sh_socket_open_int (void);
+int    sh_socket_remove (void);
 char * sh_socket_check(const char * client_name);
-int    sh_socket_poll();
-
+int    sh_socket_poll(void);
+void   sh_socket_add2reload (const char * clt);
 
 #endif
 
