@@ -22,7 +22,12 @@
 #ifndef SH_UNIX_H
 #define SH_UNIX_H
 
+/* For PATH_MAX */
 #include <limits.h>
+#if !defined(PATH_MAX)
+#define PATH_MAX 1024
+#endif
+
 #include <unistd.h>
 #include "samhain.h"
 #include "sh_error.h"
@@ -48,6 +53,8 @@ typedef enum {
   SH_FILE_BDEV,
   SH_FILE_FIFO,
   SH_FILE_SOCKET,
+  SH_FILE_DOOR,
+  SH_FILE_PORT,
   SH_FILE_UNKNOWN
 } ShFileType;
 
@@ -76,8 +83,25 @@ typedef enum {
 #define MODI_MOD (1 << 9)
 /* hardlinks    */
 #define MODI_HLN (1 << 10)
-/* device num   */
+/* device type   */
 #define MODI_RDEV (1 << 11)
+/* size may grow   */
+#define MODI_SGROW (1 << 12)
+/* use prelink     */
+#define MODI_PREL (1 << 13)
+
+/* get content     */
+#define MODI_TXT ((1 << 14)|MODI_CHK)
+#define MODI_TXT_ENABLED(a) (((a)&(1 << 14))!=0)
+
+/* get audit record  */
+#define MODI_AUDIT (1 << 15)
+#define MODI_AUDIT_ENABLED(a) (((a)&(1 << 15))!=0)
+
+#define MODI_INIT 0xDA000000UL
+#define MODI_INITIALIZED(a) (((a) & 0xFF000000UL) == MODI_INIT)
+
+#define SH_TXT_MAX 9200
 
 #define MASK_ALLIGNORE_  0
 extern  unsigned long mask_ALLIGNORE;
@@ -85,7 +109,7 @@ extern  unsigned long mask_ALLIGNORE;
 extern  unsigned long mask_ATTRIBUTES;
 #define MASK_LOGFILES_   (MASK_ATTRIBUTES_|MODI_HLN|MODI_LNK|MODI_INO)
 extern  unsigned long mask_LOGFILES;
-#define MASK_LOGGROW_    (MASK_LOGFILES_|MODI_SIZ) 
+#define MASK_LOGGROW_    (MASK_LOGFILES_|MODI_SIZ|MODI_SGROW|MODI_CHK) 
 extern  unsigned long mask_LOGGROW;
 #define MASK_READONLY_   (MASK_LOGFILES_|MODI_CHK|MODI_SIZ|MODI_MTM|MODI_CTM)
 extern  unsigned long mask_READONLY;
@@ -94,25 +118,28 @@ extern  unsigned long mask_NOIGNORE;
 #define MASK_USER_       (MASK_READONLY_|MODI_ATM)
 extern  unsigned long mask_USER0;
 extern  unsigned long mask_USER1;
-/* like READONLY, but without MTM,CTM,SIZ,INO)
+extern  unsigned long mask_USER2;
+extern  unsigned long mask_USER3;
+extern  unsigned long mask_USER4;
+/* like READONLY, but without MTM,CTM,SIZ,INO, and with PREL)
  */
-#define MASK_PRELINK_   (MASK_ATTRIBUTES_|MODI_HLN|MODI_LNK|MODI_CHK)
+#define MASK_PRELINK_   (MASK_ATTRIBUTES_|MODI_HLN|MODI_LNK|MODI_CHK|MODI_PREL)
 extern  unsigned long mask_PRELINK;
 
 typedef struct file_struct {
   unsigned long    check_mask;
-  int              reported;
+  int              file_reported;
   char             fullpath[PATH_MAX];
   ShFileType       type;
   dev_t            dev;
   ino_t            ino;
   mode_t           mode;
   nlink_t          hardlinks;
-#if defined(__linux__)
+#if defined(__linux__) || defined(HAVE_STAT_FLAGS)
   unsigned long    attributes;
-  char             c_attributes[16];
+  char             c_attributes[ATTRBUF_SIZE];
 #endif
-  char             c_mode[11];
+  char             c_mode[CMODE_SIZE];
   uid_t            owner;
   char             c_owner[USER_MAX+2];
   gid_t            group;
@@ -125,30 +152,58 @@ typedef struct file_struct {
   time_t           mtime;
   time_t           ctime;
 
-  char             linkpath[PATH_MAX];
+  char           * link_path;
   mode_t           linkmode;
   char             link_c_mode[11];
   int              linkisok;
+  char           * attr_string;
 } file_type;
 
+extern int sh_unix_check_selinux;
+extern int sh_unix_check_acl;
+
+/* destroy userid cache 
+ */
+void sh_userid_destroy ();
+
+/* --- run a command, securely --- 
+ */
+int sh_unix_run_command (const char * str);
+
+/* Ignore SIGPIPE
+ */
+void sh_unix_ign_sigpipe();
+
 /* mlock utilities
  */
-int sh_unix_mlock(void * addr, size_t len);
+int sh_unix_mlock(const char * file, int line, void * addr, size_t len);
 int sh_unix_munlock(void * addr, size_t len);
-int sh_unix_count_mlock();
+int sh_unix_count_mlock(void);
+/* public for unit tests */
+int sh_unix_pagesize(void);
+unsigned long sh_unix_lookup_page(void * in_addr, size_t len, int * num_pages);
 
 /* chroot directory
  */
-int sh_unix_set_chroot(char * str);
+int sh_unix_set_chroot(const char * str);
 
 /* whether to use localtime for file timesatams in logs
  */
-int sh_unix_uselocaltime (char * c);
+int sh_unix_uselocaltime (const char * c);
+
+/* whether to perform selinux/acl checks
+ */ 
+#ifdef USE_XATTR
+int sh_unix_setcheckselinux (const char * c);
+#endif
+#ifdef USE_ACL
+int sh_unix_setcheckacl (const char * c);
+#endif
 
 /* set I/O limit
  */
-int  sh_unix_set_io_limit (char * c);
-void sh_unix_io_pause ();
+int  sh_unix_set_io_limit (const char * c);
+void sh_unix_io_pause (void);
 
 /* get file type
  */
@@ -156,7 +211,11 @@ int sh_unix_get_ftype(char * fullpath);
 
 /* reset masks for policies
  */
-int sh_unix_maskreset();
+int sh_unix_maskreset(void);
+
+/* return true if database is remote
+ */
+int file_is_remote (void);
 
 /* return the path to the configuration/database file
  */
@@ -168,8 +227,11 @@ unsigned long sh_unix_longtime (void);
 
 /* close all files >= fd, except possibly one
  */
-void sh_unix_closeall (int fd, int except);
+void sh_unix_closeall (int fd, int except, int inchild);
 
+/* Check whether directory for pid file exists
+ */
+int sh_unix_check_piddir (char * pidpath);
 
 /* write lock for filename
  */
@@ -181,16 +243,16 @@ int sh_unix_rm_lock_file(char * filename);
 
 /* write the PID file
  */
-int sh_unix_write_pid_file();
+int sh_unix_write_pid_file(void);
 
 /* rm the PID file
  */
-int sh_unix_rm_pid_file();
+int sh_unix_rm_pid_file(void);
 
 
 /* checksum of own binary
  */
-int sh_unix_self_hash (char * c);
+int sh_unix_self_hash (const char * c);
 
 /* return BAD on failure
  */
@@ -198,11 +260,11 @@ int sh_unix_self_check (void);
 
 /* add a trusted user to the list 
  */
-int tf_add_trusted_user(char *);
+int tf_add_trusted_user(const char *);
 
 /* check a file 
  */
-int tf_trust_check (char * file, int mode);
+int tf_trust_check (const char * file, int mode);
 
 /* initialize group vector
  */
@@ -215,7 +277,7 @@ int  sh_unix_initgroups2 (uid_t         in_pid, gid_t in_gid);
 
 /* set the timeserver address
  */
-int sh_unix_settimeserver (char * address);
+int sh_unix_settimeserver (const char * address);
 void reset_count_dev_time(void);
 
 /* lock the key
@@ -224,16 +286,16 @@ void sh_unix_memlock(void);
 
 /* deamon mode 
  */
-int sh_unix_setdeamon  (char * dummy);
-int sh_unix_setnodeamon(char * dummy);
+int sh_unix_setdeamon  (const char * dummy);
+int sh_unix_setnodeamon(const char * dummy);
 
 /* Test whether file exists
  */
-int sh_unix_file_stat(char * path);
+int sh_unix_file_exists(char * path);
 
 /* test whether file exists with proper attributes
  */
-int sh_unix_file_exists(int fd);
+int sh_unix_device_readable(int fd);
 
 /* local host
  */
@@ -248,9 +310,14 @@ int sh_unix_test_proc(void);
  */
 /* int sh_unix_is_secure_dir (ShErrLevel level, char * tmp); */
 
+/* check whether there's a rotated log with the correct inode and checksum 
+ */
+int sh_check_rotated_log (const char * path,  
+			  UINT64 old_size, UINT64 old_inode, const char * old_hash);
+
 /* obtain file info
  */
-int sh_unix_getinfo (int level, char * filename, file_type * theFile, 
+int sh_unix_getinfo (int level, const char * filename, file_type * theFile, 
 		     char * fileHash, int flagrel);
 
 /* read file, return length read
@@ -261,22 +328,25 @@ int sh_unix_getline (SL_TICKET fd, char * line, int sizeofline);
  */
 int  sh_unix_init(int goDaemon);
 
-/* for local time use thetime = 0 
+/* for local time use thetime = 0, returns pointer to buffer 
  */
-/*@owned@*/ char * sh_unix_time (time_t thetime);
+char * sh_unix_time (time_t thetime, char * buffer, size_t len);
 
-/* convert to GMT time
+/* convert to GMT time, returns pointer to buffer
  */
-char * sh_unix_gmttime (time_t thetime);
+char * sh_unix_gmttime (time_t thetime, char * buffer, size_t len);
 
 /* effective user info
  */
 int  sh_unix_getUser (void);
 
-/* get home directory
+/* get home directory, , returns pointer to out
  */
-char *  sh_unix_getUIDdir (int level, uid_t uid);
+char *  sh_unix_getUIDdir (int level, uid_t uid, char * out, size_t len);
 
+/* get a group GID 
+ */
+long sh_group_to_gid (const char * g, int * fail);
 
 #ifdef HAVE_GETTIMEOFDAY
 unsigned long sh_unix_notime (void);
@@ -298,7 +368,6 @@ void sh_unix_xor_code (char * str, int len);
 #include <errno.h>
 
 void   sh_sigtrap_handler (int signum);
-extern volatile int sh_not_traced;
 
 #ifdef HAVE_GETTIMEOFDAY
 #if TIME_WITH_SYS_TIME
@@ -311,9 +380,19 @@ extern volatile int sh_not_traced;
 #include <time.h>
 #endif
 #endif
-extern struct timeval  save_tv;
 #endif
 
+struct sh_sigtrap_variables {
+  int not_traced;
+#ifdef HAVE_GETTIMEOFDAY
+  struct timeval save_tv;
+#endif
+};
+
+struct sh_sigtrap_variables * sh_sigtrap_variables_get();
+
+int sh_sigtrap_max_duration_set (const char * str);
+
 static inline
 int  sh_sigtrap_prepare()
 {
@@ -331,21 +410,28 @@ int  sh_sigtrap_prepare()
 /*@unused@*/ static inline 
 int sh_derr(void)
 {
-  sh_not_traced = 0;
+  struct sh_sigtrap_variables * sigtrap_variables;
+  sigtrap_variables = sh_sigtrap_variables_get();
+  if (sigtrap_variables == NULL) {
+    /* Perhaps, its better to not die, and to continue using Samhain,
+       even if this part does not work. */
+    return (0);
+  }
+
+  sigtrap_variables->not_traced = 0;
 
 #ifdef HAVE_GETTIMEOFDAY
-  gettimeofday(&save_tv, NULL);
+  gettimeofday(&sigtrap_variables->save_tv, NULL);
 #endif
 
-#if defined(__linux__) && defined(__GNUC__) && defined(__i386__)
-  __asm__ __volatile__ ("int $0x03");
-#else
+  /* raise() sends to same thread, like pthread_kill(pthread_self(), sig); 
+   */
   raise(SIGTRAP);
-#endif
   
-  if (sh_not_traced == 0)
+  if (sigtrap_variables->not_traced == 0)
     _exit(5);
-  sh_not_traced = 0;
+
+  sigtrap_variables->not_traced = 0;
   return (0);
 }
 
