@@ -62,15 +62,17 @@
 #include <float.h>
 
 #include "samhain.h"
-#include "sh_files.h"
+#include "sh_pthread.h"
 #include "sh_utils.h"
 #include "sh_error.h"
 #include "sh_unix.h"
+#include "sh_files.h"
 #include "sh_getopt.h"
 #include "sh_readconf.h"
 #include "sh_hash.h"
+#include "sh_restrict.h"
 
-#include "sh_mail.h"
+#include "sh_nmail.h"
 
 #include "sh_tiger.h"
 #include "sh_gpg.h"
@@ -112,6 +114,7 @@ volatile  int      sig_debug_switch;       /* SIGUSR1 */
 volatile  int      sig_suspend_switch;     /* SIGUSR2 */
 volatile  int      sh_global_suspend_flag;
 volatile  int      sig_fresh_trail;        /* SIGIOT  */
+volatile  int      sh_thread_pause_flag = S_FALSE;
 volatile  int      sig_config_read_again;  /* SIGHUP  */
 volatile  int      sig_terminate;          /* SIGQUIT */
 volatile  int      sig_termfast;           /* SIGTERM */
@@ -136,21 +139,84 @@ char               sh_sig_msg[SH_MINIBUF];
 #define GLOB_LEN 511
 #endif
 
+#ifdef HAVE_PTHREAD
+struct gt {
+  size_t g_count;
+  char * g_glob;
+};
+
+pthread_key_t g_key;
+
+int sh_g_thread()
+{
+  struct gt * ptr = malloc(sizeof(struct gt));
+  if (!ptr)
+    return -1;
+  ptr->g_count    = 0;
+  ptr->g_glob     = calloc(1, SH_MAX_GLOBS * (GLOB_LEN+1));
+  if (!(ptr->g_glob))
+    return -1;
+  return pthread_setspecific(g_key, ptr);
+}
+
+void sh_g_destroy(void * data)
+{
+  struct gt * ptr = (struct gt *) data;
+  free(ptr->g_glob);
+  free(ptr);
+  return;
+}
+
+void sh_g_init(void)
+{
+#if !defined(USE_SYSTEM_MALLOC) && defined(USE_MALLOC_LOCK)
+  extern int dnmalloc_pthread_init(void);
+  dnmalloc_pthread_init();
+#endif
+
+  if (0 != pthread_key_create(&g_key, sh_g_destroy))
+    {
+      perror("1");
+      exit(EXIT_FAILURE);
+    }
+
+  if (0 != sh_g_thread())
+    {
+      perror("2");
+      exit(EXIT_FAILURE);
+    }
+  return;
+}
+#define SH_G_INIT sh_g_init()
+#else
+#define SH_G_INIT ((void)0)
+#endif
+
 char * globber(const char * str)
 {
   size_t i;
   size_t j;
 
-  static   size_t  items = 0;
+#ifndef HAVE_PTHREAD
   static   size_t  count = 0;
   static   char glob[SH_MAX_GLOBS * (GLOB_LEN+1)];
+#else
+  struct gt * ptr = pthread_getspecific(g_key);
+  size_t count;
+  char *  glob;
 
-  if (str == NULL)
+  if (ptr) {
+    count = ptr->g_count;
+    glob  = ptr->g_glob;
+  } else {
     return NULL;
-  else
-    j = strlen(str);
+  }
+#endif
 
-  ++items;
+  if (str != NULL)
+    j = strlen(str);
+  else
+    return NULL;
 
   ASSERT((j <= GLOB_LEN), _("j <= GLOB_LEN"))
 
@@ -162,7 +228,6 @@ char * globber(const char * str)
   if ( (count + j) >= (SH_MAX_GLOBS * (GLOB_LEN+1)))
     {
       count = 0;
-      items = 0;
     }
 
   for (i = 0; i < j; ++i)
@@ -175,7 +240,11 @@ char * globber(const char * str)
   glob[count + j] = '\0';
 
   i     = count;
+#ifdef HAVE_PTHREAD
+  ptr->g_count = count + j + 1;
+#else
   count = count + j + 1;
+#endif
   return &glob[i];
 }
 
@@ -192,6 +261,9 @@ void sh_do_encode (char * str, int len)
   return;
 }
 
+#else
+/* not stealth */
+#define SH_G_INIT ((void)0)
 #endif
 
 /**************************************************
@@ -296,12 +368,15 @@ void sh_init (void)
   sh.prg_name[7] = '\0';
 #endif
 
+  sh.pid = (UINT64) getpid();
+
   /* The flags.
    */
   if (is_samhainctl_init == S_FALSE)
     sh.flag.checkSum        = SH_CHECK_NONE;
   sh.flag.update          = S_FALSE;
   sh.flag.opts            = S_FALSE;
+  sh.flag.started         = S_FALSE;
   if (is_samhainctl_init == S_FALSE)
     sh.flag.isdaemon        = S_FALSE;
   sh.flag.isserver        = S_FALSE;
@@ -317,6 +392,7 @@ void sh_init (void)
   sh.flag.client_class    = S_FALSE;
   sh.flag.hidefile        = S_FALSE;
   sh.flag.loop            = S_FALSE;
+  sh.flag.inotify         = 0;
 
 #ifdef MKB_09
   ErrFlag[1] |= (1 << 8);
@@ -345,8 +421,12 @@ void sh_init (void)
 
   /* The stats.
    */
-  sh.statistics.bytes_speed  = 0;
-  sh.statistics.bytes_hashed = 0;
+  sh.statistics.bytes_speed   = 0;
+  sh.statistics.bytes_hashed  = 0;
+  sh.statistics.files_report  = 0;
+  sh.statistics.files_error   = 0;
+  sh.statistics.files_nodir   = 0;
+
   sh.statistics.mail_success = 0;
   sh.statistics.mail_failed  = 0;
   sh.statistics.time_start   = time(NULL);
@@ -447,16 +527,28 @@ void sh_init (void)
   /* The addresses.
    */
 #if defined(SH_WITH_MAIL)
-  if (0 == strcmp (DEFAULT_MAILADDRESS, _("NULL")))
+  if (0 != strcmp (DEFAULT_MAILADDRESS, _("NULL")))
     {
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+      char * saveptr;
+      (void) sl_strncpy(q, DEFAULT_MAILADDRESS, SH_PATHBUF);
+      p = strtok_r (q, ", \t", &saveptr);
+      if (p)
+	{
+	  (void) sh_nmail_add_compiled_recipient (p);
+	  while (NULL != (p = strtok_r (NULL, ", \t", &saveptr)))
+	    (void) sh_nmail_add_compiled_recipient (p);
+	}
+#else
       (void) sl_strncpy(q, DEFAULT_MAILADDRESS, SH_PATHBUF);
       p = strtok (q, ", \t");
       if (p)
 	{
-	  (void) sh_mail_setaddress_int (p);
+	  (void) sh_nmail_add_compiled_recipient (p);
 	  while (NULL != (p = strtok (NULL, ", \t")))
-	    (void) sh_mail_setaddress_int (p);
+	    (void) sh_nmail_add_compiled_recipient (p);
 	}
+#endif
     }
 #endif
 
@@ -541,51 +633,57 @@ void sh_init (void)
 
   sh.looptime     = 60;
 
+#ifdef SCREW_IT_UP
+  sh.sigtrap_max_duration = 500000; /* 500ms */
+#endif
+
   /* The struct to hold privileged information.
    */
   skey = (sh_key_t *) malloc (sizeof(sh_key_t));
-  if (skey == NULL) 
+  if (skey != NULL)
+    {
+
+      skey->mlock_failed = SL_FALSE;
+      skey->rngI         = BAD;
+      /* properly initialized later 
+       */
+      skey->rng0[0] = 0x03; skey->rng0[1] = 0x09; skey->rng0[2] = 0x17;
+      skey->rng1[0] = 0x03; skey->rng1[1] = 0x09; skey->rng1[2] = 0x17;
+      skey->rng2[0] = 0x03; skey->rng2[1] = 0x09; skey->rng2[2] = 0x17;
+      
+      for (i = 0; i < KEY_BYT; ++i)
+	skey->poolv[i] = '\0';
+      
+      skey->poolc        = 0;
+      
+      skey->ErrFlag[0]   = ErrFlag[0];
+      ErrFlag[0]         = 0;
+      skey->ErrFlag[1]   = ErrFlag[1];
+      ErrFlag[1]         = 0;
+      
+      dez = &(TcpFlag[POS_TF-1][0]);
+      for (i = 0; i < PW_LEN; ++i)
+	{ 
+	  skey->pw[i] = (char) (*dez); 
+	  (*dez)      = '\0';
+	  ++dez; 
+	}
+      
+      skey->sh_sockpass[0]  = '\0';
+      skey->sigkey_old[0]   = '\0';
+      skey->sigkey_new[0]   = '\0';
+      skey->mailkey_old[0]  = '\0';
+      skey->mailkey_new[0]  = '\0';
+      skey->crypt[0]        = '\0'; /* flawfinder: ignore *//* ff bug */
+      skey->session[0]      = '\0';
+      skey->vernam[0]       = '\0';
+    }
+  else
     {
       perror(_("sh_init"));
       _exit (EXIT_FAILURE);
     }
 
-  skey->mlock_failed = SL_FALSE;
-  skey->rngI         = BAD;
-  /* properly initialized later 
-   */
-  skey->rng0[0] = 0x03; skey->rng0[1] = 0x09; skey->rng0[2] = 0x17;
-  skey->rng1[0] = 0x03; skey->rng1[1] = 0x09; skey->rng1[2] = 0x17;
-  skey->rng2[0] = 0x03; skey->rng2[1] = 0x09; skey->rng2[2] = 0x17;
-
-  for (i = 0; i < KEY_BYT; ++i)
-    skey->poolv[i] = '\0';
-
-  skey->poolc        = 0;
-
-  skey->ErrFlag[0]   = ErrFlag[0];
-  ErrFlag[0]         = 0;
-  skey->ErrFlag[1]   = ErrFlag[1];
-  ErrFlag[1]         = 0;
-
-  dez = &(TcpFlag[POS_TF-1][0]);
-  for (i = 0; i < PW_LEN; ++i)
-    { 
-       skey->pw[i] = (char) (*dez); 
-      (*dez)      = '\0';
-      ++dez; 
-    }
-
-  skey->sh_sockpass[0]  = '\0';
-  skey->sigkey_old[0]   = '\0';
-  skey->sigkey_new[0]   = '\0';
-  skey->mailkey_old[0]  = '\0';
-  skey->mailkey_new[0]  = '\0';
-  skey->crypt[0]        = '\0'; /* flawfinder: ignore *//* ff bug */
-  skey->session[0]      = '\0';
-  skey->vernam[0]       = '\0';
-
-
   sh_unix_memlock();
   SL_RET0(_("sh_init"));
 }
@@ -612,7 +710,8 @@ static void exit_handler(void)
   int modnum;
 #endif
 #if defined(SH_WITH_SERVER)
-  extern int sh_socket_remove ();
+  extern int sh_socket_remove (void);
+  extern int sh_html_zero();
 #endif
 
   SL_ENTER(_("exit_handler"));
@@ -624,9 +723,12 @@ static void exit_handler(void)
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
   for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
     {
-      if (modList[modnum].initval == GOOD)
+      if (modList[modnum].initval == SH_MOD_ACTIVE)
 	(void) modList[modnum].mod_cleanup();
     }
+#ifdef HAVE_PTHREAD
+  sh_pthread_cancel_all();
+#endif
 #endif
 
   /* --- Push out all pending messages. ---
@@ -634,14 +736,18 @@ static void exit_handler(void)
 #if defined(SH_WITH_MAIL)
   if (sh.mailNum.alarm_last > 0) 
     {
-      (void) sh_mail_msg (NULL);
+      (void) sh_nmail_flush ();
     }
 #endif
 
   /* --- Write the server stat. ---
    */
 #if defined(SH_WITH_SERVER)
-  sh_forward_html_write();
+  /* zero out the status file at exit, such that the status
+   * of client becomes unknown in the beltane interface
+   */
+  sh_html_zero();
+  /* sh_forward_html_write(); */
 #endif
 
   /* --- Clean up memory to check for problems. ---
@@ -650,12 +756,28 @@ static void exit_handler(void)
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
   sh_files_deldirstack ();
   sh_files_delfilestack ();
+  sh_files_delglobstack ();
   sh_hash_hashdelete();
+  sh_files_hle_reg (NULL);
+  /*
+   * Only flush on exit if running as deamon.
+   * Otherwise we couldn't run another instance
+   * while the deamon is running (would leave the
+   * deamon with flushed ruleset).
+   */
+  if (sh.flag.isdaemon == S_TRUE)
+    {
+      sh_audit_delete_all ();
+    }
 #endif
 #if defined(SH_WITH_SERVER)
   sh_forward_free_all ();
+#endif
+#if defined(SH_WITH_MAIL)
+  sh_nmail_free();
 #endif
   delete_cache();
+  sh_userid_destroy ();
   sh_mem_stat();
 #endif
 
@@ -688,7 +810,8 @@ static void exit_handler(void)
   /* --- Remove lock, delete critical information. ---
    */
   (void) sh_unix_rm_lock_file (sh.srvlog.name);
-  (void) sh_unix_rm_pid_file ();
+  if (sh.flag.isdaemon == S_TRUE)
+    (void) sh_unix_rm_pid_file ();
   if (skey != NULL)
     memset (skey, (int) '\0', sizeof(sh_key_t));
   
@@ -712,37 +835,39 @@ static void exit_handler(void)
 
 #if defined(__linux__) || defined(sun) || defined(__sun) || defined(__sun__)
 #include <dirent.h>
-static pid_t * procdirSamhain ()
+static pid_t * procdirSamhain (void)
 {
-  pid_t        * pidlist = malloc(sizeof(pid_t) * 65535);
+  pid_t        * pidlist;
   struct dirent * d;
   DIR *        dp;
   long         ino;
   struct stat  buf;
-  int          i = 0;
+  int          i;
   pid_t        pid, mypid = getpid();
   char       * tail;
   char         exef[128];
 
-  if (!pidlist)
-    return NULL;
-
-  for (i = 0; i < 65535; ++i) pidlist[i] = 0;
-  i = 0;
-
   if (0 != stat(SH_INSTALL_PATH, &buf))
     {
-      free(pidlist);
       return NULL;
     }
 
   ino = (long) buf.st_ino;
     
-  if (NULL == (dp = opendir("/proc")))
+  if (NULL == (dp = opendir(_("/proc"))))
     {
-      free(pidlist);
       return NULL;
     }
+
+  SH_MUTEX_LOCK(mutex_readdir);
+
+  pidlist =  malloc(sizeof(pid_t) * 65535);
+  if (!pidlist)
+    goto unlock_and_out;
+
+  for (i = 0; i < 65535; ++i) pidlist[i] = 0;
+
+  i = 0;
   while (NULL != (d = readdir(dp)) && i < 65535)
     {
       if (0 != strcmp(d->d_name, ".") && 0 != strcmp(d->d_name, ".."))
@@ -763,11 +888,16 @@ static pid_t * procdirSamhain ()
 	    { pidlist[i] = (pid_t) pid; ++i; }
 	}
     }
+
+ unlock_and_out:
+  ;
+  SH_MUTEX_UNLOCK(mutex_readdir);
+
   closedir(dp);
   return pidlist;
 }
 #else
-static pid_t * procdirSamhain ()
+static pid_t * procdirSamhain (void)
 {
   return NULL;
 }
@@ -817,8 +947,8 @@ static pid_t pidofSamhain (int flag)
   if (!fp)
     { if (errno != ENOENT) perror(_("fopen")); return 0; }
   if (NULL == fgets(line, sizeof(line), fp))
-    { perror(_("fgets")); (void) fclose(fp); return 0; }
-  (void) fclose(fp); 
+    { perror(_("fgets")); (void) sl_fclose(FIL__, __LINE__, fp); return 0; }
+  (void) sl_fclose(FIL__, __LINE__, fp); 
   p = line; 
   while (*p == ' '  || *p == '\f' || *p == '\n' || 
 	 *p == '\r' || *p == '\t' || *p == '\v')
@@ -866,7 +996,11 @@ static int samhainctl(int ctl, int * argc, char * argv[])
   char * argp[32];
   pid_t       * pidlist;
   int         i;
-
+#ifdef WCONTINUED
+      int wflags = WNOHANG|WUNTRACED|WCONTINUED;
+#else
+      int wflags = WNOHANG|WUNTRACED;
+#endif
 
   fullpath = strdup (SH_INSTALL_PATH);
   if (fullpath == NULL)
@@ -900,7 +1034,7 @@ static int samhainctl(int ctl, int * argc, char * argv[])
 	perror(_("fork"));
 	exit (1);
       case  0:
-	if (0 != close (0))
+	if (0 != sl_close_fd (FIL__, __LINE__, 0))
 	  {
 	    _exit(4);
 	  }
@@ -912,8 +1046,8 @@ static int samhainctl(int ctl, int * argc, char * argv[])
 	_exit (1);
       default:
 	times = 0;
-	while (times < 64) {
-	  respid = waitpid(pid, &status, WNOHANG|WUNTRACED);
+	while (times < 300) {
+	  respid = waitpid(pid, &status, wflags);
 	  if ((pid_t)-1 == respid)
 	    {
 	      perror(_("waitpid"));
@@ -921,6 +1055,7 @@ static int samhainctl(int ctl, int * argc, char * argv[])
 	    }
 	  else if (pid == respid)
 	    {
+#ifndef USE_UNO
 	      if (0 != WIFEXITED(status))
 		{
 		  res = WEXITSTATUS(status);
@@ -928,11 +1063,14 @@ static int samhainctl(int ctl, int * argc, char * argv[])
 		}
 	      else
 		exit (1);
+#else
+	      exit (1);
+#endif
 	    }
 	  ++times;
 	  (void) retry_msleep(1, 0);
 	}
-	exit (1);
+	exit (0); /* assume that it runs ok */
       }
     }
 
@@ -1040,7 +1178,7 @@ static sh_schedule_t * sh_set_schedule_int (const char * str,
       (void) free_sched(FileSchedIn);
       FileSchedIn = NULL;
       *status = 0;
-      return 0;
+      return NULL;
     }
 
   FileSched = SH_ALLOC(sizeof(sh_schedule_t));
@@ -1090,7 +1228,7 @@ int undef_main(int argc, char * argv[])
   extern int    create_syslog_socket (int flag);
 #endif
 #if defined(SH_WITH_SERVER)
-  extern int    sh_create_tcp_socket();
+  extern int    sh_create_tcp_socket(void);
 #endif
 
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
@@ -1098,16 +1236,16 @@ int undef_main(int argc, char * argv[])
   time_t        runtim;
   float         st_1, st_2;
   int           status;
-  long          cct = 0; /* main loop iterations */
+  volatile long          cct = 0; /* main loop iterations */
 
-  int           flag_check_1 = 0;
-  int           flag_check_2 = 0;
+  volatile int           flag_check_1 = 0;
+  volatile int           flag_check_2 = 0;
 
   int           check_done   = 0;
 #endif
 
-  time_t        told;
-  time_t        tcurrent;
+  volatile time_t        told;
+  volatile time_t        tcurrent;
   size_t        tzlen;
   char *        tzptr;
   int           res;
@@ -1118,8 +1256,21 @@ int undef_main(int argc, char * argv[])
   char  * my_argv[32];
 #endif
 
+#if !defined(USE_SYSTEM_MALLOC)
+  typedef void assert_handler_tp(const char * error, const char *file, int line);
+  extern assert_handler_tp *dnmalloc_set_handler(assert_handler_tp *new);
+  (void) dnmalloc_set_handler(safe_fatal);
+#endif
+
+  SH_G_INIT; /* Must precede any use of _() */
+
   SL_ENTER(_("main"));
 
+  /* --- Close all but first three file descriptors. ---
+   */
+  sh_unix_closeall(3, -1, SL_FALSE); /* at program start */
+
+
   if (argc >= 2 && 0 != getuid() &&
       (0 == strcmp(argv[1], _("start")) ||
        0 == strcmp(argv[1], _("stop")) ||
@@ -1237,14 +1388,12 @@ int undef_main(int argc, char * argv[])
 
   /* --------  INIT  --------    
    */
+  sh_unix_ign_sigpipe();
 
   /* Restrict error logging to stderr.
    */
   sh_error_only_stderr (S_TRUE);
 
-  BREAKEXIT(sh_derr);
-  (void) sh_derr();
-
   /* Check that first three descriptors are open.
    */
   if ( retry_fcntl(FIL__, __LINE__, 0, F_GETFL, 0) == (-1))
@@ -1262,6 +1411,12 @@ int undef_main(int argc, char * argv[])
   sh.flag.isserver = S_TRUE;
 #endif
 
+  /* --- First check for an attached debugger (after setting
+         sh.sigtrap_max_duration which has to be done before). ---
+   */
+  BREAKEXIT(sh_derr);
+  (void) sh_derr();
+
   /* --- Get local hostname. ---
    */
   BREAKEXIT(sh_unix_localhost);
@@ -1281,13 +1436,22 @@ int undef_main(int argc, char * argv[])
     {
       if ( 0 == strcmp(argv[1], NOCL_CODE) )
 	{
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+	  char * saveptr;
+#endif
 	  my_argv[0] = argv[0]; ++my_argc;  
 	  command_line[0] = '\0';
-	  (void*) fgets (command_line, sizeof(command_line), stdin);
-	  command_line[sizeof(command_line)-1] = '\0';
+	  if (NULL != fgets (command_line, sizeof(command_line), stdin))
+	    command_line[sizeof(command_line)-1] = '\0';
+
 	  do {
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
 	    my_argv[my_argc] = 
-	      strtok( (my_argc == 1) ? command_line : NULL, " \n"); 
+	      strtok_r( (my_argc == 1) ? command_line : NULL, " \n", &saveptr);
+#else
+	    my_argv[my_argc] = 
+	      strtok( (my_argc == 1) ? command_line : NULL, " \n");
+#endif 
 	    if (my_argv[my_argc] != NULL) {
 	      ++my_argc;
 	    } else {
@@ -1310,11 +1474,6 @@ int undef_main(int argc, char * argv[])
   sh.flag.opts = S_FALSE;
   
 
-  /* close all other files
-   */
-  sh_unix_closeall(3, -1); /* after processing CL options */
-
-
   /* --- Get user info. ---
    */
   TPT((0, FIL__, __LINE__, _("msg=<Get user name.>\n")))
@@ -1336,6 +1495,8 @@ int undef_main(int argc, char * argv[])
   BREAKEXIT(sh_readconf_read);
   (void) sh_readconf_read ();
 
+  sh_calls_enable_sub();
+
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
   if (sh.flag.checkSum == SH_CHECK_NONE)
     {
@@ -1368,11 +1529,30 @@ int undef_main(int argc, char * argv[])
       sh.flag.loop     = S_FALSE;
     }
 
-  /* initialize signal handling etc
+  /* --- load database; checksum of database
    */
-  if (sh.flag.isdaemon == S_TRUE)
-    sh_error_only_stderr (BAD);
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+  TPT((0, FIL__, __LINE__, _("msg=<Get checksum of the database.>\n")))
+  if (sh.flag.checkSum == SH_CHECK_CHECK) 
+    {
+      if (0 != sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
+	{
+	  char hashbuf[KEYBUF_SIZE];
+	  (void) sl_strlcpy(sh.data.hash,
+			    sh_tiger_hash (file_path('D', 'R'), 
+					   TIGER_FILE, TIGER_NOLIM, 
+					   hashbuf, sizeof(hashbuf)), 
+			    KEY_LEN+1);
+	}
 
+      /* this eventually fetches the file from server to get checksum
+       */
+      sh_hash_init ();
+    }
+#endif
+
+  /* --- initialize signal handling etc.; fork daemon
+   */
   if (sh_unix_init(sh.flag.isdaemon) == -1) 
     {
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
@@ -1411,34 +1591,12 @@ int undef_main(int argc, char * argv[])
 #endif
 
 
-  /* checksum of database
-   */
-#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
-  TPT((0, FIL__, __LINE__, _("msg=<Get checksum of the database.>\n")))
-  if (sh.flag.checkSum == SH_CHECK_CHECK) 
-    {
-      if (0 == sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
-	{
-	  /* fetch the file from server to get checksum
-	   */
-	  sh_hash_init ();
-	  /* sh_hash_hashdelete (); */
-	}
-      else
-	{
-	  (void) sl_strlcpy(sh.data.hash,
-			    sh_tiger_hash (file_path('D', 'R'), 
-					   TIGER_FILE, 0), 
-			    KEY_LEN+1);
-	}
-    }
-#endif
-
-
   /* --- Enable full error logging --- 
    */
   sh_error_only_stderr (S_FALSE);
 
+  sh.flag.started = S_TRUE;
+
   /****************************************************
    *
    *   SERVER 
@@ -1448,7 +1606,8 @@ int undef_main(int argc, char * argv[])
 #if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
 
 #if (defined(WITH_GPG) || defined(WITH_PGP))
-  /* do nothing -- we exit earlier if error */
+  /* log startup */
+  sh_gpg_log_startup ();
 #else
   sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_1H,
 		   sh.prg_name, (long) sh.real.uid, 
@@ -1470,7 +1629,8 @@ int undef_main(int argc, char * argv[])
   if (sh.flag.checkSum == SH_CHECK_CHECK) 
     {
 #if (defined(WITH_GPG) || defined(WITH_PGP))
-      /* do nothing -- we exit earlier if error */
+      /* log startup */
+      sh_gpg_log_startup ();
 #else
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_2H,
 		       sh.prg_name, (long) sh.real.uid,
@@ -1481,7 +1641,8 @@ int undef_main(int argc, char * argv[])
   else
     {
 #if (defined(WITH_GPG) || defined(WITH_PGP))
-      /* do nothing -- we exit earlier if error */
+      /* log startup */
+      sh_gpg_log_startup ();
 #else
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_1H,
 		       sh.prg_name, (long) sh.real.uid,
@@ -1538,25 +1699,26 @@ int undef_main(int argc, char * argv[])
   TPT((0, FIL__, __LINE__, _("msg=<Initialize modules.>\n")))
   for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
     {
-      if (0 != (status = modList[modnum].mod_init()))
+      status = modList[modnum].mod_init(&(modList[modnum]));
+      if ( status < 0 )
 	{
 	  if (status == (-1)) {
 	    sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, status, 
 			     MSG_MOD_FAIL,
 			     _(modList[modnum].name),
-			     status);
+			     status+SH_MOD_OFFSET);
 	  } else {
 	    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_MOD_FAIL,
 			     _(modList[modnum].name),
-			     status);
+			     status+SH_MOD_OFFSET);
 	  }
-	  modList[modnum].initval = S_FALSE;
+	  modList[modnum].initval = SH_MOD_FAILED;
 	}
       else
 	{
 	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_MOD_OK,
 			   _(modList[modnum].name));
-	  modList[modnum].initval = S_TRUE;
+	  modList[modnum].initval = status;
 	}
     }
     
@@ -1564,7 +1726,7 @@ int undef_main(int argc, char * argv[])
    */
   (void) sh_files_setrec();
   (void) sh_files_test_setup();
-
+  sh_audit_commit ();
 
   /* --------  NICE LEVEL   ---------
    */
@@ -1581,8 +1743,11 @@ int undef_main(int argc, char * argv[])
 
   /*  --------  MAIN LOOP  ---------
    */
-  sh.statistics.bytes_speed  = 0;
-  sh.statistics.bytes_hashed = 0;
+  sh.statistics.bytes_speed   = 0;
+  sh.statistics.bytes_hashed  = 0;
+  sh.statistics.files_report  = 0;
+  sh.statistics.files_error   = 0;
+  sh.statistics.files_nodir   = 0;
 
   while (1 == 1) 
     {
@@ -1616,28 +1781,39 @@ int undef_main(int argc, char * argv[])
 	      --sig_raised; 
 	    }
 	  
-	  if (sig_config_read_again == 1) /* SIGHUP */
+	  if (sig_config_read_again == 1 && /* SIGHUP */
+	      sh_global_suspend_flag == 0)
 	    {
 	      TPT((0, FIL__, __LINE__, _("msg=<Re-read configuration.>\n")))
 	      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_RECONF);
 
+	      sh_thread_pause_flag = S_TRUE;
+
 #if defined(WITH_EXTERNAL)
 	      /* delete list of external tasks
 	       */
 	      (void) sh_ext_cleanup();
 #endif
+#if defined(SH_WITH_MAIL)
+	      sh_nmail_free();
+#endif
+
 	      /* delete the file list, make all database
 	       * entries visible (allignore = FALSE)
 	       */
 	      (void) sh_files_deldirstack ();
 	      (void) sh_files_delfilestack ();
+	      (void) sh_files_delglobstack ();
 	      (void) sh_ignore_clean ();
 	      (void) hash_full_tree ();
+	      sh_audit_delete_all ();
+
 
 #if defined(SH_WITH_CLIENT)
 	      reset_count_dev_server();
 #endif
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
+	      sh_restrict_purge ();
 
 
 	      FileSchedOne = free_sched(FileSchedOne);
@@ -1645,14 +1821,15 @@ int undef_main(int argc, char * argv[])
 
 	      for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
 		{
-		  if (modList[modnum].initval == GOOD)
-		    (void) modList[modnum].mod_reconf();
+		  /* sh_thread_pause_flag is true, and we block in lock
+		   * until check has returned, so we are sure check will
+		   * not run until sh_thread_pause_flag is set to false
+		   */
+		  /* if (modList[modnum].initval >= SH_MOD_ACTIVE) */
+		  (void) modList[modnum].mod_reconf();
 		}
 #endif
 
-#if defined(SH_WITH_MAIL)
-	      reset_count_dev_mail();
-#endif
 	      reset_count_dev_console();
 	      reset_count_dev_time();
 
@@ -1663,6 +1840,16 @@ int undef_main(int argc, char * argv[])
 	       */
 #ifdef RELOAD_DATABASE
 	      sh_hash_hashdelete();
+
+	      if (0 != sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
+		{
+		  char hashbuf[KEYBUF_SIZE];
+		  (void) sl_strlcpy(sh.data.hash,
+				    sh_tiger_hash (file_path('D', 'R'), 
+						   TIGER_FILE, TIGER_NOLIM, 
+						   hashbuf, sizeof(hashbuf)), 
+				    KEY_LEN+1);
+		}
 #endif
 	      (void) sl_trust_purge_user();
 	      (void) sh_files_hle_reg (NULL);
@@ -1676,6 +1863,8 @@ int undef_main(int argc, char * argv[])
 	      sig_config_read_again = 0;
 	      (void) sh_files_setrec();
 	      (void) sh_files_test_setup();
+	      sh_audit_commit ();
+
 	      if (0 != sh.flag.nice)
 		{
 #ifdef HAVE_SETPRIORITY
@@ -1691,52 +1880,66 @@ int undef_main(int argc, char * argv[])
 		  sh.flag.loop     = S_FALSE;
 		}
 
+
 	      /* --- Initialize modules. ---
 	       */
 	      TPT((0, FIL__, __LINE__, _("msg=<Initialize modules.>\n")));
 	      for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
 		{
-		  if (0 != (status = modList[modnum].mod_init()))
+		  status = modList[modnum].mod_init(&(modList[modnum]));
+
+		  if (status < 0)
 		    {
 		      if (status == (-1)) {
 			sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, 
 					 status, MSG_MOD_FAIL,
 					 _(modList[modnum].name),
-					 status);
+					 status+SH_MOD_OFFSET);
 		      } else {
 			sh_error_handle ((-1), FIL__, __LINE__, 
 					 status, MSG_MOD_FAIL,
 					 _(modList[modnum].name),
-					 status);
+					 status+SH_MOD_OFFSET);
 		      }
-		      modList[modnum].initval = S_FALSE;
+		      modList[modnum].initval = SH_MOD_FAILED;
 		    }
 		  else
 		    {
 		      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_MOD_OK,
 				       _(modList[modnum].name));
-		      modList[modnum].initval = S_TRUE;
+		      modList[modnum].initval = status;
 		    }
 		}
+
+	      /* module is properly set up now
+	       */
+	      sh_thread_pause_flag = S_FALSE;
 	      
 	      --sig_raised;
 	    }
 	  
 	  if (sig_fresh_trail == 1) /* SIGIOT */
 	    {
-	      /* Logfile access 
-	       */
+	      if (sh_global_suspend_flag == 0)
+		{
+		  SH_MUTEX_LOCK(mutex_thread_nolog);
+
+		  /* Logfile access 
+		   */
 #ifdef SH_USE_XML
-	      (void) sh_log_file (NULL, NULL);
+		  (void) sh_log_file (NULL, NULL);
 #endif
-	      TPT((0, FIL__, __LINE__, _("msg=<Logfile stop/restart.>\n")));
-	      sh_error_only_stderr (S_TRUE);
-	      (void) sh_unix_rm_lock_file(sh.srvlog.name);
-	      (void) retry_msleep(3, 0);
-	      sh.flag.log_start = S_TRUE;
-	      sh_error_only_stderr (S_FALSE);
-	      sig_fresh_trail       = 0;
-	      --sig_raised;
+		  TPT((0, FIL__, __LINE__, _("msg=<Logfile stop/restart.>\n")));
+		  sh_error_only_stderr (S_TRUE);
+		  (void) sh_unix_rm_lock_file(sh.srvlog.name);
+		  (void) retry_msleep(3, 0);
+		  sh.flag.log_start = S_TRUE;
+		  sh_error_only_stderr (S_FALSE);
+		  sh_thread_pause_flag = S_FALSE;
+		  sig_fresh_trail       = 0;
+		  --sig_raised;
+		  SH_MUTEX_UNLOCK(mutex_thread_nolog);
+		}
 	    }
 	  
 	  if (sig_terminate == 1)  /* SIGQUIT */
@@ -1758,12 +1961,14 @@ int undef_main(int argc, char * argv[])
 	  if (sig_suspend_switch > 0)  /* SIGUSR2 */
 	    {
 	      TPT((0, FIL__, __LINE__, _("msg=<Suspend switch.>\n")));
-	      if (sh_global_suspend_flag == 1) {
-		sh_global_suspend_flag = 0;
-	      } else {
+	      if (sh_global_suspend_flag != 1) {
+		SH_MUTEX_LOCK_UNSAFE(mutex_thread_nolog);
+		sh_global_suspend_flag = 1;
 		sh_error_handle((-1), FIL__, __LINE__, 0, MSG_SUSPEND, 
 				sh.prg_name);
-		sh_global_suspend_flag = 1;
+	      } else {
+		sh_global_suspend_flag = 0;
+		SH_MUTEX_UNLOCK_UNSAFE(mutex_thread_nolog);
 	      }
 	      --sig_suspend_switch;
 	      --sig_raised; --sig_urgent;
@@ -1782,6 +1987,7 @@ int undef_main(int argc, char * argv[])
       /* see whether its time to check files
        */
       if      (sh.flag.checkSum == SH_CHECK_INIT ||
+	       (sh.flag.inotify & SH_INOTIFY_DOSCAN) != 0 ||
 	       (sh.flag.checkSum == SH_CHECK_CHECK &&
 		(sh.flag.isdaemon == S_FALSE && sh.flag.loop == S_FALSE)))
 	{
@@ -1816,6 +2022,12 @@ int undef_main(int argc, char * argv[])
       if (sh.flag.checkSum != SH_CHECK_NONE &&
 	  (flag_check_1 == 1 || flag_check_2 == 1))
 	{
+	  SH_INOTIFY_IFUSED( sh.flag.inotify |= SH_INOTIFY_INSCAN; );
+	  /* Refresh list files matching glob patterns.
+	   */
+	  if (sh.flag.checkSum != SH_CHECK_INIT)
+	    sh_files_check_globPatterns();
+
 	  /* 
 	   * check directories and files
 	   * ORDER IS IMPORTANT -- DIRZ FIRST
@@ -1824,6 +2036,9 @@ int undef_main(int argc, char * argv[])
 	  sh.statistics.time_start     = time (NULL);
 	  sh.statistics.dirs_checked   = 0;
 	  sh.statistics.files_checked  = 0;
+	  sh.statistics.files_report   = 0;
+	  sh.statistics.files_error    = 0;
+	  sh.statistics.files_nodir    = 0;
 
 	  TPT((0, FIL__, __LINE__, _("msg=<Check directories.>\n")))
 	  BREAKEXIT(sh_dirs_chk);
@@ -1831,14 +2046,14 @@ int undef_main(int argc, char * argv[])
 	    {
 	      (void) sh_dirs_chk  (1);
 #ifndef SH_PROFILE
-	      (void) chdir ("/");
+	      (void) retry_aud_chdir (FIL__, __LINE__, "/");
 #endif
 	    }
 	  if (flag_check_2 == 1)
 	    {
 	      (void) sh_dirs_chk  (2); 
 #ifndef SH_PROFILE
-	      (void) chdir ("/");
+	      (void) retry_aud_chdir (FIL__, __LINE__, "/");
 #endif
 	    }
 	  TPT((0, FIL__, __LINE__, _("msg=<Check files.>\n")))
@@ -1872,6 +2087,8 @@ int undef_main(int argc, char * argv[])
 	  flag_check_1 = 0;
 	  flag_check_2 = 0;
 	  check_done   = 1;
+	  SH_INOTIFY_IFUSED( sh.flag.inotify &= ~SH_INOTIFY_INSCAN; );
+	  SH_INOTIFY_IFUSED( sh.flag.inotify &= ~SH_INOTIFY_DOSCAN; );
 
 	  (void) sh_prelink_run (NULL, NULL, 0);
 
@@ -1903,6 +2120,9 @@ int undef_main(int argc, char * argv[])
 	      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_CHECK_1,
 			       (long) runtim, 
 			       0.001 * st_1);
+
+	      if (sh.flag.checkSum != SH_CHECK_INIT)
+		sh_efile_report();
 	    }
 	  sh.fileCheck.alarm_last = time (NULL);
 
@@ -1914,7 +2134,7 @@ int undef_main(int argc, char * argv[])
 	   */
 #if defined(SH_WITH_MAIL)
 	  TPT((0, FIL__, __LINE__, _("msg=<Flush mail queue.>\n")))
-	  (void) sh_mail_msg (NULL);
+	  (void) sh_nmail_flush ();
 #endif
 	}
       
@@ -1926,11 +2146,11 @@ int undef_main(int argc, char * argv[])
       TPT((0, FIL__, __LINE__, _("msg=<Execute modules.>\n")))
       for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
 	{
-	  if (modList[modnum].initval == GOOD &&
+	  if (modList[modnum].initval == SH_MOD_ACTIVE &&
 	      0 != modList[modnum].mod_timer(tcurrent))
 	    if (0 != (status = modList[modnum].mod_check()))
 	      sh_error_handle ((-1), FIL__, __LINE__, status, MSG_MOD_EXEC,
-			       _(modList[modnum].name), (long) status);
+			       _(modList[modnum].name), (long) (status+SH_MOD_OFFSET));
 	}
       
       /* 27.05.2002 avoid empty database
@@ -1946,6 +2166,30 @@ int undef_main(int argc, char * argv[])
 	  check_done == 1)
 	sh_hash_writeout ();
 
+      /* no-op unless MEM_LOG is defined in sh_mem.c
+       */
+#ifdef MEM_DEBUG
+      sh_mem_dump ();
+#endif
+
+      {
+	char * stale;
+
+	stale = sl_check_stale();
+	if (stale)
+	  {
+	    sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, 0, MSG_E_SUBGEN,
+			    stale, _("sl_check_stale"));
+	  }
+
+	stale = sl_check_badfd();
+	if (stale)
+	  {
+	    sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, 0, MSG_E_SUBGEN,
+			    stale, _("sl_check_stale"));
+	  }
+      }
+
       /* no loop if not daemon
        */
       if (sh.flag.isdaemon != S_TRUE && sh.flag.loop == S_FALSE)
@@ -1959,7 +2203,7 @@ int undef_main(int argc, char * argv[])
       if (tcurrent - sh.mailTime.alarm_last >= sh.mailTime.alarm_interval) 
 	{
 	  TPT((0, FIL__, __LINE__, _("msg=<Flush mail queue.>\n")))
-	  (void) sh_mail_msg (NULL);
+	  (void) sh_nmail_flush ();
 	  sh.mailTime.alarm_last = time (NULL);
 	}
 #endif
@@ -1987,6 +2231,10 @@ int undef_main(int argc, char * argv[])
       if (sig_urgent > 0)
 	continue;
       
+      /* reset cache
+       */
+      sh_userid_destroy();
+
       /* go to sleep
        */
       (void) retry_msleep (1, 0);
@@ -2008,12 +2256,21 @@ int undef_main(int argc, char * argv[])
 
 #if defined(SH_WITH_MAIL)
   if (sh.mailNum.alarm_last > 0) 
-    (void)sh_mail_msg (NULL);
+    (void)sh_nmail_flush ();
 #endif
 
   /* #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) */
 #endif
 
+#if 0
+  {
+    char command[128];
+    sprintf(command, "/bin/cat /proc/%d/status", (int) getpid());
+    system(command); /* flawfinder: ignore *//* debug code */
+    malloc_stats();
+  }
+#endif
+
   aud_exit (FIL__, __LINE__, EXIT_SUCCESS);
   SL_RETURN(0, _("main"));
 }
