@@ -19,6 +19,10 @@
 
 #include "config_xor.h"
 
+#ifdef HOST_IS_HPUX          
+#define _XOPEN_SOURCE_EXTENDED
+#endif                       
+
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
@@ -29,6 +33,7 @@
 #include <fcntl.h>
 #include <signal.h>
 #include <sys/socket.h>
+#include <netdb.h>
 #include <netinet/in.h>
 
 #ifndef S_SPLINT_S
@@ -46,15 +51,20 @@
 #include "samhain.h"
 #include "sh_error.h"
 #include "sh_calls.h"
+#include "sh_ipvx.h"
+#include "sh_sub.h"
+#include "sh_utils.h"
 
 #undef  FIL__
 #define FIL__  _("sh_calls.c")
 
+extern int flag_err_debug;
+
 char aud_err_message[64];
 
 typedef struct cht_struct 
 {
-  char           * str;
+  const char           * str;
   unsigned long    val;
 } cht_type;
 
@@ -77,7 +87,7 @@ static cht_type aud_tab[] =
 
 /* Set aud functions
  */
-int sh_aud_set_functions(char * str_s)
+int sh_aud_set_functions(const char * str_s)
 {
   int i = 0;
 
@@ -104,12 +114,13 @@ int sh_aud_set_functions(char * str_s)
 
 /* Need to catch EINTR for these functions.
  */
-long int retry_sigaction(char * file, int line,
+long int retry_sigaction(const char * file, int line,
 			 int signum,  const  struct  sigaction  *act,
 			 struct sigaction *oldact)
 {
   int error;
   long int val_retry = -1;
+  char errbuf[SH_ERRBUF_SIZE];
   errno              = 0;
 
   SL_ENTER(_("retry_sigaction"));
@@ -121,33 +132,45 @@ long int retry_sigaction(char * file, int line,
   error = errno;
   if (val_retry < 0) {
       sh_error_handle ((-1), file, line, error, MSG_ERR_SIGACT, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) signum );
   }
   errno = error;    
   SL_RETURN(val_retry, _("retry_sigaction"));
 }
 
-static struct in_addr bind_addr;
+static struct sh_sockaddr bind_addr;
 static int        use_bind_addr = 0;
 
-int sh_calls_set_bind_addr (char * str)
+int sh_calls_set_bind_addr (const char * str)
 {
-  if (0 == /*@-unrecog@*/inet_aton(str, &bind_addr)/*@+unrecog@*/) 
-    {
-      return -1;
-    }
+  static int reject = 0;
+
+  if (reject == 1)
+    return (0);
+
+  if (sh.flag.opts == S_TRUE)  
+    reject = 1;
+
+#if defined(USE_IPVX)
+  if (0 == sh_ipvx_aton(str, &bind_addr)) 
+    return -1;
+#else
+  if (0 == inet_aton(str, &(bind_addr.sin.sin_addr))) 
+    return -1;
+#endif
+
   use_bind_addr = 1;
   return 0;
 }
 
 
-long int retry_connect(char * file, int line, int sockfd, 
+long int retry_connect(const char * file, int line, int sockfd, 
 		       struct sockaddr *serv_addr, int addrlen)
 {
   int error;
   long int val_retry = 0;
-  static struct sockaddr_in new_addr;
+  char errbuf[SH_ERRBUF_SIZE];
 
   SL_ENTER(_("retry_connect"));
 
@@ -155,123 +178,240 @@ long int retry_connect(char * file, int line, int sockfd,
 
   if (0 != use_bind_addr) 
     {
-      memcpy(&new_addr.sin_addr, &bind_addr, sizeof(struct in_addr));
-      new_addr.sin_family = AF_INET;
-      
-      val_retry = /*@-unrecog@*/bind(sockfd, 
-				     (struct sockaddr*)&new_addr, 
-				     sizeof(struct sockaddr_in))/*@+unrecog@*/;
+      int slen = SH_SS_LEN(bind_addr);
+
+      val_retry = bind(sockfd, sh_ipvx_sockaddr_cast(&bind_addr), slen);
     }
 
   if (val_retry == 0)
     {
       do {
-	val_retry = 
-	  /*@-unrecog@*/connect(sockfd, serv_addr, addrlen)/*@+unrecog@*/;
-      } while (val_retry < 0 && errno == EINTR);
+	val_retry = connect(sockfd, serv_addr, addrlen);
+      } while (val_retry < 0 && (errno == EINTR || errno == EINPROGRESS));
     }
 
   error = errno;
   if (val_retry != 0) {
-    /* ugly cast back to struct sockaddr_in :-(
-     */
+    long eport;
+    char eaddr[SH_IP_BUF];
+
+    struct sh_sockaddr ss;
+    sh_ipvx_save(&ss, serv_addr->sa_family, serv_addr);
+    sh_ipvx_ntoa(eaddr, sizeof(eaddr), &ss);
+    
+    if (serv_addr->sa_family == AF_INET)
+      eport = (long) ntohs(((struct sockaddr_in *)serv_addr)->sin_port);
+    else
+      eport = (long) ntohs(((struct sockaddr_in6 *)serv_addr)->sin6_port);
+
     sh_error_handle ((-1), file, line, error, MSG_ERR_CONNECT, 
-		     sh_error_message(error),
-		     (long) sockfd,
-		     /*@-unrecog -type@*/
-		     (long) ntohs(((struct sockaddr_in *)serv_addr)->sin_port),
-		     /*@+unrecog +type@*/
-#ifdef HAVE_INET_ATON
-		     /*@-unrecog -type@*/
-		     inet_ntoa( ((struct sockaddr_in *)serv_addr)->sin_addr )
-		     /*@+unrecog +type@*/
-#else
-		     _("unknown")
-#endif
-		     );
+		     sh_error_message(error, errbuf, sizeof(errbuf)),
+		     (long) sockfd, eport, eaddr);
   }
   errno = error;    
   SL_RETURN(val_retry, _("retry_connect"));
 }
 
-long int retry_accept(char * file, int line, int fd, 
-		      struct sockaddr *serv_addr, int * addrlen)
+long int retry_accept(const char * file, int line, int fd, 
+		      struct sh_sockaddr *serv_addr, int * addrlen)
 {
   int  error;
   long int val_retry = -1;
-#ifdef HOST_IS_AIX
-  /* on AIX, addrlen is a size_t
-   */
-  size_t my_addrlen  = (size_t) *addrlen;
-#else
-  int    my_addrlen  = *addrlen;
-#endif
+  char errbuf[SH_ERRBUF_SIZE];
+  struct sockaddr_storage ss;
+
+  ACCEPT_TYPE_ARG3 my_addrlen = sizeof(ss);
 
   errno              = 0;
 
   SL_ENTER(_("retry_accept"));
 
   do {
-    val_retry = /*@-unrecog@*/accept(fd, serv_addr, &my_addrlen)/*@+unrecog@*/;
+    val_retry = accept(fd, (struct sockaddr *)&ss, &my_addrlen);
   } while (val_retry < 0 && errno == EINTR);
-  *addrlen = (int) my_addrlen;
+
   error = errno;
   if (val_retry < 0) {
-      sh_error_handle ((-1), file, line, error, MSG_ERR_ACCEPT, sh_error_message(error),
+      sh_error_handle ((-1), file, line, error, MSG_ERR_ACCEPT, 
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) fd );
   }
-  errno = error;    
+  errno = error;
+
+  if (flag_err_debug == SL_TRUE)
+    {
+      char ipbuf[SH_IP_BUF];
+      char buf[SH_BUFSIZE];
+#if defined(USE_IPVX)
+      sl_strlcpy(errbuf, _("Address family: "), sizeof(errbuf));
+      sl_strlcat(errbuf, 
+		 (ss.ss_family == AF_INET6) ? _("AF_INET6") : _("AF_INET"),
+		 sizeof(errbuf));
+      getnameinfo((struct sockaddr *)&ss, my_addrlen,
+		  ipbuf, sizeof(ipbuf), NULL, 0, NI_NUMERICHOST);
+#else
+      struct sockaddr_in sa;
+      char * p;
+      memcpy(&(sa), (struct sockaddr_in*)&ss, sizeof(struct sockaddr_in));
+      p = inet_ntoa(sa.sin_addr);
+      sl_strlcpy(ipbuf, p, sizeof(ipbuf));
+      sl_strlcpy(errbuf, _("Address family: AF_INET"), sizeof(errbuf));
+#endif
+      sl_strlcpy(buf, _("Address: "), sizeof(buf));
+      sl_strlcat(buf, ipbuf, sizeof(buf));
+      sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN,
+		       errbuf, _("retry_accept"));
+      sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN,
+		       buf, _("retry_accept"));
+    }
+
+  sh_ipvx_save(serv_addr, ss.ss_family, (struct sockaddr *) &ss);
+
+  if (flag_err_debug == SL_TRUE)
+    {
+      char ipbuf[SH_IP_BUF];
+      char ipbuf2[SH_IP_BUF];
+      char buf[SH_BUFSIZE];
+#if defined(USE_IPVX)
+      int len = (serv_addr->ss_family == AF_INET) ? 
+	sizeof(struct sockaddr_in) :
+	sizeof(struct sockaddr_in6);
+      getnameinfo(sh_ipvx_sockaddr_cast(serv_addr), len,
+		  ipbuf2, sizeof(ipbuf2), NULL, 0, NI_NUMERICHOST);
+#else
+      char * p = inet_ntoa((serv_addr->sin).sin_addr);
+      sl_strlcpy(ipbuf2, p, sizeof(ipbuf2));
+#endif
+      sh_ipvx_ntoa (ipbuf, sizeof(ipbuf), serv_addr);
+      sl_snprintf(buf, sizeof(buf), _("Address: %s / %s"),
+		  ipbuf, ipbuf2);
+      sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN,
+		       buf, _("retry_accept"));
+    }
+
+  *addrlen = (int) my_addrlen;
   SL_RETURN(val_retry, _("retry_accept"));
 }
 
-long int retry_lstat(char * file, int line, 
-		     const char *file_name, struct stat *buf)
+static int sh_enable_use_sub = 0;
+
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+static int sh_use_sub = 1;
+#else
+static int sh_use_sub = 0;
+#endif
+
+void sh_calls_enable_sub()
+{
+  sh_enable_use_sub = 1;
+  return;
+}
+
+int sh_calls_set_sub (const char * str)
+{
+  int ret = sh_util_flagval(str, &sh_use_sub);
+
+  if ((ret == 0) && (!sh_use_sub))
+    {
+      sh_kill_sub();
+    }
+  return ret;
+}
+
+long int retry_lstat_ns(const char * file, int line, 
+			const char *file_name, struct stat *buf)
 {
   int error;
   long int val_retry = -1;
+  char errbuf[SH_ERRBUF_SIZE];
  
-  SL_ENTER(_("retry_lstat"));
+  SL_ENTER(_("retry_lstat_ns"));
 
   do {
     val_retry = /*@-unrecog@*/lstat (file_name, buf)/*@+unrecog@*/;
   } while (val_retry < 0 && errno == EINTR);
+
   error = errno;
   if (val_retry < 0) {
-      (void) sl_strlcpy(aud_err_message, sh_error_message(error), 64);
+      (void) sh_error_message(error, aud_err_message, 64);
       sh_error_handle ((-1), file, line, error, MSG_ERR_LSTAT, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       file_name );
   }
   errno = error;    
+
+  SL_RETURN(val_retry, _("retry_lstat_ns"));
+}
+
+long int retry_lstat(const char * file, int line, 
+		     const char *file_name, struct stat *buf)
+{
+  int error;
+  long int val_retry = -1;
+  char errbuf[SH_ERRBUF_SIZE];
+ 
+  SL_ENTER(_("retry_lstat"));
+
+  if (sh_use_sub && sh_enable_use_sub)
+    {
+      val_retry = sh_sub_lstat (file_name, buf);
+    }
+  else
+    {
+      do {
+	val_retry = /*@-unrecog@*/lstat (file_name, buf)/*@+unrecog@*/;
+      } while (val_retry < 0 && errno == EINTR);
+    }
+
+  error = errno;
+  if (val_retry < 0) {
+      (void) sh_error_message(error, aud_err_message, 64);
+      sh_error_handle ((-1), file, line, error, MSG_ERR_LSTAT, 
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
+		       file_name );
+  }
+  errno = error;    
+
   SL_RETURN(val_retry, _("retry_lstat"));
 }
 
-long int retry_stat(char * file, int line, 
+long int retry_stat(const char * file, int line, 
 		    const char *file_name, struct stat *buf)
 {
   int error;
   long int val_retry = -1;
+  char errbuf[SH_ERRBUF_SIZE];
  
   SL_ENTER(_("retry_stat"));
 
-  do {
-    val_retry = stat (file_name, buf);
-  } while (val_retry < 0 && errno == EINTR);
+  if (sh_use_sub && sh_enable_use_sub)
+    {
+      val_retry = sh_sub_stat (file_name, buf);
+    }
+  else
+    {
+      do {
+	val_retry = stat (file_name, buf);
+      } while (val_retry < 0 && errno == EINTR);
+    }
+
   error = errno;
   if (val_retry < 0) {
-      (void) sl_strlcpy(aud_err_message, sh_error_message(error), 64);
+      (void) sh_error_message(error, aud_err_message, 64);
       sh_error_handle ((-1), file, line, error, MSG_ERR_STAT, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       file_name );
   }
   errno = error;    
+
   SL_RETURN(val_retry, _("retry_stat"));
 }
 
-long int retry_fstat(char * file, int line, int filed, struct stat *buf)
+long int retry_fstat(const char * file, int line, int filed, struct stat *buf)
 {
   int error;
   long int val_retry = -1;
+  char errbuf[SH_ERRBUF_SIZE];
  
   SL_ENTER(_("retry_fstat"));
 
@@ -281,17 +421,18 @@ long int retry_fstat(char * file, int line, int filed, struct stat *buf)
   error = errno;
   if (val_retry < 0) {
       sh_error_handle ((-1), file, line, error, MSG_ERR_FSTAT, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) filed );
   }
   errno = error;    
   SL_RETURN(val_retry, _("retry_fstat"));
 }
 
-long int retry_fcntl(char * file, int line, int fd, int cmd, long arg)
+long int retry_fcntl(const char * file, int line, int fd, int cmd, long arg)
 {
   int error;
   long int val_retry = -1;
+  char errbuf[SH_ERRBUF_SIZE];
   errno              = 0;
 
   SL_ENTER(_("retry_fcntl"));
@@ -311,7 +452,7 @@ long int retry_fcntl(char * file, int line, int fd, int cmd, long arg)
   error = errno;
   if (val_retry < 0) {
       sh_error_handle ((-1), file, line, error, MSG_ERR_FCNTL, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) fd, (long) cmd, arg );
   }
   errno = error;    
@@ -325,7 +466,7 @@ long int retry_msleep (int sec, int millisec)
   struct timespec req, rem;
 #endif
 
-  SL_ENTER(_("retry_fcntl"));
+  SL_ENTER(_("retry_msleep"));
 
   errno  = 0;
   if (millisec > 999) millisec = 999;
@@ -347,7 +488,7 @@ long int retry_msleep (int sec, int millisec)
 #else
   if (sec > 0)
     {
-      sleep (sec);
+      sleep (sec); /* nanosleep not available */
     }
   else
     {
@@ -373,7 +514,7 @@ long int retry_msleep (int sec, int millisec)
  *
  ***************************************************/
 
-long int retry_aud_execve  (char * file, int line, 
+long int retry_aud_execve  (const char * file, int line, 
 			    const  char *dateiname, char * argv[],
 			    char * envp[])
 {
@@ -381,6 +522,7 @@ long int retry_aud_execve  (char * file, int line,
   gid_t b = getegid();
   int   i;
   int   error;
+  char errbuf[SH_ERRBUF_SIZE];
 
   SL_ENTER(_("retry_aud_execve"));
 
@@ -393,7 +535,8 @@ long int retry_aud_execve  (char * file, int line,
 
   error = errno;
   if (i < 0) {
-      sh_error_handle ((-1), file, line, error, MSG_ERR_EXEC, sh_error_message(error),
+      sh_error_handle ((-1), file, line, error, MSG_ERR_EXEC, 
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       dateiname, (long) a, (long) b );
   }
   errno = error;    
@@ -401,11 +544,12 @@ long int retry_aud_execve  (char * file, int line,
 }
 
 
-long int retry_aud_utime (char * file, int line, 
-			   char * path, struct utimbuf *buf)
+long int retry_aud_utime (const char * file, int line, 
+			  char * path, struct utimbuf *buf)
 {
   long int val_return;
   int  error;
+  char errbuf[SH_ERRBUF_SIZE];
   errno      = 0;
 
   SL_ENTER(_("retry_aud_utime"));
@@ -422,7 +566,7 @@ long int retry_aud_utime (char * file, int line,
 		     (unsigned long) buf->modtime);
   if (val_return < 0) {
       sh_error_handle ((-1), file, line, error, MSG_ERR_UTIME, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       path, 
 		       (unsigned long) buf->actime, 
 		       (unsigned long) buf->modtime);
@@ -431,11 +575,12 @@ long int retry_aud_utime (char * file, int line,
   SL_RETURN(val_return, _("retry_aud_utime"));
 }
 
-long int retry_aud_unlink (char * file, int line, 
+long int retry_aud_unlink (const char * file, int line, 
 			   char * path)
 {
   long int val_return;
   int error;
+  char errbuf[SH_ERRBUF_SIZE];
   errno      = 0;
 
   SL_ENTER(_("retry_aud_unlink"));
@@ -449,18 +594,20 @@ long int retry_aud_unlink (char * file, int line,
     sh_error_handle ((-1), file, line, 0, MSG_AUD_UNLINK,
 		     path);
   if (val_return < 0) {
-      sh_error_handle ((-1), file, line, error, MSG_ERR_UNLINK, sh_error_message(error),
+      sh_error_handle ((-1), file, line, error, MSG_ERR_UNLINK, 
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       path);
   }
   errno = error;
   SL_RETURN(val_return, _("retry_aud_unlink"));
 }
 
-long int retry_aud_dup2 (char * file, int line, 
+long int retry_aud_dup2 (const char * file, int line, 
 			 int fd, int fd2)
 {
   long int val_return;
   int error;
+  char errbuf[SH_ERRBUF_SIZE];
   errno      = 0;
 
   SL_ENTER(_("retry_aud_dup2"));
@@ -475,18 +622,19 @@ long int retry_aud_dup2 (char * file, int line,
 		      (long) fd, val_return);
   if (val_return < 0) {
       sh_error_handle ((-1), file, line, error, MSG_ERR_DUP, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) fd, val_return);
   }
   errno = error;
   SL_RETURN(val_return, _("retry_aud_dup2"));
 }
 
-long int retry_aud_dup (char * file, int line, 
+long int retry_aud_dup (const char * file, int line, 
 			int fd)
 {
   long int val_return;
   int error;
+  char errbuf[SH_ERRBUF_SIZE];
   errno      = 0;
 
   SL_ENTER(_("retry_aud_dup"));
@@ -500,7 +648,7 @@ long int retry_aud_dup (char * file, int line,
 		     (long) fd, val_return);
   if (val_return < 0) {
       sh_error_handle ((-1), file, line, error, MSG_ERR_DUP, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) fd, val_return);
   }
   errno = error;
@@ -509,11 +657,12 @@ long int retry_aud_dup (char * file, int line,
 
 
   
-long int retry_aud_chdir (char * file, int line, 
+long int retry_aud_chdir (const char * file, int line, 
 			  const char *path)
 {
   long int val_return;
   int      error      = 0;
+  char errbuf[SH_ERRBUF_SIZE];
   errno      = 0;
 
   SL_ENTER(_("retry_aud_chdir"));
@@ -527,31 +676,57 @@ long int retry_aud_chdir (char * file, int line,
     sh_error_handle ((-1), file, line, 0, MSG_AUD_CHDIR,
 		     path);
   if (val_return < 0) {
-      sh_error_handle ((-1), file, line, error, MSG_ERR_CHDIR, sh_error_message(error),
+      sh_error_handle ((-1), file, line, error, MSG_ERR_CHDIR, 
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       path);
   }
   errno = error;
   SL_RETURN(val_return, _("retry_aud_chdir"));
 }
-  
 
-long int aud_open (char * file, int line, int privs,
-		   const char *pathname, int flags, mode_t mode)
+
+long int aud_open_noatime (const char * file, int line, int privs,
+			   const char *pathname, int flags, mode_t mode,
+			   int * o_noatime)
 {
   long int val_return;
   int error;
+  char errbuf[SH_ERRBUF_SIZE];
 
   SL_ENTER(_("aud_open"));
 
-  val_return = open (pathname, flags, mode);
-  error = errno;
+#ifdef USE_SUID
+  if (0 == strcmp(pathname, "/usr/bin/sudo"))
+    {
+      uid_t ruid; uid_t euid; uid_t suid;
+      getresuid(&ruid, &euid, &suid);
+    }
+  if (privs == SL_YESPRIV)
+    sl_set_suid();
+#else
   /*@-noeffect@*/
   (void) privs; /* fix compiler warning */
   /*@+noeffect@*/
+#endif
+
+  val_return = open (pathname, *o_noatime|flags, mode);
+
+#ifdef USE_SUID
+  if (privs == SL_YESPRIV)
+    sl_unset_suid();
+#endif
+
+  if ((val_return < 0) && (*o_noatime != 0))
+    {
+      val_return = open (pathname, flags, mode);
+      if (val_return >= 0)
+	*o_noatime = 0;
+    }
+  error = errno;
 
   if (val_return < 0)
     {
-      (void) sl_strlcpy(aud_err_message, sh_error_message(error), 64);
+      (void) sh_error_message(error, aud_err_message, 64);
     }
 
   if (sh.flag.audit != 0 && (sh.flag.aud_mask & AUD_OPEN) != 0)
@@ -561,17 +736,64 @@ long int aud_open (char * file, int line, int privs,
     }
   if (val_return < 0) {
     sh_error_handle ((-1), file, line, error, MSG_ERR_OPEN, 
-		     sh_error_message(error),
+		     sh_error_message(error, errbuf, sizeof(errbuf)),
+		     pathname, (long) flags, (long) mode, val_return);
+  }
+  errno = error;
+  SL_RETURN(val_return, _("aud_open"));
+}
+
+long int aud_open (const char * file, int line, int privs,
+		   const char *pathname, int flags, mode_t mode)
+{
+  long int val_return;
+  int error;
+  char errbuf[SH_ERRBUF_SIZE];
+
+  SL_ENTER(_("aud_open"));
+
+#ifdef USE_SUID
+  if (privs == SL_YESPRIV)
+    sl_set_suid();
+#else
+  /*@-noeffect@*/
+  (void) privs; /* fix compiler warning */
+  /*@+noeffect@*/
+#endif
+
+  val_return = open (pathname, flags, mode);
+
+#ifdef USE_SUID
+  if (privs == SL_YESPRIV)
+    sl_unset_suid();
+#endif
+
+  error = errno;
+
+  if (val_return < 0)
+    {
+      (void) sh_error_message(error, aud_err_message, 64);
+    }
+
+  if (sh.flag.audit != 0 && (sh.flag.aud_mask & AUD_OPEN) != 0)
+    {
+      sh_error_handle ((-1), file, line, 0, MSG_AUD_OPEN,
+		       pathname, (long) flags, (long) mode, val_return);
+    }
+  if (val_return < 0) {
+    sh_error_handle ((-1), file, line, error, MSG_ERR_OPEN, 
+		     sh_error_message(error, errbuf, sizeof(errbuf)),
 		     pathname, (long) flags, (long) mode, val_return);
   }
   errno = error;
   SL_RETURN(val_return, _("aud_open"));
 }
   
-long int aud_kill (char * file, int line, pid_t pid, int sig)
+long int aud_kill (const char * file, int line, pid_t pid, int sig)
 {
   int  myerror;
   long int val_return = kill (pid, sig);
+  char errbuf[SH_ERRBUF_SIZE];
   myerror = errno;
 
   SL_ENTER(_("aud_kill"));
@@ -581,7 +803,7 @@ long int aud_kill (char * file, int line, pid_t pid, int sig)
 		      (long) pid, (long) sig);
   if (val_return < 0) {
       sh_error_handle ((-1), file, line, myerror, MSG_ERR_KILL, 
-		       sh_error_message(myerror),
+		       sh_error_message(myerror, errbuf, sizeof(errbuf)),
 		       (long) pid, (long) sig);
   }
   errno = myerror;
@@ -589,7 +811,7 @@ long int aud_kill (char * file, int line, pid_t pid, int sig)
 }
   
 /*@noreturn@*/
-void aud_exit (char * file, int line, int fd)
+void aud_exit (const char * file, int line, int fd)
 {
   if (sh.flag.audit != 0 && (sh.flag.aud_mask & AUD_EXIT) != 0)
     sh_error_handle ((-1), file, line, 0, MSG_AUD_EXIT,
@@ -602,7 +824,7 @@ void aud_exit (char * file, int line, int fd)
 }
 
 /*@noreturn@*/
-void aud__exit (char * file, int line, int fd)
+void aud__exit (const char * file, int line, int fd)
 {
   if (sh.flag.audit != 0 && (sh.flag.aud_mask & AUD_EXIT) != 0)
     sh_error_handle ((-1), file, line, 0, MSG_AUD_EXIT,
@@ -614,10 +836,11 @@ void aud__exit (char * file, int line, int fd)
   _exit(fd);
 }
 
-pid_t aud_fork (char * file, int line)
+pid_t aud_fork (const char * file, int line)
 {
   int error;
   pid_t i = fork();
+  char errbuf[SH_ERRBUF_SIZE];
 
   error = errno;
   SL_ENTER(_("aud_fork"));
@@ -627,17 +850,18 @@ pid_t aud_fork (char * file, int line)
 		      (long) i);
   if (i == (pid_t) -1) {
     sh_error_handle ((-1), file, line, error, MSG_ERR_FORK, 
-		     sh_error_message(error),
+		     sh_error_message(error, errbuf, sizeof(errbuf)),
 		     (long) i);
   }
   errno = error;
   SL_RETURN(i, _("aud_fork"));
 }
 
-int aud_setuid (char * file, int line, uid_t uid)
+int aud_setuid (const char * file, int line, uid_t uid)
 {
   int error = 0;
   int i = 0;
+  char errbuf[SH_ERRBUF_SIZE];
 
   SL_ENTER(_("aud_setuid"));
 
@@ -654,17 +878,18 @@ int aud_setuid (char * file, int line, uid_t uid)
   }
   if (i < 0) {
     sh_error_handle ((-1), file, line, error, MSG_ERR_SETUID, 
-		     sh_error_message(error),
+		     sh_error_message(error, errbuf, sizeof(errbuf)),
 		     (long) uid);
   }
   errno = error;
   SL_RETURN(i, _("aud_setuid"));
 }
 
-int aud_setgid (char * file, int line, gid_t gid)
+int aud_setgid (const char * file, int line, gid_t gid)
 {
   int error = 0;
   int i = 0;
+  char errbuf[SH_ERRBUF_SIZE];
 
   SL_ENTER(_("aud_setgid"));
 
@@ -682,17 +907,18 @@ int aud_setgid (char * file, int line, gid_t gid)
   }
   if (i < 0) {
     sh_error_handle ((-1), file, line, error, MSG_ERR_SETGID, 
-		     sh_error_message(error),
+		     sh_error_message(error, errbuf, sizeof(errbuf)),
 		     (long) gid);
   }
   errno = error;
   SL_RETURN(i, _("aud_setgid"));
 }
 
-int aud_pipe (char * file, int line, int * modus)
+int aud_pipe (const char * file, int line, int * modus)
 {
   int error;
   int i = pipe (modus);
+  char errbuf[SH_ERRBUF_SIZE];
 
   SL_ENTER(_("aud_pipe"));
 
@@ -709,11 +935,11 @@ int aud_pipe (char * file, int line, int * modus)
   if (i < 0) {
     if (i < 0)
       sh_error_handle ((-1), file, line, error, MSG_ERR_PIPE, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) 0, (long) 0);
     else
       sh_error_handle ((-1), file, line, error, MSG_ERR_PIPE, 
-		       sh_error_message(error),
+		       sh_error_message(error, errbuf, sizeof(errbuf)),
 		       (long) modus[0], (long) modus[1]);
   }
   SL_RETURN(i, _("aud_pipe"));
