@@ -39,16 +39,16 @@
 static FILE * pdbg = NULL;
 static FILE * pdbgc = NULL;
 #define PDBG_OPEN    if (pdbg == NULL) pdbg = fopen(PDGBFILE"main",  "a")  
-#define PDBG_CLOSE   fclose (pdbg); pdbg = NULL
+#define PDBG_CLOSE   sl_fclose (FIL__, __LINE__, pdbg); pdbg = NULL
 #define PDBG(arg)    fprintf(pdbg,  "PDBG: step %d\n", arg); fflush(pdbg)
 #define PDBG_D(arg)  fprintf(pdbg,  "PDBG: %d\n", arg); fflush(pdbg)
 #define PDBG_S(arg)  fprintf(pdbg,  "PDBG: %s\n", arg); fflush(pdbg)
 
 #define PDBGC_OPEN   if (pdbgc == NULL) pdbgc = fopen(PDGBFILE"child", "a")  
-#define PDBGC_CLOSE  fclose (pdbgc); pdbgc = NULL
-#define PDBGC(arg)   fprintf(pdbgc, "PDBG: step %d\n", arg); fflush(pdbgc)
-#define PDBGC_D(arg) fprintf(pdbgc, "PDBG: %d\n", arg); fflush(pdbgc)
-#define PDBGC_S(arg) fprintf(pdbgc, "PDBG: %s\n", arg); fflush(pdbgc)
+#define PDBGC_CLOSE  sl_fclose (FIL__, __LINE__, pdbgc); pdbgc = NULL
+#define PDBGC(arg)   fprintf(pdbgc, "PDBGC: step %d\n", arg); fflush(pdbgc)
+#define PDBGC_D(arg) fprintf(pdbgc, "PDBGC: %d\n", arg); fflush(pdbgc)
+#define PDBGC_S(arg) fprintf(pdbgc, "PDBGC: %s\n", arg); fflush(pdbgc)
 #else
 #define PDBG_OPEN    
 #define PDBG_CLOSE   
@@ -91,6 +91,7 @@ static FILE * pdbgc = NULL;
 #include "sh_tiger.h"
 #include "sh_extern.h"
 #include "sh_calls.h"
+#include "sh_filter.h"
 #define SH_NEED_PWD_GRP 1
 #include "sh_static.h"
 
@@ -114,9 +115,10 @@ int sh_ext_popen (sh_tas_t * task)
   int    pipedes[2];
   FILE * outf = NULL;
   char * envp[1];
-  char * argp[1];
+  char * argp[2];
 
   char * errfile;
+  char errbuf[SH_ERRBUF_SIZE];
 
   static int some_error = 0;
 
@@ -125,24 +127,31 @@ int sh_ext_popen (sh_tas_t * task)
   char        pname[128];
   int         pfd = -1;
 #endif
-  
+
   SL_ENTER(_("sh_ext_popen"));
 
   /* Linux, HP-UX and FreeBSD will happily accept envp = argp = NULL
+   * (newer Linuxes (gcc 4.4.4) warn on argp == NULL, but accept it,
+   * as reported by T. Luettgert)
    * Solaris (and probably some other Unices) 
    *         needs a valid *envp[] with envp[0] = NULL;
    *         and similarly for argp
+   * OpenBSD finally needs non-null argp[0] ...
    */
+  argp[0] = task->command;
+  argp[1] = NULL;
   envp[0] = NULL;
-  argp[0] = NULL;
 
   /* 
    * --  check whether path is trustworthy
    */
+  status = sl_trustfile(task->command, NULL, NULL);
+#if 0
   if ((uid_t) -1 != task->trusted_users[0])
     {
       status = sl_trustfile(task->command, task->trusted_users, NULL);
     }
+#endif
 
   PDBG_OPEN;
   PDBG_D( (int) status);
@@ -177,12 +186,15 @@ int sh_ext_popen (sh_tas_t * task)
   /* 
    * --  check whether the checksum is correct; with linux emulate fdexec
    */
-#if !defined(__linux__) && !defined(SL_DEBUG)
+#if ( !defined(__linux__) || ( defined(__linux__) && defined(HAVE_PTHREAD)) ) && !defined(SL_DEBUG)
   if (task->checksum[0]  != '\0')
     {
+      char hashbuf[KEYBUF_SIZE];
       PDBG_S("checksum test");
       if (0 != sl_strcmp(task->checksum, 
-			 sh_tiger_hash (task->command, TIGER_FILE, 0))
+			 sh_tiger_hash (task->command, TIGER_FILE, TIGER_NOLIM,
+					hashbuf, sizeof(hashbuf))
+			 )
 	  )
 	{
 	  PDBG_S("checksum mismatch");
@@ -210,27 +222,29 @@ int sh_ext_popen (sh_tas_t * task)
       PDBG_S("pipe() failure");
       errnum = errno;
       sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_E_SUBGEN, 
-		      sh_error_message(errnum), _("pipe"));
+		      sh_error_message(errnum, errbuf, sizeof(errbuf)), _("pipe"));
       SL_RETURN ((-1), _("sh_ext_popen"));
     }
 
   PDBG(3);
-  
+
   /* 
-   * -- Fork 
+   * -- Flush streams and fork 
    */
+  fflush (NULL);
+
   task->pid = aud_fork(FIL__, __LINE__);
 
   if (task->pid == (pid_t) - 1) 
     {
       PDBG_S("fork() failure");
       /*@-usedef@*/
-      (void) close(pipedes[0]);
-      (void) close(pipedes[1]);
+      (void) sl_close_fd(FIL__, __LINE__, pipedes[0]);
+      (void) sl_close_fd(FIL__, __LINE__, pipedes[1]);
       /*@+usedef@*/
       errnum = errno;
       sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_E_SUBGEN, 
-		      sh_error_message(errnum), _("fork"));
+		      sh_error_message(errnum, errbuf, sizeof(errbuf)), _("fork"));
       SL_RETURN ((-1), _("sh_ext_popen"));
     }
   
@@ -243,16 +257,18 @@ int sh_ext_popen (sh_tas_t * task)
        */
       if (S_TRUE == task->fork_twice)
 	{
-	  task->pid = aud_fork(FIL__, __LINE__);
+	  task->pid = fork();
 
 	  if (task->pid == (pid_t) - 1) 
 	    {
-	      aud__exit (FIL__, __LINE__, EXIT_FAILURE);
+	      _exit (EXIT_FAILURE);
 	    }
 	}
 
       if (task->pid == (pid_t) 0)
 	{
+	  int val_return;
+
 	  PDBGC_OPEN;
 	  PDBGC(1);
 
@@ -261,28 +277,34 @@ int sh_ext_popen (sh_tas_t * task)
 	   */
 	  if (task->rw == 'w')
 	    {
-	      if (retry_aud_dup2(FIL__, __LINE__, 
-				 pipedes[STDIN_FILENO], STDIN_FILENO) < 0)
-		aud__exit(FIL__, __LINE__,EXIT_FAILURE);
+	      do {
+		val_return = dup2 (pipedes[STDIN_FILENO], STDIN_FILENO);
+	      } while (val_return < 0 && errno == EINTR);
+
+	      if (val_return < 0)
+		_exit(EXIT_FAILURE);
 	    }
 	  else
 	    {
-	      if (retry_aud_dup2(FIL__, __LINE__,
-				 pipedes[STDOUT_FILENO], STDOUT_FILENO) < 0)
-		aud__exit(FIL__, __LINE__,EXIT_FAILURE);
+	      do {
+		val_return = dup2 (pipedes[STDOUT_FILENO], STDOUT_FILENO);
+	      } while (val_return < 0 && errno == EINTR);
+
+	      if (val_return < 0)
+		_exit(EXIT_FAILURE);
 	    }
 	  PDBGC(2);
 	    
 	  
 	  /* close the pipe descriptors 
 	   */
-	  (void) close   (pipedes[STDIN_FILENO]);
-	  (void) close   (pipedes[STDOUT_FILENO]);
+	  (void) sl_close_fd   (FIL__, __LINE__, pipedes[STDIN_FILENO]);
+	  (void) sl_close_fd   (FIL__, __LINE__, pipedes[STDOUT_FILENO]);
 	  
 	  /* don't leak file descriptors
 	   */
 #if !defined(PDGBFILE)
-	  sh_unix_closeall (3, task->com_fd); /* in child process */
+	  sh_unix_closeall (3, task->com_fd, SL_TRUE); /* in child process */
 #endif
 
 	  /* drop root privileges, if possible && requested
@@ -295,12 +317,15 @@ int sh_ext_popen (sh_tas_t * task)
 	       */
 	      memset(skey, 0, sizeof(sh_key_t));
 
-	      (void) aud_setgid(FIL__, __LINE__,(gid_t) task->run_user_gid);
-	      (void) aud_setuid(FIL__, __LINE__,(uid_t) task->run_user_uid);
+	      if (setgid((gid_t) task->run_user_gid) != 0)
+		_exit(EXIT_FAILURE);
+	      if (setuid((uid_t) task->run_user_uid) != 0)
+		_exit(EXIT_FAILURE);
+
 	      /* make sure we cannot get root again
 	       */
-	      if (aud_setuid(FIL__, __LINE__,0) >= 0)
-		aud__exit(FIL__, __LINE__,EXIT_FAILURE);
+	      if (setuid(0) >= 0)
+		_exit(EXIT_FAILURE);
 	    }
 	  
 	  PDBGC(3);
@@ -319,8 +344,10 @@ int sh_ext_popen (sh_tas_t * task)
 	  else
 	    {
 	      PDBGC_S("r");
-	      (void) retry_aud_dup2 (FIL__, __LINE__, 
-				     STDOUT_FILENO, STDERR_FILENO);
+	      do {
+		val_return = dup2 (STDOUT_FILENO, STDERR_FILENO);
+	      } while (val_return < 0 && errno == EINTR);
+
 	      (void) fcntl  (STDIN_FILENO, F_SETFD, FD_CLOEXEC);
 	      /*
 	      freopen(_("/dev/null"), "r+", stdin);
@@ -334,47 +361,103 @@ int sh_ext_popen (sh_tas_t * task)
 	  /* 
 	   * --  emulate an fdexec with checksum testing
 	   */
+
+#if !defined(HAVE_PTHREAD)
 	  if (task->checksum[0]  != '\0')
+#endif
 	    {
 	      PDBGC_S("fexecve");
 	      if (task->com_fd != (-1))
 		{
-		  pfd = retry_aud_dup(FIL__, __LINE__, task->com_fd);
+		  do {
+		    val_return = dup (task->com_fd);
+		  } while (val_return < 0 && errno == EINTR);
+		  pfd = val_return;
 		  if (pfd < 0)
 		    {
-		      PDBGC_S("fexecve: dup2 failed");
-		      aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+		      PDBGC_S("fexecve: dup failed");
+		      _exit(EXIT_FAILURE);
 		    }
 		}
+#if !defined(HAVE_PTHREAD)
 	      else
 		{
+		  char hashbuf[KEYBUF_SIZE];
+
 		  fd = 
-		    sl_open_read(task->command, 
+		    sl_open_read(FIL__, __LINE__, task->command, 
 				 task->privileged==0 ? SL_NOPRIV : SL_YESPRIV);
-		  tiger_fd = fd;
+
 		  if (0 != sl_strcmp(task->checksum, 
 				     sh_tiger_hash (task->command, 
-						    TIGER_FD, 0)))
+						    fd, TIGER_NOLIM, hashbuf, sizeof(hashbuf))))
 		    {
 		      PDBGC_S("fexecve: checksum mismatch");
-		      aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+		      sl_close(fd);
+		      _exit(EXIT_FAILURE);
 		    }
+
 		  pfd = get_the_fd(fd);
+
+		  do {
+		    val_return = dup (pfd);
+		  } while (val_return < 0 && errno == EINTR);
+		  pfd = val_return;
+
+		  sl_close(fd);
+		  fd = -1;
+
+		  if (pfd < 0)
+		    {
+		      PDBGC_S("fexecve: dup (2) failed");
+		      _exit(EXIT_FAILURE);
+		    }
 		}
+#endif
               
 	      PDBGC(5);
-	      sprintf(pname, _("/proc/self/fd/%d"),      /* known to fit  */
-			   pfd);
-              if (access(pname, R_OK|X_OK) == 0) 
+	      /* Cannot use sprintf because of deadlock in malloc/free */
+	      {
+		static char digit[] = "0123456789";
+		char str0[128];
+		char str1[128];
+		int ival = pfd;
+		int n = 0;
+		int m = 0;
+
+		if (ival < 0) ival = -ival;
+		do {
+		  str0[n] = digit[ival % 10];
+		  ++n;
+		  ival /= 10;
+		} while (ival);
+
+		if (pfd < 0)
+		  {
+		    str0[n] = '-';
+		    ++n;
+		  }
+		str0[n] = '\0';
+		str1[n] = '\0';
+		while (n > 0)
+		  {
+		    str1[m] = str0[n-1];
+		    ++m; --n;
+		  }
+		sl_strlcpy(pname, _("/proc/self/fd/"), sizeof(pname));
+		sl_strlcat(pname, str1, sizeof(pname));
+	      }
+              if (access(pname, R_OK|X_OK) == 0) /* flawfinder: ignore */
 		{
 		  PDBGC(6);
 		  PDBGC_CLOSE;
 		  fcntl  (pfd, F_SETFD, FD_CLOEXEC);
-		  retry_aud_execve (FIL__, __LINE__, 
-				    pname, 
-				    (task->argc == 0) ? NULL : task->argv, 
-				    (task->envc == 0) ? NULL : task->envv
-				    );
+		  do {
+		    val_return = execve (pname, 
+					 (task->argc == 0) ? argp : task->argv, 
+					 (task->envc == 0) ? NULL : task->envv
+					 );
+		  } while (val_return < 0 && errno == EINTR);
 		  
 		  errnum = errno;
 		  PDBGC_OPEN;
@@ -384,7 +467,7 @@ int sh_ext_popen (sh_tas_t * task)
 		  PDBGC_CLOSE;
 		  /* failed 
 		   */
-		  aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+		  _exit((errnum == 0) ? (EXIT_SUCCESS) : (EXIT_FAILURE));
               }
 	      PDBGC_S("fexecve: not working");
 	      /* 
@@ -393,7 +476,7 @@ int sh_ext_popen (sh_tas_t * task)
 	      if (fd != -1)
 		sl_close(fd);
 	      else if (pfd != -1)
-		close(fd);
+		sl_close_fd(FIL__, __LINE__, pfd);
 	    }
 #endif
 
@@ -401,15 +484,16 @@ int sh_ext_popen (sh_tas_t * task)
 	  /* 
 	   * --  execute path if executable
 	   */
-	  if (0 == access(task->command, R_OK|X_OK))
+	  if (0 == access(task->command, R_OK|X_OK)) /* flawfinder: ignore */
 	    {
 	      PDBGC(5);
 	      PDBGC_CLOSE;
-	      (void) retry_aud_execve (FIL__, __LINE__, 
-				       task->command, 
-				       (task->argc == 0) ? argp : task->argv, 
-				       (task->envc == 0) ? envp : task->envv
-				       );
+	      do {
+		val_return = execve (task->command, 
+				     (task->argc == 0) ? argp : task->argv, 
+				     (task->envc == 0) ? envp : task->envv
+				     );
+	      } while (val_return < 0 && errno == EINTR);
 	    }
 	  errnum = errno;
 	  PDBGC_OPEN;
@@ -419,14 +503,14 @@ int sh_ext_popen (sh_tas_t * task)
 	  PDBGC_CLOSE;
 	  /* failed 
 	   */
-	  aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+	  _exit((errnum == 0) ? (EXIT_SUCCESS) : (EXIT_FAILURE));
 	}
       /* 
        * if we have forked twice, this is parent::detached_subprocess
        */
       if (S_TRUE == task->fork_twice)
 	{
-	  aud__exit (FIL__, __LINE__, 0);
+	  _exit (0);
 	}
     }
 
@@ -446,7 +530,7 @@ int sh_ext_popen (sh_tas_t * task)
   if (task->rw == 'w')
     {
       PDBG_S("is w");
-      (void) close (pipedes[STDIN_FILENO]);
+      (void) sl_close_fd (FIL__, __LINE__, pipedes[STDIN_FILENO]);
       (void) retry_fcntl (FIL__, __LINE__, pipedes[STDOUT_FILENO], 
 			  F_SETFD, FD_CLOEXEC);
       outf = fdopen (pipedes[STDOUT_FILENO], "w");
@@ -454,7 +538,7 @@ int sh_ext_popen (sh_tas_t * task)
   else
     {
       PDBG_S("is r");
-      (void) close (pipedes[STDOUT_FILENO]);
+      (void) sl_close_fd (FIL__, __LINE__, pipedes[STDOUT_FILENO]);
       (void) retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], 
 			  F_SETFD, FD_CLOEXEC);
       outf = fdopen (pipedes[STDIN_FILENO], "r");
@@ -476,8 +560,8 @@ int sh_ext_popen (sh_tas_t * task)
       SH_FREE(tmp);
 
       (void) aud_kill (FIL__, __LINE__, task->pid, SIGKILL);
-      (void) close (pipedes[STDOUT_FILENO]);
-      (void) close (pipedes[STDIN_FILENO]);
+      (void) sl_close_fd (FIL__, __LINE__, pipedes[STDOUT_FILENO]);
+      (void) sl_close_fd (FIL__, __LINE__, pipedes[STDIN_FILENO]);
       (void) waitpid (task->pid, NULL, 0);
       task->pid = 0;
 
@@ -491,7 +575,7 @@ int sh_ext_popen (sh_tas_t * task)
 
   PDBG_D(task->pipeFD);
 
-  task->pipeTI = sl_make_ticket(task->pipeFD, _("pipe"));
+  task->pipeTI = sl_make_ticket(FIL__, __LINE__, task->pipeFD, _("pipe"), outf);
 
   flags = (int) retry_fcntl (FIL__, __LINE__, task->pipeFD, F_GETFL, 0);
   if (flags != (-1))
@@ -517,12 +601,17 @@ int sh_ext_pclose (sh_tas_t * task)
   pid_t retval;
   char  infomsg[256];
 
+#ifdef WCONTINUED
+      int wflags = WNOHANG|WUNTRACED|WCONTINUED;
+#else
+      int wflags = WNOHANG|WUNTRACED;
+#endif
+
   SL_ENTER(_("sh_ext_pclose"));
 
-  PDBGC_OPEN;
+  PDBG_OPEN;
   PDBG_S(" -> pclose");
   (void) fflush(task->pipe);
-  (void) fclose(task->pipe);
   if (!SL_ISERROR(task->pipeTI))
     (void) sl_close(task->pipeTI);
 
@@ -535,41 +624,45 @@ int sh_ext_pclose (sh_tas_t * task)
       infomsg[0] = '\0';
 
     nochmal:
-      retval = waitpid(task->pid, &(task->exit_status), WNOHANG|WUNTRACED);
+      retval = waitpid(task->pid, &(task->exit_status), wflags);
       /*@-bufferoverflowhigh@*/
       if (task->pid == retval)
 	{
+#ifndef USE_UNO
 	  if (WIFEXITED(task->exit_status) != 0)
 	    {
 	      task->exit_status = WEXITSTATUS(task->exit_status);
 	      if ((flag_err_debug == SL_TRUE) || (task->exit_status != 0))
-		sprintf(infomsg,                         /* known to fit  */
-			_("Subprocess exited normally with status %d"),
-			task->exit_status);
+		sl_snprintf(infomsg, sizeof(infomsg),
+			    _("Subprocess exited normally with status %d"),
+			    task->exit_status);
 	    }
 	  else if (WIFSIGNALED(task->exit_status) != 0)
 	    {
-	      sprintf(infomsg,                           /* known to fit  */
-		      _("Subprocess terminated by signal %d"),
-		      WTERMSIG(task->exit_status));
+	      sl_snprintf(infomsg, sizeof(infomsg),
+			  _("Subprocess terminated by signal %d"),
+			  WTERMSIG(task->exit_status));
 	      task->exit_status = EXIT_FAILURE;
 	    }
 	  else if (WIFSTOPPED(task->exit_status) != 0)
 	    {
-	      sprintf(infomsg,                           /* known to fit  */
-		      _("Subprocess stopped by signal %d, killing"),
-		      WSTOPSIG(task->exit_status));
+	      sl_snprintf(infomsg, sizeof(infomsg),
+			  _("Subprocess stopped by signal %d, killing"),
+			  WSTOPSIG(task->exit_status));
 	      task->exit_status = EXIT_FAILURE;
 	      (void) aud_kill (FIL__, __LINE__, task->pid, 9);
 	      (void) retry_msleep (0, 30);
-	      (void) waitpid (task->pid, NULL, WNOHANG|WUNTRACED);
+	      (void) waitpid (task->pid, NULL, wflags);
 	    }
 	  else
 	    {
-	      sprintf(infomsg,                           /* known to fit  */
-		      _("Subprocess exit status unknown"));
+	      sl_snprintf(infomsg, sizeof(infomsg),
+			  _("Subprocess exit status unknown"));
 	      task->exit_status = EXIT_FAILURE;
 	    }
+#else
+	  task->exit_status = EXIT_FAILURE;
+#endif 
 	}
       else if (0 == retval)
 	{
@@ -580,15 +673,15 @@ int sh_ext_pclose (sh_tas_t * task)
 	      goto nochmal;
 	    }
 	  (void) aud_kill (FIL__, __LINE__, task->pid, 9);
-	  sprintf(infomsg,                               /* known to fit  */
-		  _("Subprocess not yet exited, killing"));
+	  sl_snprintf(infomsg, sizeof(infomsg),
+		      _("Subprocess not yet exited, killing"));
 	  task->exit_status = EXIT_FAILURE;
 	  (void) waitpid (task->pid, NULL, 0);
 	}
       else
 	{
-	  sprintf(infomsg,                               /* known to fit  */
-		  _("Waitpid returned error %d\n"), errno);
+	  sl_snprintf(infomsg, sizeof(infomsg),
+		      _("Waitpid returned error %d\n"), errno);
 	  task->exit_status = EXIT_FAILURE;
 	}
       /*@+bufferoverflowhigh@*/
@@ -633,7 +726,9 @@ void sh_ext_tas_init (sh_tas_t * tas)
     {
       tas->argv[i]          = NULL;
       tas->envv[i]          = NULL;
+#if 0
       tas->trusted_users[i] = (uid_t) -1;
+#endif
     }
 
   tas->run_user_uid     = (uid_t) getuid();
@@ -645,7 +740,7 @@ void sh_ext_tas_init (sh_tas_t * tas)
 }
 
 
-int sh_ext_tas_add_envv(sh_tas_t * tas, char * key, char * val)
+int sh_ext_tas_add_envv(sh_tas_t * tas, const char * key, const char * val)
 {
   size_t sk = 0, sv = 0;
   int    si;
@@ -662,6 +757,10 @@ int sh_ext_tas_add_envv(sh_tas_t * tas, char * key, char * val)
   if (val != NULL)
     sv = strlen(val) + 1;
 
+  if (!sl_ok_adds(sk, sv))
+    {
+      SL_RETURN (-1, _("sh_ext_tas_add_envv"));
+    }
   si = tas->envc;
   tas->envv[si] = SH_ALLOC(sk + sv);
 
@@ -696,7 +795,7 @@ int sh_ext_tas_rm_argv(sh_tas_t * tas)
   SL_RETURN ((tas->argc), _("sh_ext_tas_rm_argv"));
 }
 
-int sh_ext_tas_add_argv(sh_tas_t * tas, char * val)
+int sh_ext_tas_add_argv(sh_tas_t * tas, const char * val)
 {
   size_t sv = 0;
   int    si;
@@ -721,7 +820,7 @@ int sh_ext_tas_add_argv(sh_tas_t * tas, char * val)
   SL_RETURN ((tas->argc), _("sh_ext_tas_add_argv"));
 }
 
-void sh_ext_tas_command(sh_tas_t * tas, char * command)
+void sh_ext_tas_command(sh_tas_t * tas, const char * command)
 {
   size_t len = sl_strlen(command);
   tas->command = SH_ALLOC(len+1);
@@ -750,6 +849,174 @@ void sh_ext_tas_free(sh_tas_t * tas)
   return;
 }
 
+static void task_init (sh_tas_t * task)
+{
+  sh_ext_tas_init(task);
+
+  (void) sh_ext_tas_add_envv (task, _("SHELL"), 
+			      _("/bin/sh")); 
+  (void) sh_ext_tas_add_envv (task, _("PATH"),  
+			      _("/sbin:/bin:/usr/sbin:/usr/bin:/usr/ucb")); 
+  (void) sh_ext_tas_add_envv (task, _("IFS"), " \n\t"); 
+
+  if (sh.timezone != NULL)
+    {
+      (void) sh_ext_tas_add_envv(task,  "TZ", sh.timezone);
+    }
+  return;
+}
+
+int sh_ext_popen_init (sh_tas_t * task, char * command, char * argv0, ...)
+{
+  va_list vl;
+  int status;
+
+  task_init (task);
+  
+  if (!argv0)
+    {
+      sh_ext_tas_command(task,  _("/bin/sh"));
+
+      (void) sh_ext_tas_add_argv(task,  _("/bin/sh"));
+      (void) sh_ext_tas_add_argv(task,  _("-c"));
+      (void) sh_ext_tas_add_argv(task,  command);
+    }
+  else
+    {
+      char * s;
+
+      sh_ext_tas_command(task,  command);
+
+      (void) sh_ext_tas_add_argv(task, argv0);
+
+      va_start (vl, argv0);
+      s = va_arg (vl, char * );
+      while (s != NULL)
+	{
+	  (void) sh_ext_tas_add_argv(task, s);
+	  s = va_arg (vl, char * );
+	}
+      va_end (vl);
+
+    }
+  task->rw = 'r';
+  task->fork_twice = S_FALSE;
+
+  status = sh_ext_popen(task);
+
+  return status;
+}
+
+/* Execute a system command */
+
+int sh_ext_system (char * command, char * argv0, ...)
+{
+  sh_tas_t task;
+  int    status;
+  va_list vl;
+  char * s;
+
+  SL_ENTER(_("sh_ext_system"));
+
+  task_init (&task);
+  
+  sh_ext_tas_command(&task,  command);
+
+  (void) sh_ext_tas_add_argv(&task, argv0);
+  
+  va_start (vl, argv0);
+  s = va_arg (vl, char * );
+  while (s != NULL)
+    {
+      (void) sh_ext_tas_add_argv(&task, s);
+      s = va_arg (vl, char * );
+    }
+  va_end (vl);
+
+  task.rw = 'r';
+  task.fork_twice = S_FALSE;
+
+  status = sh_ext_popen(&task);
+
+  if (status != 0)
+    {
+      sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, status, MSG_E_SUBGEN, 
+                      _("Could not execute command"), _("sh_ext_system"));
+      SL_RETURN ((-1), _("sh_ext_system"));
+    }
+
+  /* close pipe and return exit status
+   */
+  (void) sh_ext_pclose(&task);
+  sh_ext_tas_free (&task);
+  SL_RETURN ((status), _("sh_ext_system"));
+}
+
+/* Execute command, return first line of output
+ * ifconfig | grep -1 lo | tail -n 1 | sed s/.*inet addr:\([0-9.]*\)\(.*\)/\1/
+ */
+char * sh_ext_popen_str (char * command)
+{
+  sh_tas_t task;
+  struct  sigaction  new_act;
+  struct  sigaction  old_act;
+  char * out = NULL;
+  int    status;
+
+  SL_ENTER(_("sh_ext_popen_str"));
+
+  status = sh_ext_popen_init (&task, command, NULL, NULL);
+
+  if (status != 0)
+    {
+      sh_error_handle(SH_ERR_ALL, FIL__, __LINE__, status, MSG_E_SUBGEN, 
+                      _("Could not open pipe"), _("sh_ext_popen_str"));
+      SL_RETURN ((NULL), _("sh_ext_popen_str"));
+    }
+
+   /* ignore SIGPIPE (instead get EPIPE if connection is closed)
+    */
+   new_act.sa_handler = SIG_IGN;
+   (void) retry_sigaction (FIL__, __LINE__, SIGPIPE, &new_act, &old_act);
+   
+  /* read from the open pipe
+   */
+  if (task.pipe != NULL)
+    {
+      int try = 1200; /* 1000 * 0.1 = 120 sec */
+      sh_string * s = sh_string_new(0);
+      do {
+	sh_string_read(s, task.pipe, 0);
+	if (sh_string_len(s) == 0)
+	  {
+	    --try; retry_msleep(0, 100);
+	  }
+      } while (sh_string_len(s) == 0 && try != 0);
+
+      if (sh_string_len(s) == 0)
+	{
+	  sh_error_handle(SH_ERR_ALL, FIL__, __LINE__, status, MSG_E_SUBGEN, 
+			  _("No output from command"), _("sh_ext_popen_str"));
+	}
+
+      out = sh_util_strdup(sh_string_str(s));
+      sh_string_destroy(&s);
+    }
+
+  /* restore old signal handler
+   */
+  (void) retry_sigaction (FIL__, __LINE__, SIGPIPE, &old_act, NULL);
+
+  /* close pipe and return exit status
+   */
+  (void) sh_ext_pclose(&task);
+  sh_ext_tas_free (&task);
+  SL_RETURN ((out), _("sh_ext_popen_str"));
+}
+
+
+
+
 /* ---------------  EXTERN STUFF ------------------- */
 
 #if defined(WITH_EXTERNAL)
@@ -758,12 +1025,8 @@ typedef struct _sh_com_t
 {
   char     type[4];
 
-  int      for_c;
-  char   * for_v[32];
-  int      fand_c;
-  char   * fand_v[32];
-  int      fnot_c;
-  char   * fnot_v[32];
+  sh_filter_type * filter;
+
   time_t   deadtime;
   time_t   last_run;
 
@@ -773,6 +1036,7 @@ typedef struct _sh_com_t
 
 } sh_com_t;
 
+
 static
 void set3 (char * pos, char c1, char c2, char c3)
 {
@@ -790,7 +1054,6 @@ void set3 (char * pos, char c1, char c2, char c3)
 static
 sh_com_t * command_init(void)
 {
-  int         i;
   uid_t       ff_euid;
   sh_com_t  * ext_com = NULL;
 
@@ -800,33 +1063,27 @@ sh_com_t * command_init(void)
 
   if (!ext_com)
     {
-      SL_RETURN( NULL, ("command_init"));
+      SL_RETURN( NULL, _("command_init"));
     }
 
   sh_ext_tas_init (&(ext_com->tas));
 
   (void) sl_get_euid(&ff_euid);
+#if 0
   ext_com->tas.trusted_users[0] = (uid_t) 0;
   ext_com->tas.trusted_users[1] = (uid_t) (ff_euid);
+#endif
 
   /* ------------------------------------------------- */
 
   set3(ext_com->type, 'l', 'o', 'g');
-  ext_com->for_c        = 0;
-  ext_com->fand_c       = 0;
-  ext_com->fnot_c       = 0;
+  ext_com->filter       = NULL;
   ext_com->deadtime     = 0;
   ext_com->last_run     = 0;
 
-  for (i = 0; i < 32; ++i)
-    {
-      ext_com->for_v[i]         = NULL;
-      ext_com->fand_v[i]        = NULL;
-      ext_com->fnot_v[i]        = NULL;
-    }
   ext_com->next             = NULL;
 
-  SL_RETURN( ext_com, ("command_init"));
+  SL_RETURN( ext_com, _("command_init"));
 }
 
 /* the list of external commands
@@ -839,8 +1096,10 @@ static sh_com_t * ext_coms   = NULL;
 static int ext_failed = -1;
 
 static
-int sh_ext_add_envv(char * key, char * val)
+int sh_ext_add_envv(const char * key, const char * val)
 {
+  int retval; 
+
   SL_ENTER(_("sh_ext_add_envv"));
 
   if (ext_coms == NULL || ext_failed == (-1) || 
@@ -850,15 +1109,18 @@ int sh_ext_add_envv(char * key, char * val)
       SL_RETURN (-1, _("sh_ext_add_envv"));
     }
 
-  (void) sh_ext_tas_add_envv(&(ext_coms->tas), key, val);
+  retval = sh_ext_tas_add_envv(&(ext_coms->tas), key, val);
 
-  SL_RETURN (0, _("sh_ext_add_envv"));
+  if (retval >= 0) 
+    retval = 0;
+
+  SL_RETURN (retval, _("sh_ext_add_envv"));
 }
 
 
 
 static 
-int sh_ext_init(char * command)
+int sh_ext_init(const char * command)
 {
   sh_com_t * retval;
   size_t     size;
@@ -893,9 +1155,13 @@ int sh_ext_init(char * command)
 }
 
 static
-int sh_ext_uid (char * user, /*@out@*/uid_t * uid, /*@out@*/gid_t * gid)
+int sh_ext_uid (const char * user, /*@out@*/uid_t * uid, /*@out@*/gid_t * gid)
 {
-  struct passwd * tempres;
+  struct passwd *  tempres;
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+  struct passwd    pwd;
+  char           * buffer = SH_ALLOC(SH_PWBUF_SIZE);
+#endif
 
   SL_ENTER(_("sh_ext_uid"));
 
@@ -903,27 +1169,43 @@ int sh_ext_uid (char * user, /*@out@*/uid_t * uid, /*@out@*/gid_t * gid)
 
   if (user == NULL)
     {
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      SH_FREE(buffer);
+#endif
       SL_RETURN (-1, _("sh_ext_uid"));
     }
+
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+  sh_getpwnam_r(user, &pwd, buffer, SH_PWBUF_SIZE, &tempres);
+#else
   tempres = sh_getpwnam(user);
+#endif
 
   if (NULL != tempres) 
     {
       *uid = tempres->pw_uid;  
       *gid = tempres->pw_gid;
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      SH_FREE(buffer);
+#endif
       SL_RETURN (0, _("sh_ext_uid"));
     } 
 
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+  SH_FREE(buffer);
+#endif
   SL_RETURN (-1, _("sh_ext_uid"));
 }
 
 
 static
-int sh_ext_add (char * argstring, int * ntok, char * stok[])
+int sh_ext_add (const char * argstring, int * ntok, char * stok[])
 {
   int    i = 0;
   size_t s;
   char * p;
+  char * new;
+  size_t len;
 
   SL_ENTER(_("sh_ext_add"));
 
@@ -932,12 +1214,25 @@ int sh_ext_add (char * argstring, int * ntok, char * stok[])
       SL_RETURN((-1), _("sh_ext_add")); 
     }
 
+  len = strlen(argstring) + 1;
+  new = SH_ALLOC(len);
+  sl_strlcpy(new, argstring, len); 
+
   do
     {
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+      char * saveptr;
       if (i == 0)
-	p = strtok (argstring, ", \t");
+	p = strtok_r (new, ", \t", &saveptr);
+      else
+	p = strtok_r (NULL, ", \t", &saveptr);
+#else
+      if (i == 0)
+	p = strtok (new, ", \t");
       else
 	p = strtok (NULL, ", \t");
+#endif
+
       if (p == NULL)
 	break;
 
@@ -954,6 +1249,7 @@ int sh_ext_add (char * argstring, int * ntok, char * stok[])
   while (p != NULL);
 
   *ntok = i;
+  SH_FREE(new);
 
   SL_RETURN (0, _("sh_ext_add"));
 }
@@ -968,7 +1264,7 @@ int sh_ext_add (char * argstring, int * ntok, char * stok[])
 /* 
  * -- start a new external command, and add it to the list
  */ 
-int sh_ext_setcommand(char * cmd)
+int sh_ext_setcommand(const char * cmd)
 {
   int i;
 
@@ -986,7 +1282,6 @@ int sh_ext_setcommand(char * cmd)
  */
 int sh_ext_cleanup(void)
 {
-  int i;
   sh_com_t * retval;
 
   SL_ENTER(_("sh_ext_cleanup"));
@@ -998,12 +1293,8 @@ int sh_ext_cleanup(void)
 
       sh_ext_tas_free (&(retval->tas));
 
-      for (i = 0; i < 32; ++i)
-	{
-	  if (NULL != retval->for_v[i])  SH_FREE(retval->for_v[i]);
-	  if (NULL != retval->fand_v[i]) SH_FREE(retval->fand_v[i]);
-	  if (NULL != retval->fnot_v[i]) SH_FREE(retval->fnot_v[i]);
-	}
+      if (retval->filter)
+	sh_filter_free (retval->filter);
 
       SH_FREE(retval);
 
@@ -1012,40 +1303,58 @@ int sh_ext_cleanup(void)
   SL_RETURN (0, _("sh_ext_cleanup"));
 }
 
+/*
+ * -- explicitely close a command
+ */
+int sh_ext_close_command (const char * str)
+{
+  (void) str;
+  if (ext_coms == NULL || ext_failed == (-1))
+    return (-1);
+  ext_failed = (-1);
+  return 0;
+}
+
 /*
  * -- add keywords to the OR filter
  */
-int sh_ext_add_or (char * str)
+int sh_ext_add_or (const char * str)
 {
   if (ext_coms == NULL || ext_failed == (-1))
     return (-1);
-  return (sh_ext_add (str, &(ext_coms->for_c), ext_coms->for_v));
+  if (ext_coms->filter == NULL)
+    ext_coms->filter = sh_filter_alloc();
+  return (sh_filter_add(str, ext_coms->filter, SH_FILT_OR));
 }
 
 /*
  * -- add keywords to the AND filter
  */
-int sh_ext_add_and (char * str)
+int sh_ext_add_and (const char * str)
 {
   if (ext_coms == NULL || ext_failed == (-1))
     return (-1);
-  return (sh_ext_add (str, &(ext_coms->fand_c), ext_coms->fand_v));
+  if (ext_coms->filter == NULL)
+    ext_coms->filter = sh_filter_alloc();
+  return (sh_filter_add(str, ext_coms->filter, SH_FILT_AND));
 }
 
 /*
  * -- add keywords to the NOT filter
  */
-int sh_ext_add_not (char * str)
+int sh_ext_add_not (const char * str)
 {
   if (ext_coms == NULL || ext_failed == (-1))
     return (-1);
-  return (sh_ext_add (str, &(ext_coms->fnot_c), ext_coms->fnot_v));
+  if (ext_coms->filter == NULL)
+    ext_coms->filter = sh_filter_alloc();
+  return (sh_filter_add(str, ext_coms->filter, SH_FILT_NOT));
 }
 
 /*
  * -- add keywords to the CL argument list
  */
-int sh_ext_add_argv (char * str)
+int sh_ext_add_argv (const char * str)
 {
   if (ext_coms == NULL || ext_failed == (-1))
     return (-1);
@@ -1055,13 +1364,11 @@ int sh_ext_add_argv (char * str)
 /*
  * -- add a path to the environment
  */
-int sh_ext_add_default (char * dummy)
+int sh_ext_add_default (const char * dummy)
 {
-  /* while this assignment looks ridiculous, it is here to avoid
-   * an 'unused parameter' warning
-   */
-  char * p = (dummy == NULL ? dummy : NULL);
+  char * p = NULL;
   int    i;
+  char   dir[SH_PATHBUF];
 
   SL_ENTER(_("sh_ext_add_default"));
   if (dummy[0] == 'n' ||  dummy[0] == 'N' ||
@@ -1069,11 +1376,13 @@ int sh_ext_add_default (char * dummy)
     {
       SL_RETURN(0, _("sh_ext_add_default"));
     }
-  p = sh_unix_getUIDdir (SH_ERR_ERR, (uid_t) ext_coms->tas.run_user_uid);
+  p = sh_unix_getUIDdir (SH_ERR_ERR, (uid_t) ext_coms->tas.run_user_uid, 
+			 dir, sizeof(dir));
   if (p)
     (void) sh_ext_add_envv (_("HOME"), p);
   (void) sh_ext_add_envv (_("SHELL"), _("/bin/sh")); 
-  (void) sh_ext_add_envv (_("PATH"),  _("/sbin:/usr/sbin:/bin:/usr/bin")); 
+  (void) sh_ext_add_envv (_("PATH"),  _("/sbin:/bin:/usr/sbin:/usr/bin")); 
+  (void) sh_ext_add_envv (_("IFS"), " \n\t"); 
   i = (p == NULL ? (-1) :  0);
   SL_RETURN(i, _("sh_ext_add_default"));
 }
@@ -1081,9 +1390,10 @@ int sh_ext_add_default (char * dummy)
 /*
  * -- add an environment variable
  */
-int sh_ext_add_environ (char * str)
+int sh_ext_add_environ (const char * str)
 {
   int i;
+
   SL_ENTER(_("sh_ext_add_environ"));
   i = sh_ext_add_envv (NULL, str);
   SL_RETURN(i, _("sh_ext_add_environ"));
@@ -1092,7 +1402,7 @@ int sh_ext_add_environ (char * str)
 /*
  * -- set deadtime
  */
-int sh_ext_deadtime (char * str)
+int sh_ext_deadtime (const char * str)
 {
   long    deadtime = 0;
   char  * tail     = NULL;
@@ -1101,22 +1411,22 @@ int sh_ext_deadtime (char * str)
 
   if (ext_coms == NULL || ext_failed == (-1) || str == NULL)
     {
-      SL_RETURN (-1, ("sh_ext_deadtime"));
+      SL_RETURN (-1, _("sh_ext_deadtime"));
     }
   deadtime = strtol(str, &tail, 10);
   if (tail == str || deadtime < 0 || deadtime == LONG_MAX)
     {
-      SL_RETURN (-1, ("sh_ext_deadtime"));
+      SL_RETURN (-1, _("sh_ext_deadtime"));
     }
   
   ext_coms->deadtime = (time_t) deadtime;  
-  SL_RETURN (0, ("sh_ext_deadtime"));  
+  SL_RETURN (0, _("sh_ext_deadtime"));  
 }
 
 /*
  * -- define type
  */
-int sh_ext_type (char * str)
+int sh_ext_type (const char * str)
 {
   SL_ENTER(_("sh_ext_type"));
 
@@ -1151,7 +1461,7 @@ int sh_ext_type (char * str)
 /*
  * -- define checksum
  */
-int sh_ext_checksum (char * str)
+int sh_ext_checksum (const char * str)
 {
   SL_ENTER(_("sh_ext_checksum"));
   if (ext_coms == NULL || ext_failed == (-1) || str == NULL)
@@ -1172,7 +1482,7 @@ int sh_ext_checksum (char * str)
 /*
  * -- choose privileges
  */
-int sh_ext_priv (char * c)
+int sh_ext_priv (const char * c)
 {
 
   uid_t me_uid;
@@ -1199,71 +1509,36 @@ int sh_ext_priv (char * c)
  */
 static int sh_ext_filter (char * message, sh_com_t * task)
 {
-  int i;
-  int j = 0;
   time_t now_time;
 
   SL_ENTER(_("sh_ext_filter"));
 
-  /* Presence of any of these keywords prevents execution.
-   */
-  if (task->fnot_c > 0)
+  if (task->filter)
     {
-      for (i = 0; i < task->fnot_c; ++i)
-	{
-	  if (NULL != sl_strstr(message, task->fnot_v[i]))
-	    {
-	      SL_RETURN ((-1), _("sh_ext_filter"));
-	    }
-	}
-    }
-
-  /* Presence of all of these keywords is required for execution.
-   */
-  if (task->fand_c > 0)
-    {
-      j = 0;
-
-      for (i = 0; i < task->fand_c; ++i)
-	if (NULL != sl_strstr(message, task->fand_v[i]))
-	  ++j;
-
-      if (j != task->fand_c)
+      if (0 != sh_filter_filter (message, task->filter))
 	{
 	  SL_RETURN ((-1), _("sh_ext_filter"));
 	}
-
     }
 
-  /* Presence of at least one of these keywords is required for execution.
-   */
-  if (task->for_c > 0)
+  /* Filter passed, check deadtime */
+
+  if (task->deadtime != (time_t) 0)
     {
-      for (i = 0; i < task->for_c; ++i)
+      now_time = time (NULL);
+      
+      if (task->last_run == (time_t) 0)
 	{
-	  if (NULL != sl_strstr(message, task->for_v[i]))
-	    {
-	      if (task->deadtime != (time_t) 0)   /* deadtime */
-		{
-		  now_time = time (NULL);
-		  
-		  if (task->last_run == (time_t) 0)
-		    {
-		      task->last_run = now_time;
-		    }
-		  else if ((time_t)(now_time-task->last_run) < task->deadtime)
-		    {
-		      SL_RETURN ((-1), _("sh_ext_filter"));
-		    }
-		  else
-		    {
-		      task->last_run = now_time;
-		    }
-		}
-	      SL_RETURN ((0), _("sh_ext_filter"));
-	    }
+	  task->last_run = now_time;
+	}
+      else if ((time_t)(now_time-task->last_run) < task->deadtime)
+	{
+	  SL_RETURN ((-1), _("sh_ext_filter"));
+	}
+      else
+	{
+	  task->last_run = now_time;
 	}
-      SL_RETURN ((-1), _("sh_ext_filter"));
     }
 
   SL_RETURN ((0), _("sh_ext_filter"));
@@ -1281,6 +1556,7 @@ int sh_ext_execute (char t1, char t2, char t3, /*@null@*/char * message,
   sh_com_t * listval = ext_coms;
   int        status = 0;
   char     * tmp;
+  char errbuf[SH_ERRBUF_SIZE];
 
   static  int some_error = 0;
 
@@ -1309,6 +1585,7 @@ int sh_ext_execute (char t1, char t2, char t3, /*@null@*/char * message,
 
   while (listval != NULL)
     {
+      PDBG_OPEN;
       PDBG(-2);
       if (t1 == listval->type[0] &&
 	  t2 == listval->type[1] &&
@@ -1320,11 +1597,12 @@ int sh_ext_execute (char t1, char t2, char t3, /*@null@*/char * message,
 	  if (0 != (caperr = sl_get_cap_sub()))
 	    {
 	      sh_error_handle((-1), FIL__, __LINE__, caperr, MSG_E_SUBGEN,
-			      sh_error_message (caperr), 
+			      sh_error_message (caperr, errbuf, sizeof(errbuf)), 
 			      _("sl_get_cap_sub"));
 	    }
 	  if (0 == sh_ext_popen (&(listval->tas)))
 	    {
+	      PDBG_OPEN;
 	      PDBG(-4);
 	      if (NULL != listval->tas.pipe && listval->tas.rw == 'w')
 		{
@@ -1392,18 +1670,20 @@ int sh_ext_execute (char t1, char t2, char t3, /*@null@*/char * message,
 	    }
 	  else
 	    {
+	      PDBG_OPEN;
 	      PDBG_S("0 != sh_ext_popen()");
 	    }
 	  if (0 != (caperr = sl_drop_cap_sub()))
 	    {
 	      sh_error_handle((-1), FIL__, __LINE__, caperr, MSG_E_SUBGEN,
-			      sh_error_message (caperr), 
+			      sh_error_message (caperr, errbuf, sizeof(errbuf)), 
 			      _("sl_drop_cap_sub"));
 	    }
 
 	}
       listval = listval->next;
     }
+  PDBG_OPEN;
   PDBG_S("no more commands");
 
   /* restore old signal handler
