@@ -87,13 +87,13 @@ typedef struct {
 FILE * pdbg;
 FILE * pdbgc;
 #define PDBG_OPEN    pdbg = fopen(PDGBFILE"main",  "a")  
-#define PDBG_CLOSE   fclose (pdbg)
+#define PDBG_CLOSE   sl_fclose (FIL__, __LINE__, pdbg)
 #define PDBG(arg)    fprintf(pdbg,  "PDBG: step %d\n", arg); fflush(pdbg)
 #define PDBG_D(arg)  fprintf(pdbg,  "PDBG: %d\n", arg); fflush(pdbg)
 #define PDBG_S(arg)  fprintf(pdbg,  "PDBG: %s\n", arg); fflush(pdbg)
 
 #define PDBGC_OPEN   pdbgc = fopen(PDGBFILE"child", "a")  
-#define PDBGC_CLOSE  fclose (pdbgc)
+#define PDBGC_CLOSE  sl_fclose (FIL__, __LINE__, pdbgc)
 #define PDBGC(arg)   fprintf(pdbgc, "PDBG: step %d\n", arg); fflush(pdbgc)
 #define PDBGC_D(arg) fprintf(pdbgc, "PDBG: %d\n", arg); fflush(pdbgc)
 #define PDBGC_S(arg) fprintf(pdbgc, "PDBG: %s\n", arg); fflush(pdbgc)
@@ -114,6 +114,7 @@ FILE * pdbgc;
 #define FIL__  _("sh_gpg.c")
 
 #ifdef GPG_HASH
+
 static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
 {
   char * test_gpg;
@@ -126,11 +127,10 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
 
   SL_ENTER(_("sh_gpg_checksum"));
 
-  tiger_fd = checkfd;
 #if defined(WITH_PGP)
-  test_gpg = sh_tiger_hash_gpg (DEFAULT_PGP_PATH, TIGER_FD, 0);
+  test_gpg = sh_tiger_hash_gpg (DEFAULT_PGP_PATH, checkfd, TIGER_NOLIM);
 #else
-  test_gpg = sh_tiger_hash_gpg (DEFAULT_GPG_PATH, TIGER_FD, 0);
+  test_gpg = sh_tiger_hash_gpg (DEFAULT_GPG_PATH, checkfd, TIGER_NOLIM);
 #endif
   
   test_ptr1 = strchr(GPG_HASH, ':');
@@ -172,14 +172,17 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
     }
 
   k = 0;
-  for (i = 0; i < 127; ++i)
+  if (test_ptr2)
     {
-      if (test_ptr2[i] == '\0')
-	break;
-      if (test_ptr2[i] != ' ')
+      for (i = 0; i < 127; ++i)
 	{
-	  wstrip2[k] = test_ptr2[i];
-	  ++k;
+	  if (test_ptr2[i] == '\0')
+	    break;
+	  if (test_ptr2[i] != ' ')
+	    {
+	      wstrip2[k] = test_ptr2[i];
+	      ++k;
+	    }
 	}
     }
   wstrip2[k] = '\0';
@@ -202,12 +205,49 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
 }
 #endif
 
+struct startup_info {
+  long   line;
+  char * program;
+  long   uid;
+  char * path;
+  char * key_uid;
+  char * key_id;
+};
+
+static struct startup_info startInfo = { 0, NULL, 0, NULL, NULL, NULL };
+
+void sh_gpg_log_startup (void)
+{
+  if (startInfo.program != NULL)
+    {
+      sh_error_handle ((-1), FIL__, startInfo.line, 0, MSG_START_GH,
+		       startInfo.program, startInfo.uid,
+		       startInfo.path,
+		       startInfo.key_uid, startInfo.key_id);
+    }
+  return;
+}
+
+static void sh_gpg_fill_startup (long line, char * program, long uid, char * path, 
+				 char * key_uid, char * key_id)
+{
+  startInfo.line    = line;
+  startInfo.program = sh_util_strdup(program);
+  startInfo.uid     = uid;
+  startInfo.path    = sh_util_strdup(path);
+  startInfo.key_uid = sh_util_strdup(key_uid);
+  startInfo.key_id  = sh_util_strdup(key_id);
+  return;
+}
+
 static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd, 
 			    int mode, char * id, char * homedir)
 {
+  extern int flag_err_debug;
   int pipedes[2];
   FILE * outf = NULL;
   char * envp[2];
+  size_t len;
   char   path[256];
   char   cc1[32];
   char   cc2[32];
@@ -219,19 +259,20 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
   char   cc0[2] = "-";
   char   cc3[32];
   char   cc4[SH_PATHBUF+32];
-  char   cc5[32] = "--no-tty";
+  char   cc5[32];
 #endif
 
   char * arg[9];
 
 #if defined(HAVE_GPG_CHECKSUM)
-  SL_TICKET   checkfd;
+  SL_TICKET   checkfd = -1;
   int         myrand;
   int         i;
 #if defined(__linux__)
   int         get_the_fd(SL_TICKET);
   char        pname[128];
   int         pfd;
+  int         val_return;
 #endif
 #endif
 
@@ -245,7 +286,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
   sl_strlcpy (cc3,   _("--homedir"),    32);
   /* sl_strlcpy (cc4,   sh.effective.home, SH_PATHBUF+32); */
   sl_strlcpy (cc4,   homedir,           SH_PATHBUF+32);
-  sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
+  sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32);
+  sl_strlcpy (cc5,   _("--no-tty"),     32);
 
   /* fprintf(stderr, "YULE: homedir=%s\n", homedir); */
 
@@ -254,9 +296,15 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
     {
       struct stat lbuf;
       int         status_stat = 0;
-      struct passwd * tempres = getpwnam(DEFAULT_IDENT);
-      if (!tempres)
-	tempres = sh_getpwnam(DEFAULT_IDENT);
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      struct passwd    pwd;
+      char          *  buffer = SH_ALLOC(SH_PWBUF_SIZE);
+      struct passwd *  tempres;
+      sh_getpwnam_r(DEFAULT_IDENT, &pwd, buffer, SH_PWBUF_SIZE, &tempres);
+#else
+      struct passwd * tempres = sh_getpwnam(DEFAULT_IDENT);
+#endif
+
       if (!tempres)
 	{
 	  dlog(1, FIL__, __LINE__, 
@@ -318,6 +366,9 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 	}
       sl_strlcpy (cc4, tempres->pw_dir, SH_PATHBUF+32); 
       sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      SH_FREE(buffer);
+#endif
     }
 #endif
 
@@ -369,10 +420,10 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
    */
   if (sh.effective.home != NULL)
     {
-      envp[0] = malloc (sl_strlen(sh.effective.home) + 6); /* free() ok   */
+      len = sl_strlen(sh.effective.home) + 6;
+      envp[0] = malloc (len); /* free() ok   */
       if (envp[0] != NULL)
-	sprintf (envp[0], "HOME=%s",                     /* known to fit  */
-		 sh.effective.home); 
+	sl_snprintf (envp[0], len, _("HOME=%s"), sh.effective.home); 
       envp[1] = NULL;
     }
   else
@@ -388,6 +439,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 	free(envp[0]);
       SL_RETURN( (NULL), _("sh_gpg_popen"));
     }
+
+  fflush (NULL);
   
   source->pid = aud_fork(FIL__, __LINE__);
   
@@ -395,8 +448,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
    */
   if (source->pid == (pid_t) - 1) 
     {
-      close(pipedes[0]);
-      close(pipedes[1]);
+      sl_close_fd(FIL__, __LINE__, pipedes[0]);
+      sl_close_fd(FIL__, __LINE__, pipedes[1]);
       if (envp[0] != NULL) 
 	free(envp[0]);
       SL_RETURN( (NULL), _("sh_gpg_popen"));
@@ -417,8 +470,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
       
       /* close the pipe descriptors 
        */
-      close (pipedes[STDIN_FILENO]);
-      close (pipedes[STDOUT_FILENO]);
+      sl_close_fd (FIL__, __LINE__, pipedes[STDIN_FILENO]);
+      sl_close_fd (FIL__, __LINE__, pipedes[STDOUT_FILENO]);
       
 
 #if defined(WITH_PGP)
@@ -442,9 +495,17 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
  
       /* don't leak file descriptors
        */
-      sh_unix_closeall (3, -1);
+      sh_unix_closeall (3, -1, SL_TRUE); /* in child process */
+
+      if (flag_err_debug != SL_TRUE)
+	{
+	  if (NULL == freopen(_("/dev/null"), "r+", stderr))
+	    {
+	      dlog(1, FIL__, __LINE__, _("Internal error: freopen failed\n"));
+	      aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+	    }
+	}
 
-      freopen(_("/dev/null"), "r+", stderr); 
 
       /* We should become privileged if SUID,
        * to be able to read the keyring.
@@ -476,18 +537,28 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
        * --  emulate an fexecve with checksum testing
        */
 #if defined(WITH_PGP)
-      checkfd = sl_open_read(DEFAULT_PGP_PATH, SL_NOPRIV);
+      checkfd = sl_open_read(FIL__, __LINE__, DEFAULT_PGP_PATH, SL_NOPRIV);
 #else
-      checkfd = sl_open_read(DEFAULT_GPG_PATH, SL_NOPRIV);
+      checkfd = sl_open_read(FIL__, __LINE__, DEFAULT_GPG_PATH, SL_NOPRIV);
 #endif
 
       if (0 != sh_gpg_checksum(checkfd, 0))
-	aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+	{
+	  sl_close(checkfd);
+	  aud__exit(FIL__, __LINE__, EXIT_FAILURE);
+	}
 
       pfd = get_the_fd(checkfd);
-      sprintf(pname, _("/proc/self/fd/%d"),             /* known to fit  */
-		   pfd);
-      if (0 == access(pname, R_OK|X_OK))
+      do {
+	val_return = dup (pfd);
+      } while (val_return < 0 && errno == EINTR);
+      pfd = val_return;
+      sl_close(checkfd);
+      /* checkfd = -1; *//* never read */
+
+      sl_snprintf(pname, sizeof(pname), _("/proc/self/fd/%d"), pfd);
+      if (0 == access(pname, R_OK|X_OK))               /* flawfinder: ignore */
+
 	{
 	  fcntl  (pfd, F_SETFD, FD_CLOEXEC);
 	  retry_aud_execve (FIL__, __LINE__,  pname, arg, envp);
@@ -501,7 +572,6 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 	  
       /* procfs not working, go ahead 
        */
-      sl_close(checkfd);
 #endif
 
 #if defined(HAVE_GPG_CHECKSUM)
@@ -509,17 +579,17 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
        * from knowing when it is safe to slip in a fake executable
        * between the integrity check and the execve
        */
-      myrand = (int) taus_get (&(skey->rng0[0]), &(skey->rng1[0]), 
-			       &(skey->rng2[0]));
+      myrand = (int) taus_get ();
+
       myrand = (myrand < 0) ? (-myrand) : myrand;
       myrand = (myrand % 32) + 2;
 
       for (i = 0; i < myrand; ++i)
 	{
 #if defined(WITH_PGP)
-	  checkfd = sl_open_fastread(DEFAULT_PGP_PATH, SL_NOPRIV);
+	  checkfd = sl_open_fastread(FIL__, __LINE__, DEFAULT_PGP_PATH, SL_NOPRIV);
 #else
-	  checkfd = sl_open_fastread(DEFAULT_GPG_PATH, SL_NOPRIV);
+	  checkfd = sl_open_fastread(FIL__, __LINE__, DEFAULT_GPG_PATH, SL_NOPRIV);
 #endif
 	  if (0 != sh_gpg_checksum(checkfd, 0)) {
 	    aud__exit(FIL__, __LINE__, EXIT_FAILURE);
@@ -550,7 +620,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
   if (envp[0] != NULL) 
     free(envp[0]);
 
-  close (pipedes[STDOUT_FILENO]);
+  sl_close_fd (FIL__, __LINE__, pipedes[STDOUT_FILENO]);
   retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
   retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFL,  O_NONBLOCK);
 
@@ -559,7 +629,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
   if (outf == NULL) 
     {
       aud_kill (FIL__, __LINE__, source->pid, SIGKILL);
-      close (pipedes[STDOUT_FILENO]);
+      sl_close_fd (FIL__, __LINE__, pipedes[STDOUT_FILENO]);
       waitpid (source->pid, NULL, 0);
       source->pid = 0;
       SL_RETURN( (NULL), _("sh_gpg_popen"));
@@ -575,7 +645,7 @@ static int sh_gpg_pclose (sh_gpg_popen_t *source)
   
   SL_ENTER(_("sh_gpg_pclose"));
 
-  status = fclose(source->pipe);
+  status = sl_fclose(FIL__, __LINE__, source->pipe);
   if (status)
     SL_RETURN( (-1), _("sh_gpg_pclose"));
   
@@ -614,9 +684,11 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 
   if (0 != retry_lstat(FIL__, __LINE__, DEFAULT_GPG_PATH, &buf))
     {
+      char errbuf[SH_ERRBUF_SIZE];
+
       status = errno;
       sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, status, MSG_ERR_LSTAT,
-		      sh_error_message(status), DEFAULT_GPG_PATH);
+		      sh_error_message(status, errbuf, sizeof(errbuf)), DEFAULT_GPG_PATH);
       SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
 
@@ -624,7 +696,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
     SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
 
 #ifdef HAVE_GPG_CHECKSUM
-  checkfd = sl_open_read(DEFAULT_GPG_PATH, SL_YESPRIV);
+  checkfd = sl_open_read(FIL__, __LINE__, DEFAULT_GPG_PATH, SL_YESPRIV);
 
   if (0 != sh_gpg_checksum(checkfd, 1))
     {
@@ -644,16 +716,18 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 
   if (0 != retry_lstat(FIL__, __LINE__, DEFAULT_PGP_PATH, &buf))
     {
+      char errbuf[SH_ERRBUF_SIZE];
+
       status = errno;
       sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, status, MSG_ERR_LSTAT,
-		      sh_error_message(status), DEFAULT_PGP_PATH);
+		      sh_error_message(status, errbuf, sizeof(errbuf)), DEFAULT_PGP_PATH);
       SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
   if (0 != tf_trust_check (DEFAULT_PGP_PATH, SL_YESPRIV))
     SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
 
 #ifdef HAVE_GPG_CHECKSUM
-  checkfd = sl_open_read(DEFAULT_PGP_PATH, SL_YESPRIV);
+  checkfd = sl_open_read(FIL__, __LINE__, DEFAULT_PGP_PATH, SL_YESPRIV);
 
   if (0 != sh_gpg_checksum(checkfd, 1))
     {
@@ -688,7 +762,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 
   errno = 0;
 
-  while (NULL != fgets(line, 255, source.pipe))
+  while (NULL != fgets(line, sizeof(line), source.pipe))
     {
 
       TPT(((0), FIL__, __LINE__, _("msg=<gpg out: %s>\n"), line));
@@ -743,7 +817,8 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
       if (0 == sl_strncmp(_("GOODSIG"), &line[9], 7))
 	{
 	  sl_strlcpy (sign_id, &line[25], SH_MINIBUF+1);
-	  sign_id[sl_strlen(sign_id)-1] = '\0';  /* remove trailing '"' */
+	  if (sign_id)
+	    sign_id[sl_strlen(sign_id)-1] = '\0';  /* remove trailing '"' */
 	  have_id = GOOD;
 	} 
       if (0 == sl_strncmp(_("VALIDSIG"), &line[9], 8))
@@ -767,9 +842,16 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 	{
 	  ptr = strchr ( line, '"');
 	  ++ptr;
-	  sl_strlcpy (sign_id, ptr, SH_MINIBUF+1);
-	  sign_id[sl_strlen(sign_id)-1] = '\0'; /* remove trailing dot */
-	  sign_id[sl_strlen(sign_id)-2] = '\0'; /* remove trailing '"' */
+	  if (ptr)
+	    {
+	      sl_strlcpy (sign_id, ptr, SH_MINIBUF+1);
+	      sign_id[sl_strlen(sign_id)-1] = '\0'; /* remove trailing dot */
+	      sign_id[sl_strlen(sign_id)-2] = '\0'; /* remove trailing '"' */
+	    }
+	  else
+	    {
+	      sl_strlcpy (sign_id, _("(null)"), SH_MINIBUF+1);
+	    }
 	  have_id = GOOD;
 	}
 #endif
@@ -777,6 +859,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 
   if (ferror(source.pipe) && errno == EAGAIN) 
     {
+      retry_msleep(0,10); /* sleep 10 ms to avoid starving the gpg child writing to the pipe */
       clearerr(source.pipe);
       goto xagain;
     }
@@ -803,8 +886,11 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 
   errno = 0;
 
-  while (NULL != fgets(line, 255, source.pipe))
+  while (NULL != fgets(line, sizeof(line), source.pipe))
     {
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+      char * saveptr = NULL;
+#endif
       if (line[strlen(line)-1] == '\n')
 	line[strlen(line)-1] = ' ';
       sh_error_handle(SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN, 
@@ -813,17 +899,33 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 
       if (sl_strlen(line) < 18) 
 	continue;
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+      ptr = strtok_r (line, " ", &saveptr);
+#else
       ptr = strtok (line, " ");
+#endif
       while (ptr)
 	{
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+	  ptr = strtok_r (NULL, " ", &saveptr);
+#else
 	  ptr = strtok (NULL, " ");
+#endif
 	  if (ptr && 0 == sl_strncmp (ptr, _("fingerprint"), 11))
 	    {
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+	      ptr = strtok_r (NULL, " ", &saveptr); /* to '=' */
+#else
 	      ptr = strtok (NULL, " "); /* to '=' */
+#endif
 	      sign_fp[0] = '\0';
 	      while (ptr)
 		{
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_STRTOK_R)
+		  ptr = strtok_r (NULL, " ", &saveptr); /* part of fingerprint */
+#else
 		  ptr = strtok (NULL, " "); /* part of fingerprint */
+#endif
 		  sl_strlcat (sign_fp, ptr, SH_MINIBUF+1);
 		}
 	      /* sign_fp[sl_strlen(sign_fp)-1] = '\0'; remove trailing '\n' */
@@ -836,6 +938,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp,
 
   if (ferror(source.pipe) && errno == EAGAIN) 
     {
+      retry_msleep(0,10); /* sleep 10 ms to avoid starving the gpg child writing to the pipe */
       clearerr(source.pipe);
       goto yagain;
     }
@@ -884,7 +987,12 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
   char  * homedir = sh.effective.home;
 #if defined(SH_WITH_SERVER)
   struct passwd * tempres;
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+  struct passwd    pwd;
+  char           * buffer = SH_ALLOC(SH_PWBUF_SIZE);
 #endif
+#endif
+
 #ifdef USE_FINGERPRINT
 #include "sh_gpg_fp.h"
 #endif
@@ -906,6 +1014,11 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	   _("This looks like an unexpected internal error.\n"));
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1, sh.prg_name);
       aud_exit (FIL__, __LINE__, EXIT_FAILURE);
+#if defined(SH_WITH_SERVER)
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      SH_FREE(buffer);
+#endif
+#endif
       SL_RETURN( (-1), _("sh_gpg_check_sign"));
     }
   
@@ -913,9 +1026,12 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD1 = %d>\n"), fd1));
 #if defined(SH_WITH_SERVER)
-      tempres = getpwnam(DEFAULT_IDENT);
-      if (!tempres)
-	tempres = sh_getpwnam(DEFAULT_IDENT);
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      sh_getpwnam_r(DEFAULT_IDENT, &pwd, buffer, SH_PWBUF_SIZE, &tempres);
+#else
+      tempres = sh_getpwnam(DEFAULT_IDENT);
+#endif
+
       if ((tempres != NULL) && (0 == sl_ret_euid()))
 	{
 	  /* privileges not dropped yet*/
@@ -931,9 +1047,12 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD2 = %d>\n"), fd2));
 #if defined(SH_WITH_SERVER)
-      tempres = getpwnam(DEFAULT_IDENT);
-      if (!tempres)
-	tempres = sh_getpwnam(DEFAULT_IDENT);
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      sh_getpwnam_r(DEFAULT_IDENT, &pwd, buffer, SH_PWBUF_SIZE, &tempres);
+#else
+      tempres = sh_getpwnam(DEFAULT_IDENT);
+#endif
+
       if ((tempres != NULL) && (0 == sl_ret_euid()))
 	{
 	  /* privileges not dropped yet*/
@@ -966,7 +1085,8 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	  if (smsg == S_FALSE)
 	    {
 	      tmp  = sh_util_safe_name(gp.conf_id);
-	      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH,
+	      sh_gpg_fill_startup (__LINE__,
+				   /* sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH, */
 			       sh.prg_name, sh.real.uid,
 			       (sh.flag.hidefile == S_TRUE) ? 
 			       _("(hidden)") : file_path('C', 'R'), 
@@ -975,6 +1095,11 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	      SH_FREE(tmp);
 	    }
 	  smsg = S_TRUE;
+#if defined(SH_WITH_SERVER)
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+	  SH_FREE(buffer);
+#endif
+#endif
 	  SL_RETURN(0, _("sh_gpg_check_sign"));
 	}
       else
@@ -993,7 +1118,8 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
       if (smsg == S_FALSE)
 	{
 	  tmp = sh_util_safe_name(gp.conf_id);
-	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH,
+	  sh_gpg_fill_startup (__LINE__,
+	  /* sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH, */
 			   sh.prg_name, sh.real.uid,
 			   (sh.flag.hidefile == S_TRUE) ? 
 			   _("(hidden)") : file_path('C', 'R'), 
@@ -1002,6 +1128,11 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	  SH_FREE(tmp);
 	}
       smsg = S_TRUE;
+#if defined(SH_WITH_SERVER)
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      SH_FREE(buffer);
+#endif
+#endif
       SL_RETURN(0, _("sh_gpg_check_sign"));
 #endif
     }
@@ -1011,6 +1142,11 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
       if ((sl_strcmp(gp.data_id, gp.conf_id) == 0) &&
 	  (sl_strcmp(gp.data_fp, gp.conf_fp) == 0))
 	{
+#if defined(SH_WITH_SERVER)
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+	  SH_FREE(buffer);
+#endif
+#endif
 	  SL_RETURN(0, _("sh_gpg_check_sign"));
 	}
       else
@@ -1038,9 +1174,15 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
       char  * e_home = sh.effective.home;
 
 #if defined(SH_WITH_SERVER)
-      struct passwd * e_tempres = getpwnam(DEFAULT_IDENT);
-      if (!e_tempres)
-	e_tempres = sh_getpwnam(DEFAULT_IDENT);
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      struct passwd    e_pwd;
+      char          *  e_buffer = SH_ALLOC(SH_PWBUF_SIZE);
+      struct passwd *  e_tempres;
+      sh_getpwnam_r(DEFAULT_IDENT, &e_pwd, e_buffer, SH_PWBUF_SIZE, &e_tempres);
+#else
+      struct passwd * e_tempres = sh_getpwnam(DEFAULT_IDENT);
+#endif
+
       if ((e_tempres != NULL) && (0 == sl_ret_euid()))   
 	{
 	  /* privileges not dropped yet */
@@ -1056,6 +1198,12 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	   DEFAULT_PGP_PATH,
 #endif
 	   (int) e_uid, e_home);
+
+#if defined(SH_WITH_SERVER)
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      SH_FREE(e_buffer);
+#endif
+#endif
     }
 
   TPT(((0), FIL__, __LINE__, _("msg=<Status = %d>\n"), status));
@@ -1066,6 +1214,82 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
   return (-1); /* make compiler happy */
 }  
 
+#define FGETS_BUF 16384
+
+SL_TICKET sh_gpg_extract_signed(SL_TICKET fd)
+{
+  FILE * fin_cp = NULL;
+  char * buf    = NULL;
+  int    bufc;
+  int    flag_pgp    = S_FALSE;
+  int    flag_nohead = S_FALSE;
+  SL_TICKET fdTmp = (-1);
+  SL_TICKET open_tmp (void);
+
+  /* extract the data and copy to temporary file
+   */
+  fdTmp = open_tmp();
+
+  fin_cp = fdopen(dup(get_the_fd(fd)), "rb");
+  buf = SH_ALLOC(FGETS_BUF);
+
+  while (NULL != fgets(buf, FGETS_BUF, fin_cp))
+    {
+      bufc = 0; 
+      while (bufc < FGETS_BUF) { 
+	if (buf[bufc] == '\n') { ++bufc; break; }
+	++bufc;
+      }
+
+      if (flag_pgp == S_FALSE &&
+	  (0 == sl_strcmp(buf, _("-----BEGIN PGP SIGNED MESSAGE-----\n"))||
+	   0 == sl_strcmp(buf, _("-----BEGIN PGP MESSAGE-----\n")))
+	  )
+	{
+	  flag_pgp = S_TRUE;
+	  sl_write(fdTmp, buf, bufc);
+	  continue;
+	}
+      
+      if (flag_pgp == S_TRUE && flag_nohead == S_FALSE)
+	{
+	  if (buf[0] == '\n')
+	    {
+	      flag_nohead = S_TRUE;
+	      sl_write(fdTmp, buf, 1);
+	      continue;
+	    }
+	  else if (0 == sl_strncmp(buf, _("Hash:"), 5) ||
+		   0 == sl_strncmp(buf, _("NotDashEscaped:"), 15))
+	    {
+	      sl_write(fdTmp, buf, bufc);
+	      continue;
+	    }
+	  else
+	    continue;
+	}
+    
+      if (flag_pgp == S_TRUE && buf[0] == '\n')
+	{
+	  sl_write(fdTmp, buf, 1);
+	}
+      else if (flag_pgp == S_TRUE)
+	{
+	  /* sl_write_line(fdTmp, buf, bufc); */
+	  sl_write(fdTmp, buf, bufc);
+	}
+      
+      if (flag_pgp == S_TRUE && 
+	  0 == sl_strcmp(buf, _("-----END PGP SIGNATURE-----\n")))
+	break;
+    }
+  SH_FREE(buf);
+  sl_fclose(FIL__, __LINE__, fin_cp); /* fin_cp = fdopen(dup(), "rb"); */
+  sl_rewind (fdTmp);
+
+  return fdTmp;
+}
+
 /* #ifdef WITH_GPG */
 #endif
 
