@@ -51,6 +51,12 @@
 #include <pwd.h>
 #include <grp.h>
 
+#include "sh_pthread.h"
+
+extern  int sl_close_fd (const char * file, int line, int fd);
+extern  int sl_fclose (const char * file, int line, FILE * fp);
+
+
 #ifndef _PATH_PASSWD
 #define _PATH_PASSWD "/etc/passwd"
 #endif
@@ -58,13 +64,19 @@
 #define _PATH_GROUP "/etc/group"
 #endif
 
+#undef  FIL__
+#define FIL__  _("sh_static.c")
+
+extern  int sl_strlcpy(char * dst, /*@null@*/const char * src, size_t siz);
+extern  int sl_strlcat(char * dst, /*@null@*/const char * src, size_t siz);
 
 
 /**********************************************************************/
 /* Sizes for staticly allocated buffers. */
 
 #define PWD_BUFFER_SIZE 256
-#define GRP_BUFFER_SIZE 256
+#define GRP_BUFFER_SIZE 3584
+#define GRP_BUFFER_SIZE_MALLOC 32768
 
 /**********************************************************************/
 /* Prototypes for internal functions. */
@@ -89,10 +101,10 @@ static int __pgsreader(int (*__parserfunc)(void *d, char *line), void *data,
 #define DO_GETXXKEY_R_KEYTYPE	const char *__restrict
 #define DO_GETXXKEY_R_PATHNAME  _PATH_PASSWD
 
-static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
-			   GETXXKEY_R_ENTTYPE *__restrict resultbuf,
-			   char *__restrict buffer, size_t buflen,
-			   GETXXKEY_R_ENTTYPE **__restrict result)
+int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
+		    GETXXKEY_R_ENTTYPE *__restrict resultbuf,
+		    char *__restrict buffer, size_t buflen,
+		    GETXXKEY_R_ENTTYPE **__restrict result)
 {
   FILE *stream;
   int rv;
@@ -118,7 +130,7 @@ static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
 	break;
       }
     } while (1);
-    fclose(stream);
+    sl_fclose(FIL__, __LINE__, stream);
   }
   
   return rv;
@@ -137,10 +149,10 @@ static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
 #define DO_GETXXKEY_R_KEYTYPE	const char *__restrict
 #define DO_GETXXKEY_R_PATHNAME  _PATH_GROUP
 
-static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
-			   GETXXKEY_R_ENTTYPE *__restrict resultbuf,
-			   char *__restrict buffer, size_t buflen,
-			   GETXXKEY_R_ENTTYPE **__restrict result)
+int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
+		    GETXXKEY_R_ENTTYPE *__restrict resultbuf,
+		    char *__restrict buffer, size_t buflen,
+		    GETXXKEY_R_ENTTYPE **__restrict result)
 {
   FILE *stream;
   int rv;
@@ -166,7 +178,7 @@ static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
 	break;
       }
     } while (1);
-    fclose(stream);
+    sl_fclose(FIL__, __LINE__, stream);
   }
   
   return rv;
@@ -185,10 +197,10 @@ static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
 #define DO_GETXXKEY_R_KEYTYPE	uid_t
 #define DO_GETXXKEY_R_PATHNAME  _PATH_PASSWD
 
-static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
-			   GETXXKEY_R_ENTTYPE *__restrict resultbuf,
-			   char *__restrict buffer, size_t buflen,
-			   GETXXKEY_R_ENTTYPE **__restrict result)
+int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
+		    GETXXKEY_R_ENTTYPE *__restrict resultbuf,
+		    char *__restrict buffer, size_t buflen,
+		    GETXXKEY_R_ENTTYPE **__restrict result)
 {
   FILE *stream;
   int rv;
@@ -214,7 +226,7 @@ static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
 	break;
       }
     } while (1);
-    fclose(stream);
+    sl_fclose(FIL__, __LINE__, stream);
   }
   
   return rv;
@@ -233,10 +245,10 @@ static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
 #define DO_GETXXKEY_R_KEYTYPE	gid_t
 #define DO_GETXXKEY_R_PATHNAME  _PATH_GROUP
 
-static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
-			   GETXXKEY_R_ENTTYPE *__restrict resultbuf,
-			   char *__restrict buffer, size_t buflen,
-			   GETXXKEY_R_ENTTYPE **__restrict result)
+int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
+		    GETXXKEY_R_ENTTYPE *__restrict resultbuf,
+		    char *__restrict buffer, size_t buflen,
+		    GETXXKEY_R_ENTTYPE **__restrict result)
 {
   FILE *stream;
   int rv;
@@ -262,7 +274,7 @@ static int GETXXKEY_R_FUNC(DO_GETXXKEY_R_KEYTYPE key,
 	break;
       }
     } while (1);
-    fclose(stream);
+    sl_fclose(FIL__, __LINE__, stream);
   }
   
   return rv;
@@ -278,6 +290,11 @@ struct passwd * sh_getpwuid(uid_t uid)
 	return result;
 }
 
+struct passwd * getpwuid(uid_t uid)
+{
+        return sh_getpwuid(uid);
+}
+
 struct group * sh_getgrgid(gid_t gid)
 {
 	static char buffer[GRP_BUFFER_SIZE];
@@ -288,6 +305,10 @@ struct group * sh_getgrgid(gid_t gid)
 	return result;
 }
 
+struct group * getgrgid(gid_t gid)
+{
+        return sh_getgrgid(gid);
+}
 
 struct passwd * sh_getpwnam(const char *name)
 {
@@ -309,28 +330,28 @@ struct group * sh_getgrnam(const char *name)
 	return result;
 }
 
-#define LOCK		((void) 0)
-#define UNLOCK		((void) 0)
+SH_MUTEX_STATIC(pwf_lock, PTHREAD_MUTEX_INITIALIZER);
+
 
 static FILE *pwf = NULL;
 
 void  sh_setpwent(void)
 {
-	LOCK;
+        SH_MUTEX_LOCK(pwf_lock);
 	if (pwf) {
 		rewind(pwf);
 	}
-	UNLOCK;
+	SH_MUTEX_UNLOCK(pwf_lock);
 }
 
 void  sh_endpwent(void)
 {
-	LOCK;
+        SH_MUTEX_LOCK(pwf_lock);
 	if (pwf) {
-		fclose(pwf);
+		sl_fclose(FIL__, __LINE__, pwf);
 		pwf = NULL;
 	}
-	UNLOCK;
+	SH_MUTEX_UNLOCK(pwf_lock);
 }
 
 
@@ -340,7 +361,7 @@ static int  sh_getpwent_r(struct passwd *__restrict resultbuf,
 {
 	int rv;
 
-	LOCK;
+        SH_MUTEX_LOCK(pwf_lock);
 
 	*result = NULL;				/* In case of error... */
 
@@ -358,33 +379,33 @@ static int  sh_getpwent_r(struct passwd *__restrict resultbuf,
 	}
 
  ERR:
-	UNLOCK;
+	; /* 'label at end of compound statement' */
+	SH_MUTEX_UNLOCK(pwf_lock);
 
 	return rv;
 }
 
-#define LOCK		((void) 0)
-#define UNLOCK		((void) 0)
+SH_MUTEX_STATIC(grf_lock, PTHREAD_MUTEX_INITIALIZER);
 
 static FILE *grf = NULL;
 
 void  sh_setgrent(void)
 {
-	LOCK;
+	SH_MUTEX_LOCK(grf_lock);
 	if (grf) {
 		rewind(grf);
 	}
-	UNLOCK;
+	SH_MUTEX_UNLOCK(grf_lock);
 }
 
 void  sh_endgrent(void)
 {
-	LOCK;
+	SH_MUTEX_LOCK(grf_lock);
 	if (grf) {
-		fclose(grf);
+		sl_fclose(FIL__, __LINE__, grf);
 		grf = NULL;
 	}
-	UNLOCK;
+	SH_MUTEX_UNLOCK(grf_lock);
 }
 
 static int sh_getgrent_r(struct group *__restrict resultbuf,
@@ -393,7 +414,7 @@ static int sh_getgrent_r(struct group *__restrict resultbuf,
 {
 	int rv;
 
-	LOCK;
+	SH_MUTEX_LOCK(grf_lock);
 
 	*result = NULL;				/* In case of error... */
 
@@ -411,7 +432,8 @@ static int sh_getgrent_r(struct group *__restrict resultbuf,
 	}
 
  ERR:
-	UNLOCK;
+	; /* 'label at end of compound statement' */
+	SH_MUTEX_UNLOCK(grf_lock);
 
 	return rv;
 }
@@ -434,7 +456,7 @@ struct group * sh_getgrent(void)
 	static struct group gr;
 	struct group *result;
 
-	 sh_getgrent_r(&gr, line_buff, sizeof(line_buff), &result);
+	sh_getgrent_r(&gr, line_buff, sizeof(line_buff), &result);
 	return result;
 }
 
@@ -445,7 +467,8 @@ int  sh_initgroups(const char *user, gid_t gid)
 	int num_groups, rv;
 	char **m;
 	struct group group;
-	char buff[PWD_BUFFER_SIZE];
+
+	char * buff = malloc(GRP_BUFFER_SIZE_MALLOC);
 
 	rv = -1;
 
@@ -459,7 +482,7 @@ int  sh_initgroups(const char *user, gid_t gid)
 		*group_list = gid;
 		num_groups = 1;
 
-		while (!__pgsreader(__parsegrent, &group, buff, sizeof(buff), grf)) {
+		while (!__pgsreader(__parsegrent, &group, buff, GRP_BUFFER_SIZE_MALLOC, grf)) {
 			assert(group.gr_mem); /* Must have at least a NULL terminator. */
 			if (group.gr_gid != gid) {
 				for (m=group.gr_mem ; *m ; m++) {
@@ -483,12 +506,13 @@ int  sh_initgroups(const char *user, gid_t gid)
 
 		rv = setgroups(num_groups, group_list);
 	DO_CLOSE:
-		fclose(grf);
+		sl_fclose(FIL__, __LINE__, grf);
 	}
 
 	/* group_list will be NULL if initial malloc failed, which may trigger
 	 * warnings from various malloc debuggers. */
 	free(group_list);
+	free(buff);
 	return rv;
 }
 
@@ -666,8 +690,8 @@ static int __pgsreader(int (*__parserfunc)(void *d, char *line), void *data,
 
 		skip = 0;
 		do {
-			if (!fgets_unlocked(line_buff, buflen, f)) {
-				if (feof_unlocked(f)) {
+			if (!fgets(line_buff, buflen, f)) {
+				if (feof(f)) {
 					rv = ENOENT;
 				}
 				break;
@@ -677,8 +701,12 @@ static int __pgsreader(int (*__parserfunc)(void *d, char *line), void *data,
 			if (line_buff[line_len] == '\n') {
 				line_buff[line_len] = 0;
 			} else if (line_len + 2 == buflen) { /* line too long */
+			        rv = ERANGE;
+				break;
+				/*
 				++skip;
 				continue;
+				*/
 			}
 
 			if (skip) {
@@ -848,7 +876,7 @@ static int __pgsreader(int (*__parserfunc)(void *d, char *line), void *data,
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-/* close()
+/* sl_close_fd(FIL__, __LINE__, )
  */
 #include <unistd.h>
 
@@ -860,8 +888,7 @@ static int __pgsreader(int (*__parserfunc)(void *d, char *line), void *data,
  */
 #include <arpa/nameser.h>
 
-#define BIGLOCK
-#define BIGUNLOCK
+SH_MUTEX_STATIC(resolv_lock, PTHREAD_MUTEX_INITIALIZER);
 
 #define __UCLIBC_HAS_IPV6__
 #define MAX_RECURSE 5
@@ -883,6 +910,7 @@ static char * __searchdomain[MAX_SEARCH];
 /*#define DEBUG*/
 
 #ifdef DEBUG
+/* flawfinder: ignore *//* definition of debug macro */
 #define DPRINTF(X,args...) fprintf(stderr, X, ##args)
 #else
 #define DPRINTF(X,args...)
@@ -1164,25 +1192,25 @@ static int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 
 	DPRINTF("Looking up type %d answer for '%s'\n", type, name);
 
-	LOCK;
+	SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 	ns %= nscount;
-	UNLOCK;
+	SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 
 	while (retries++ < MAX_RETRIES) {
 		if (fd != -1)
-			close(fd);
+			sl_close_fd(FIL__, __LINE__, fd);
 
 		memset(packet, 0, PACKETSZ);
 
 		memset(&h, 0, sizeof(h));
 
 		/* Mess with globals while under lock */
-		LOCK;
+		SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 		++id;
 		id &= 0xffff;
 		h.id = id;
 		dns = nsip[ns];
-		UNLOCK;
+		SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 
 		h.qdcount = 1;
 		h.rd = 1;
@@ -1193,14 +1221,14 @@ static int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 		if (i < 0)
 			goto fail;
 
-		strncpy(lookup,name,MAXDNAME);
-		BIGLOCK;
+		sl_strlcpy(lookup,name,MAXDNAME);
+		SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 		if (variant < __searchdomains && strchr(lookup, '.') == NULL)
 		{
-		    strncat(lookup,".", MAXDNAME);
-		    strncat(lookup,__searchdomain[variant], MAXDNAME);
+		    sl_strlcat(lookup,".", MAXDNAME);
+		    sl_strlcat(lookup,__searchdomain[variant], MAXDNAME);
 		}
-		BIGUNLOCK;
+		SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 		DPRINTF("lookup name: %s\n", lookup);
 		q.dotted = (char *)lookup;
 		q.qtype = type;
@@ -1277,13 +1305,13 @@ static int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 
 		DPRINTF("id = %d, qr = %d\n", h.id, h.qr);
 
-		LOCK;
+		SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 		if ((h.id != id) || (!h.qr)) {
-			UNLOCK;
+			SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 			/* unsolicited */
 			goto again;
 		}
-		UNLOCK;
+		SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 
 
 		DPRINTF("Got response %s\n", "(i think)!");
@@ -1329,7 +1357,7 @@ static int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 		DPRINTF("Answer name = |%s|\n", a->dotted);
 		DPRINTF("Answer type = |%d|\n", a->atype);
 
-		close(fd);
+		sl_close_fd(FIL__, __LINE__, fd);
 
 		if (outpacket)
 			*outpacket = packet;
@@ -1344,9 +1372,9 @@ static int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 		{
 		    int sdomains;
 
-		    BIGLOCK;
+		    SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 		    sdomains=__searchdomains;
-		    BIGUNLOCK;
+		    SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 		    variant = 0;
 		    if (retries >= nscount*(sdomains+1))
 			goto fail;
@@ -1356,18 +1384,18 @@ static int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 		/* if there are searchdomains, try them or fallback as passed */
 		{
 		    int sdomains;
-		    BIGLOCK;
+		    SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 		    sdomains=__searchdomains;
-		    BIGUNLOCK;
+		    SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 
 		    if (variant < sdomains) {
 			/* next search */
 			variant++;
 		    } else {
 			/* next server, first search */
-			LOCK;
+			SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 			ns = (ns + 1) % nscount;
-			UNLOCK;
+			SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 			variant = 0;
 		    }
 		}
@@ -1375,7 +1403,7 @@ static int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 
 fail:
 	if (fd != -1)
-	    close(fd);
+	    sl_close_fd(FIL__, __LINE__, fd);
 	if (lookup)
 	    free(lookup);
 	if (packet)
@@ -1531,12 +1559,12 @@ static int __read_etc_hosts_r(FILE * fp, const char * name, int type,
         }
         
 		if (action!=GETHOSTENT) {
-			fclose(fp);
+			sl_fclose(FIL__, __LINE__, fp);
 		}
 		return ret;
 	}
 	if (action!=GETHOSTENT) {
-		fclose(fp);
+		sl_fclose(FIL__, __LINE__, fp);
 	}
 	return ret;
 }
@@ -1554,7 +1582,7 @@ int __get_hosts_byname_r(const char * name, int type,
 	return(__read_etc_hosts_r(NULL, name, type, GET_HOSTS_BYNAME, result_buf, buf, buflen, result, h_errnop));
 }
 
-static int __open_nameservers()
+static int __open_nameservers(void)
 {
 	FILE *fp;
 	int i;
@@ -1562,10 +1590,9 @@ static int __open_nameservers()
 	char szBuffer[128], *p, *argv[RESOLV_ARGS];
 	int argc;
 
-	BIGLOCK;
-	if (__nameservers > 0) { 
-	    BIGUNLOCK;
-	    return 0;
+	SH_MUTEX_LOCK(resolv_lock);
+	if (__nameservers > 0) {
+	  goto the_end;
 	}
 
 	if ((fp = fopen("/etc/resolv.conf", "r")) ||
@@ -1605,12 +1632,14 @@ static int __open_nameservers()
 				}
 			}
 		}
-		fclose(fp);
+		sl_fclose(FIL__, __LINE__, fp);
 	} else {
 	    DPRINTF("failed to open %s\n", "resolv.conf");
 	}
 	DPRINTF("nameservers = %d\n", __nameservers);
-	BIGUNLOCK;
+ the_end:
+	; /* 'label at end of compound statement' */
+	SH_MUTEX_UNLOCK(resolv_lock);
 	return 0;
 }
 
@@ -1687,10 +1716,10 @@ static int sh_gethostbyname_r(const char * name,
 
 	for (;;) {
 
-	BIGLOCK;
+	SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 	__nameserversXX=__nameservers;
 	__nameserverXX=__nameserver;
-	BIGUNLOCK;
+	SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 		i = __dns_lookup(buf, T_A, __nameserversXX, __nameserverXX, &packet, &a);
 
 		if (i < 0) {
@@ -1880,7 +1909,7 @@ static int sh_gethostbyaddr_r (const void *addr, socklen_t len, int type,
 		return ERANGE;
 
 	if(type == AF_INET) {
-		unsigned char *tmp_addr = (unsigned char *)addr;
+		const unsigned char *tmp_addr = (const unsigned char *)addr;
 
 		memcpy(&in->s_addr, addr, len);
 
@@ -1907,10 +1936,10 @@ static int sh_gethostbyaddr_r (const void *addr, socklen_t len, int type,
 
 	for (;;) {
 
-	BIGLOCK;
+	SH_MUTEX_LOCK_UNSAFE(resolv_lock);
 	__nameserversXX=__nameservers;
 	__nameserverXX=__nameserver;
-	BIGUNLOCK;
+	SH_MUTEX_UNLOCK_UNSAFE(resolv_lock);
 		i = __dns_lookup(buf, T_PTR, __nameserversXX, __nameserverXX, &packet, &a);
 
 		if (i < 0) {
@@ -1982,5 +2011,10 @@ struct hostent * sh_gethostbyaddr (const void *addr, socklen_t len, int type)
 }
 
 /* NEED_STATIC_LIBS */
+#else
+
+/* include something to avoid empty compilation unit */
+#include <stdio.h>
+
 #endif
 
