@@ -15,9 +15,9 @@
 #include <stdint.h>
 #endif
 
-#include <unistd.h>
-#include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
 #include <fcntl.h>
 #include <signal.h>
 
@@ -258,7 +258,7 @@ int dlog (int flag, const char * file, int line,  const char *fmt, ...)
     sl_strlcpy(tmp, fmt, 512);
   else
     sl_strlcpy(tmp, fmt, 256);
-  retval = sl_strlen(tmp);
+  retval = strlen(tmp);
   if (retval > 0 && tmp[retval-1] == '\n')
     tmp[retval-1] = '\0';
   retval = 0;
@@ -273,7 +273,7 @@ int dlog (int flag, const char * file, int line,  const char *fmt, ...)
 	sl_strlcat (msg, ".  ", 256);
       sprintf      (val, _("[%2d] "), trace_level);
       sl_strlcat   (msg,     val,   256);
-      sl_vsnprintf (&msg[sl_strlen(msg)], 255, tmp, ap);
+      sl_vsnprintf (&msg[strlen(msg)], 255, tmp, ap);
       sl_snprintf  (tmp, 255, _(" \t - File %c%s%c at line %d"), 
 		    0x22, file, 0x22, line);
       sl_strlcat   (msg,     tmp,   512);
@@ -930,7 +930,7 @@ int sl_strlcpy(char * dst, /*@null@*/const char * src, size_t siz)
     }
   else if (src == NULL)
     {
-      if (siz > 0) 
+      if (dst && siz > 0) 
 	dst[0] = '\0';
       return SL_ENONE;
     }
@@ -1027,11 +1027,11 @@ int sl_strcasecmp(const char * one, const char * two)
       do {
 	if (*one && *two)
 	  {
-	    if (tolower(*one) == tolower(*two))
+	    if (tolower((int) *one) == tolower((int) *two))
 	      {
 		++one; ++two;
 	      }
-	    else if (tolower(*one) < tolower(*two))
+	    else if (tolower((int) *one) < tolower((int) *two))
 	      return -1;
 	    else
 	      return 1;
@@ -1534,6 +1534,7 @@ int sl_policy_get_user(const char * user)
   SL_REQUIRE(uids_are_stored == SL_FALSE, _("uids_are_stored == SL_FALSE"));
   SL_REQUIRE (sl_save_uids() == SL_ENONE, _("sl_save_uids() == SL_ENONE"));
 
+#ifndef SH_ALLOW_SUID
   if (euid != ruid || egid != rgid)
     {
 #if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
@@ -1555,6 +1556,7 @@ int sl_policy_get_user(const char * user)
       free(buffer);
 #endif
     }
+#endif
   SL_IRETURN(SL_ENONE, _("sl_policy_get_user"));
 }
 
@@ -1594,9 +1596,6 @@ static char badfd_orig_file[64] = { '\0' };
 static int  badfd_orig_line = -1;
 static char badfd_orig_mesg[128];
 
-SH_MUTEX_STATIC(mutex_ticket, PTHREAD_MUTEX_INITIALIZER);
-
-static unsigned int nonce_counter = TOFFSET;
 
 char * sl_check_stale()
 {
@@ -1621,12 +1620,43 @@ char * sl_check_badfd()
   return badfd_orig_mesg;
 }
 
+typedef struct { volatile unsigned int atom; } atomic_t;
+static atomic_t nonce_counter = { TOFFSET };
+
+#if defined(__GNUC__) && (defined(__i486__) || defined(__x86_64__))
+/* from linux/include/asm-i386/atomic.h */
+static unsigned int atomic_add ( unsigned int i, atomic_t *var)
+{
+  unsigned int j = i;
+
+  __asm__ __volatile__ ("lock; xaddl %0, %1"
+			: "+r" (i), "+m" (var->atom)
+			: : "memory");
+  return j+i; 
+}
+#else
+SH_MUTEX_STATIC(mutex_ticket, PTHREAD_MUTEX_INITIALIZER);
+
+static unsigned int atomic_add ( unsigned int i, atomic_t *var)
+{
+  volatile unsigned int j;
+
+  SH_MUTEX_LOCK_UNSAFE(mutex_ticket);
+  var->atom += i;
+  j = var->atom;
+  SH_MUTEX_UNLOCK_UNSAFE(mutex_ticket);
+
+  return j;
+}
+#endif
+
 static
 SL_TICKET sl_create_ticket (unsigned int myindex) 
 {
   unsigned int high; /* index */ 
   unsigned int low;  /* nonce */
   SL_TICKET    retval = SL_EINTERNAL;
+  unsigned int nonce;/* nonce */
 
   SL_ENTER(_("sl_create_ticket"));
 
@@ -1647,30 +1677,30 @@ SL_TICKET sl_create_ticket (unsigned int myindex)
       goto out_ticket;
     }
 
-  SH_MUTEX_LOCK_UNSAFE(mutex_ticket);
+  nonce = atomic_add(1, &nonce_counter);
 
-  low = nonce_counter & 0xffff;
+  /* Wrap around the nonce counter.
+   * This is a dirty trick.
+   */
+  if (nonce > 0x7fff)
+    {
+      nonce_counter.atom = TOFFSET;
+      nonce = atomic_add(1, &nonce_counter);
+    }
+
+  low = nonce & 0xffff;
 
   /* Overflow -> nonce too big.
    */
-  if ((low != nonce_counter++) || low == 0)
+  if ((low != nonce) || low == 0)
     {
       retval = SL_EINTERNAL03;
       goto out_ticket;
     }
- 
-  /* Wrap around the nonce counter.
-   * This is a dirty trick.
-   */
-  if (nonce_counter > 0x7fff)
-    nonce_counter = TOFFSET;
 
   retval = (SL_TICKET) ((high << 16) | low);
 
  out_ticket:
-  ;
-
-  SH_MUTEX_UNLOCK_UNSAFE(mutex_ticket);
   SL_RETURN (retval, _("sl_create_ticket")); 
 }
 
@@ -1775,7 +1805,7 @@ SL_TICKET sl_make_ticket (const char * ofile, int oline,
 #define SL_OPEN_MAX          119
 
 #if !defined(O_NOATIME)
-#if defined(__linux__) && (defined(__i386__) || defined(__PPC__))
+#if defined(__linux__) && (defined(__i386__) || defined(__x86_64__) || defined(__PPC__))
 #define O_NOATIME 01000000
 #else
   /* 
@@ -2582,6 +2612,7 @@ int sl_read_timeout_fd (int fd, void * buf_in, size_t count,
 	    }  
 	  else if (byteread == 0)
 	    {
+	      /* zero indicates end of file */
 	      break;
 	    }
 	  else
@@ -2617,6 +2648,8 @@ int sl_read_timeout_fd (int fd, void * buf_in, size_t count,
 	      retry_fcntl(FIL__, __LINE__, fd, F_SETFL, sflags);
 	  TPT(( 0, FIL__, __LINE__, _("msg=<timeout>")));
 	  errno = 0;
+	  if (bytes > 0)
+	    return ((int) bytes); 
 	  return (SL_TIMEOUT);
 	}
       else
@@ -2647,6 +2680,8 @@ int sl_read_timeout_fd (int fd, void * buf_in, size_t count,
 	      retry_fcntl(FIL__, __LINE__, fd, F_SETFL, sflags);
 	  TPT(( 0, FIL__, __LINE__, _("msg=<timeout>")));
 	  errno = 0;
+	  if (bytes > 0)
+	    return ((int) bytes);
 	  return (SL_TIMEOUT);
 	}
     }
@@ -2801,10 +2836,12 @@ int sl_write (SL_TICKET ticket, const void * msg_in, long nbytes)
   /* write
    */
   bytecount    = 0;
-  bytewritten  = 0;
+
   while (bytecount < nbytes) 
     {    
-      if ((bytewritten = write (fd, msg, nbytes-bytecount)) > 0) 
+      bytewritten = write (fd, msg, nbytes-bytecount);
+
+      if (bytewritten > 0) 
 	{
 	  bytecount += bytewritten;
 	  msg       += bytewritten;    /* move buffer pointer forward */
