@@ -85,6 +85,9 @@
 #include "sh_calls.h"
 #else
 #define UID_CAST long
+#define HAVE_GETPWENT
+#define SH_MUTEX_LOCK(a)   ((void)0)
+#define SH_MUTEX_UNLOCK(a) ((void)0)
 #endif
 
 #include <stdio.h>
@@ -95,19 +98,25 @@
 #include <pwd.h>
 #include <string.h>
 #include <unistd.h>
+#include <errno.h>
 
 
 #ifndef TRUST_MAIN
+
 #include "slib.h"
 #define SH_NEED_PWD_GRP 1
 #include "sh_static.h"
+#include "sh_pthread.h"
 
 #else
 
 #define sh_getgrgid   getgrgid
-#define sh_getpwent   getpwent
+#define sh_getgrgid_r getgrgid_r
 #define sh_getpwnam   getpwnam
+#define sh_getpwnam_r getpwnam_r
 #define sh_getpwuid   getpwuid
+#define sh_getpwuid_r getpwuid_r
+#define sh_getpwent   getpwent
 #define sh_endpwent   endpwent
 
 #define TRUST_DEBUG
@@ -126,6 +135,7 @@ static int sl_errno = 0;
 #define SL_ETRUNC    -1026     /* Result truncated.                    */
 #define SL_EINTERNAL -1028     /* Internal error.                      */
 #define SL_EBADFILE  -1030     /* File access error. Check errno.      */
+#define SL_EMEM      -1032     /* Out of memory.                       */
 #define SL_EBADNAME  -1040     /* Invalid name.                        */
 #define SL_ESTAT     -1041     /* stat of file failed. Check errno.    */
 #define SL_EBADUID   -1050	/* Owner not trustworthy.              */
@@ -312,8 +322,8 @@ int dirz(char *path)
 			
 
 
-static 
-int getfname(char *fname, char *rbuf, int rsz)
+/* not static to circumvent stupid gcc 4 bug */ 
+int getfname(const char *fname, char *rbuf, int rsz)
 {
 #ifndef TRUST_MAIN
   register int status;
@@ -356,8 +366,10 @@ int getfname(char *fname, char *rbuf, int rsz)
       if (status != SL_ENONE)
 	SL_IRETURN(status, _("getfname"));
 #else
-      strncat(rbuf, "/",   rsz-strlen(rbuf));
-      strncat(rbuf, fname, rsz-strlen(rbuf));
+      strncat(rbuf, "/",   rsz-strlen(rbuf)-1);
+      rbuf[rsz-1] = '\0';
+      strncat(rbuf, fname, rsz-strlen(rbuf)-1);
+      rbuf[rsz-1] = '\0';
 #endif
     }
   SL_IRETURN(dirz(rbuf), _("getfname"));
@@ -401,39 +413,68 @@ int isin(uid_t n, uid_t *list)
 /* comment added by R. Wichmann
  *  RETURN TRUE if ANYONE in ulist is group member
  */
-static 
-int isingrp(gid_t grp, uid_t *ulist)
+/* not static to circumvent stupid gcc 4 bug */ 
+int isingrp(gid_t grp, uid_t *ulist, int * errval)
 {
-  register struct passwd *w;	/* info about group member */
+  struct passwd *w;	        /* info about group member */
   register uid_t *u;		/* points to current ulist member */
   register char **p;		/* points to current group member */
-  register struct group *g;	/* pointer to group information */
+  struct group *g;	        /* pointer to group information */
+
+  int status;
   
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  struct group    gr;
+  char          * buffer = NULL;
+  struct passwd   pwd;
+  char          * pbuffer = NULL;
+#endif
+
   SL_ENTER(_("isingrp"));
 
-  if ((g = sh_getgrgid(grp)) == NULL)
+  *errval = 0;
+
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  buffer = malloc(SH_GRBUF_SIZE);
+  status = sh_getgrgid_r(grp, &gr, buffer, SH_GRBUF_SIZE, &g);
+#else
+  errno = 0;
+  g = sh_getgrgid(grp);
+  status = errno;
+#endif
+
+  if (g == NULL)
     {
-      SL_IRETURN(SL_FALSE, _("isingrp") );
+      if (status == ERANGE)
+	*errval = status;
+
+      goto end_false;
     }
-  /*
-  if(g->gr_mem == NULL || g->gr_mem[0] == NULL )
-    SL_IRETURN(SL_FALSE, _("isingrp") );
-  */
 
   /* this will return at the first match
    */
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  pbuffer = malloc(SH_PWBUF_SIZE);
+#endif
+
   for(p = g->gr_mem; *p != NULL; p++)
     {
       for(u = ulist; *u != tf_uid_neg; u++)
 	{
 	  /* map user name to UID and compare */
-#ifdef TRUST_MAIN
-	  if ((w = getpwnam(*p)) != NULL && *u == (uid_t)(w->pw_uid) )
-	    SL_IRETURN(SL_TRUE, _("isingrp"));
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+	  sh_getpwnam_r(*p, &pwd, pbuffer, SH_PWBUF_SIZE, &w);
 #else
-	  if ((w = sh_getpwnam(*p)) != NULL && *u == (uid_t)(w->pw_uid) )
+	  w = sh_getpwnam(*p);
+#endif
+
+#ifdef TRUST_MAIN
+	  if (w != NULL && *u == (uid_t)(w->pw_uid) )
+	    goto end_true;
+#else
+	  if (w != NULL && *u == (uid_t)(w->pw_uid) )
 	    {
-	      SL_IRETURN(SL_TRUE, _("isingrp"));
+	      goto end_true;
 	    }
 #endif
 	}
@@ -443,47 +484,90 @@ int isingrp(gid_t grp, uid_t *ulist)
    */
   for(u = ulist; *u != tf_uid_neg; u++)
     {
-#ifdef TRUST_MAIN
-      if ((w = getpwuid(*u)) != NULL && grp == (gid_t)(w->pw_gid) )
-	SL_IRETURN(SL_TRUE, _("isingrp"));
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWUID_R)
+      sh_getpwuid_r(*u, &pwd, pbuffer, SH_PWBUF_SIZE, &w);
 #else
-      if ((w = sh_getpwuid(*u)) != NULL && grp == (gid_t)(w->pw_gid) )
+      w = sh_getpwuid(*u);
+#endif
+#ifdef TRUST_MAIN
+      if (w != NULL && grp == (gid_t)(w->pw_gid) )
+	goto end_true;
+#else
+      if (w != NULL && grp == (gid_t)(w->pw_gid) )
 	{
-	  SL_IRETURN(SL_TRUE, _("isingrp"));
+	  goto end_true;
 	}
 #endif
     }
 
+ end_false:
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  if (buffer)  free(buffer);
+  if (pbuffer) free(pbuffer);
+#endif
   SL_IRETURN(SL_FALSE, _("isingrp"));
+
+ end_true:
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  if (buffer)  free(buffer);
+  if (pbuffer) free(pbuffer);
+#endif
+  SL_IRETURN(SL_TRUE, _("isingrp"));
 }
 
 /* added by R. Wichmann Fri Mar 30 08:16:14 CEST 2001
  *  RETURN TRUE only if ALL group members are trusted
  */
-static 
-int onlytrustedingrp(gid_t grp, uid_t *ulist)
+/* not static to circumvent stupid gcc 4 bug */ 
+int onlytrustedingrp(gid_t grp, uid_t *ulist, int * errval)
 {
-  register struct passwd *w;	/* info about group member */
+  struct passwd *w;	        /* info about group member */
   register uid_t *u;		/* points to current ulist member */
   register char **p;		/* points to current group member */
-  register struct group *g;	/* pointer to group information */
+  struct group *g;	        /* pointer to group information */
   register int flag = -1;       /* group member found */
 
+  int status;
+
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  struct group    gr;
+  char          * buffer  = NULL;
+  struct passwd   pw;
+  char          * pbuffer = NULL;
+#endif
+
+  int retval = SL_FALSE;
+
   SL_ENTER(_("onlytrustedingrp"));
 
+  *errval = 0;
+
 #ifdef TRUST_DEBUG
   fprintf(stderr, "trustfile: group writeable, group_gid: %ld\n", 
 	  (UID_CAST)grp); 
 #endif
 
-  if ((g = sh_getgrgid(grp)) == NULL)
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  buffer = malloc(SH_GRBUF_SIZE);
+  status = sh_getgrgid_r(grp, &gr, buffer, SH_GRBUF_SIZE, &g);
+#else
+  errno = 0;
+  g = sh_getgrgid(grp);
+  status = errno;
+#endif
+
+  if (g == NULL)
     {
+      if (status == ERANGE)
+	*errval = status;
+
 #ifdef TRUST_DEBUG
       fprintf(stderr, 
 	      "trustfile: group_gid: %ld, no such group --> ERROR\n", 
 	      (UID_CAST)grp); 
 #endif
-      SL_IRETURN(SL_FALSE, _("onlytrustedingrp") );
+      retval = SL_FALSE;
+      goto end_retval;
     }
 
   /* empty group -> no problem
@@ -494,6 +578,10 @@ int onlytrustedingrp(gid_t grp, uid_t *ulist)
 
   /* check for untrusted members of the group
    */
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+  pbuffer = malloc(SH_PWBUF_SIZE);
+#endif
+
   for(p = g->gr_mem; *p != NULL; p++)
     {
       flag = -1;
@@ -502,11 +590,12 @@ int onlytrustedingrp(gid_t grp, uid_t *ulist)
 #endif
       /* map user name to UID and compare 
        */
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETPWNAM_R)
+      sh_getpwnam_r(*p, &pw, pbuffer, SH_PWBUF_SIZE, &w);
+#else
       w = sh_getpwnam(*p);
-#ifndef TRUST_MAIN
-      if (!w)
-	w = sh_getpwnam(*p);
 #endif
+
       if (w == NULL)    /* not a valid user, ignore    */
 	{
 	  flag = 0; 
@@ -550,14 +639,24 @@ int onlytrustedingrp(gid_t grp, uid_t *ulist)
 		   "trustfile: user=%s (gid %ld), not a trusted user --> ERROR\n", *p, (UID_CAST)grp);
 #endif 
 	  tf_baduid = w->pw_uid;
-	  SL_IRETURN(SL_FALSE, _("onlytrustedingrp"));
+	  retval = SL_FALSE;
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+	  if (pbuffer) free(pbuffer);
+#endif
+	  goto end_retval;
 	}
     }
 
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  if (pbuffer) free(pbuffer);
+#endif
+
 #ifndef TEST_ONLY	
 #ifdef HAVE_GETPWENT
   /* now check ALL users for their GID !!!
    */
+  SH_MUTEX_LOCK(mutex_pwent);
+
   while (NULL != (w = sh_getpwent())) 
     {
       if (grp == (gid_t)(w->pw_gid))
@@ -576,7 +675,7 @@ int onlytrustedingrp(gid_t grp, uid_t *ulist)
 #ifdef TRUST_DEBUG
 		  fprintf (stderr, 
 			   "trustfile: uid=%ld, trusted_uid=%ld, match found --> OK\n", 
-			   (UID_CAST)(w->pw_uid), *u);
+			   (UID_CAST)(w->pw_uid), (UID_CAST)(*u));
 #endif 
 		  flag = 0;
 		  break;
@@ -598,41 +697,61 @@ int onlytrustedingrp(gid_t grp, uid_t *ulist)
 	      fprintf(stderr,"trustfile: group member %s not found in trusted users --> ERROR\n", w->pw_name); 
 #endif
 	      tf_baduid = w->pw_uid;
-	      SL_IRETURN(SL_FALSE, _("onlytrustedingrp"));
+	      retval = SL_FALSE;
+	      goto out;
+	      /* SL_IRETURN(SL_FALSE, _("onlytrustedingrp")); */
 	    }
 	}
     }
+  retval = SL_TRUE;
+
+ out:
 
 #ifdef HAVE_ENDPWENT
   sh_endpwent();
 #endif
+
+  SH_MUTEX_UNLOCK(mutex_pwent);
+
   /* TEST_ONLY */
 #endif
-
   /* #ifdef HAVE_GETPWENT */
 #endif
 
 #ifdef TRUST_DEBUG
-  fprintf(stderr,
-	  "trustfile: group %ld:  all members are trusted users --> OK\n", 
-	  (UID_CAST)grp);
+  if (retval == SL_TRUE)
+    fprintf(stderr,
+	    "trustfile: group %ld:  all members are trusted users --> OK\n", 
+	    (UID_CAST)grp);
 #endif
   /* all found
    */
-  SL_IRETURN(SL_TRUE, _("onlytrustedingrp"));
+ end_retval:
+#if defined(HAVE_PTHREAD) && defined (_POSIX_THREAD_SAFE_FUNCTIONS) && defined(HAVE_GETGRGID_R)
+  if (buffer)  free(buffer);
+#endif
+  SL_IRETURN(retval, _("onlytrustedingrp"));
 }
 
-int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
+int sl_trustfile(const char *fname, uid_t *okusers, uid_t *badusers)
 {
-  char fexp[MAXFILENAME];	/* file name fully expanded        */
-  register char *p = fexp;      /* used to hold name to be checked */
+  char * fexp = NULL;	        /* file name fully expanded        */
+  register char *p;             /* used to hold name to be checked */
   struct stat stbuf;	        /* used to check file permissions  */
   char c;			/* used to hold temp char          */
   
+  int errgrp = 0;
+
   SL_ENTER(_("sl_trustfile"));
   if (fname == NULL)
     SL_IRETURN(SL_EBADFILE, _("sl_trustfile"));
 
+  fexp = malloc( MAXFILENAME );
+  if (!fexp)
+    SL_IRETURN(SL_EMEM, _("sl_trustfile"));
+
+  p = fexp;
+
   /*
    * next expand to the full file name
    * getfname sets sl_errno as appropriate
@@ -640,10 +759,16 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 #ifdef TRUST_MAIN
   sl_errno = getfname(fname, fexp, MAXFILENAME);
   if (sl_errno != 0)
-    return sl_errno;
+    {
+      free(fexp);
+      return sl_errno;
+    }
 #else
   if (SL_ISERROR(getfname(fname, fexp, MAXFILENAME)))
+    {
+      free(fexp);
       SL_IRETURN(sl_errno, _("sl_trustfile"));
+    }
 #endif
 
   if (okusers == NULL && badusers == NULL)
@@ -686,13 +811,15 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
        */
       if (retry_lstat(FIL__, __LINE__, fexp, &stbuf) < 0)
 	{
-	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
+	  (void) strncpy(tf_path, fexp, sizeof(tf_path));
+	  tf_path[sizeof(tf_path)-1] = '\0';
 #ifdef TRUST_MAIN
 	  fprintf(stderr, "---------------------------------------------\n");
-	      fprintf(stderr, "trustfile: ESTAT: stat(%s) failed, maybe the file does not exist\n",
-		      fexp);
+	  fprintf(stderr, "trustfile: ESTAT: stat(%s) failed,\n", fexp);
+	  fprintf(stderr, "maybe the file does not exist\n");
 	  fprintf(stderr, "---------------------------------------------\n");
 #endif
+	  free(fexp);
 	  SL_IRETURN(SL_ESTAT, _("sl_trustfile"));
 	}
 
@@ -719,11 +846,13 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	   * to the CWD) are interpreted correctly.
 	   * got it?
 	   */
-	  char csym[MAXFILENAME];	/* contents of symlink file  */
-	  char full[MAXFILENAME];	/* "full" name of symlink    */
+	  char * csym;	                /* contents of symlink file  */
+	  char * full;	                /* "full" name of symlink    */
 	  register char *b, *t;	        /* used to copy stuff around */
 	  register int lsym;	        /* num chars in symlink ref  */
 	  register int i;		/* trustworthy or not?       */
+	  const char * t_const;
+	  char *end;
 
 	  /*
 	   * get what the symbolic link points to
@@ -733,6 +862,13 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	   * (MAXFILENAME instead of MAXFILENAME-1)
 	   * R.W. Tue May 29 22:05:16 CEST 2001
 	   */
+	  csym = malloc( MAXFILENAME );
+	  if (!csym)
+	    {
+	      free(fexp);
+	      SL_IRETURN(SL_EMEM, _("sl_trustfile"));
+	    }
+
 	  lsym = readlink(fexp, csym, MAXFILENAME-1);
 	  if (lsym >= 0) 
 	    csym[lsym] = '\0';
@@ -744,14 +880,28 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 		      fexp);
 	      fprintf(stderr, "---------------------------------------------\n");
 #endif
+	      free(csym);
+	      free(fexp);
 	      SL_IRETURN(SL_EBADNAME, _("sl_trustfile"));
 	    }
 
+	  full = malloc( MAXFILENAME );
+	  if (!full)
+	    {
+	      free(csym);
+	      free(fexp);
+	      SL_IRETURN(SL_EMEM, _("sl_trustfile"));
+	    }
+
 	  /*
 	   * relative or absolute referent?
 	   */
 	  if (csym[0] != '/')
 	    {
+	      /* pointer to one above last element
+	       */
+	      end = &full[MAXFILENAME-1]; ++end;
+
 	      /* initialize pointers 
 	       */
 	      b = full;
@@ -759,28 +909,29 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	      /* copy in base path 
 	       */
 	      t = fexp;
-	      while(*t && b < &full[MAXFILENAME])
+	      while(*t && b < end)
 		*b++ = *t++;
 
 	      /* smack on the /../ 
 	       */
-	      t = "/../";
-	      while(*t && b < &full[MAXFILENAME])
+	      t_const = "/../"; t = (char *)t_const;
+	      while(*t && b < end)
 		*b++ = *t++;
 
 	      /* append the symlink referent 
 	       */
 	      t = csym;
-	      while(*t && b < &full[MAXFILENAME])
+	      while(*t && b < end)
 		*b++ = *t++;
 
 	      /* see if we're too big 
 	       */
-	      if (*t || b == &full[MAXFILENAME])
+	      if (*t || b == end)
 		{
 		  /* yes -- error 
 		   */
-		  (void) strcpy(tf_path, fexp);          /* known to fit  */
+		  (void) strncpy(tf_path, fexp, sizeof(tf_path));
+		  tf_path[sizeof(tf_path)-1] = '\0';
 #ifdef TRUST_MAIN
 		  fprintf(stderr, "---------------------------------------------\n");
 		  fprintf(stderr, 
@@ -788,6 +939,9 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 			  fexp);
 		  fprintf(stderr, "---------------------------------------------\n");
 #endif
+		  free(full);
+		  free(csym);
+		  free(fexp);
 		  SL_IRETURN(SL_ETRUNC, _("sl_trustfile"));
 		}
 	      *b = '\0';
@@ -803,7 +957,12 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	   * now check out this file and its ancestors
 	   */
 	  if ((i = sl_trustfile(full, okusers, badusers)) != SL_ENONE)
-	    SL_IRETURN(i, _("sl_trustfile"));
+	    {
+	      free(full);
+	      free(csym);
+	      free(fexp);
+	      SL_IRETURN(i, _("sl_trustfile"));
+	    }
 
 	  /*
 	   * okay, this part is valid ... let's check the rest
@@ -822,6 +981,8 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	      if (*p)
 		p++;
 	    }
+	  free(full);
+	  free(csym);
 	  continue;
 	}
 #endif
@@ -849,8 +1010,11 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	  fprintf(stderr, "where UID is the UID of the (yet) untrusted user.\n"); 
 	  fprintf(stderr, "---------------------------------------------\n");
 #endif 
-	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
+	  (void) strncpy(tf_path, fexp, sizeof(tf_path));
+	  tf_path[sizeof(tf_path)-1] = '\0';
+
 	  tf_baduid = (uid_t) stbuf.st_uid;
+	  free(fexp);
 	  SL_IRETURN(SL_EBADUID, _("sl_trustfile"));
 	}
 
@@ -866,8 +1030,8 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
        * (untrusted) users in the group
        */
       if (((stbuf.st_mode & S_IWGRP) == S_IWGRP) &&
-	  ((okusers != NULL && !onlytrustedingrp((gid_t)stbuf.st_gid,okusers))||
-	   (badusers != NULL && isingrp((gid_t)stbuf.st_gid, badusers)))
+	  ((okusers != NULL && !onlytrustedingrp((gid_t)stbuf.st_gid,okusers,&errgrp))||
+	   (badusers != NULL && isingrp((gid_t)stbuf.st_gid, badusers,&errgrp)))
 #ifdef STICKY
 	  && ((stbuf.st_mode&S_IFDIR) != S_IFDIR ||
 	      (stbuf.st_mode&S_ISVTX) != S_ISVTX)
@@ -886,9 +1050,12 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	  fprintf(stderr, "where UID is the UID of the (yet) untrusted user.\n"); 
 	  fprintf(stderr, "---------------------------------------------\n");
 #endif 
-	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
+	  (void) strncpy(tf_path, fexp, sizeof(tf_path));
+	  tf_path[sizeof(tf_path)-1] = '\0';
+
 	  tf_badgid = (gid_t) stbuf.st_gid;
-	  SL_IRETURN(SL_EBADGID, _("sl_trustfile"));
+	  free(fexp);
+	  SL_IRETURN((errgrp == ERANGE) ? SL_ERANGE : SL_EBADGID, _("sl_trustfile"));
 	}
       /*
        * if other can write, bomb; but if the sticky
@@ -908,7 +1075,10 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 	  fprintf(stderr, "This file/directory is world writeable.\n");
 	  fprintf(stderr, "---------------------------------------------\n");
 #endif 
-	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
+	  (void) strncpy(tf_path, fexp, sizeof(tf_path));
+	  tf_path[sizeof(tf_path)-1] = '\0';
+
+	  free(fexp);
 	  SL_IRETURN(SL_EBADOTH, _("sl_trustfile"));
 	}
       /*
@@ -931,11 +1101,21 @@ int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
   /*
    * yes, it can be trusted
    */
-  (void) strcpy(tf_path, fexp);                      /* known to fit  */
+  (void) strncpy(tf_path, fexp, sizeof(tf_path));
+  tf_path[sizeof(tf_path)-1] = '\0';
+
+  free(fexp);
   SL_IRETURN(SL_ENONE, _("sl_trustfile"));
 }
 
 #ifdef TRUST_MAIN
+
+#if defined(HOST_IS_CYGWIN) || defined(__cygwin__) || defined(__CYGWIN32__) || defined(__CYGWIN__)
+int main()
+{
+  return 0;
+}
+#else
 int main (int argc, char * argv[])
 {
   int status;
@@ -950,7 +1130,7 @@ int main (int argc, char * argv[])
 
   tf_path[0] = '\0';
 #if defined(SH_WITH_SERVER)
-  pass = getpwnam(SH_IDENT);
+  pass = sh_getpwnam(SH_IDENT);  /* TESTONLY */
   if (pass != NULL)
     tf_euid = pass->pw_uid;
   else
@@ -973,6 +1153,7 @@ int main (int argc, char * argv[])
   return 0;
 }
 #endif
+#endif
 
 
 
